#!/usr/bin/perl -w
use Getopt::Long qw(GetOptions);
use POSIX qw(ceil);
use List::Util qw[min max];
use File::Spec;

# Program AIS_NMEA.pl

## See README.TXT for background commentary and version information. ##

# (c) Gary C. Kessler, 2018-2023

$build_date = "01/02/2023";
$version = "5.4.4";

# Per NMEA 0183 v4.11 (pp. 176-177, footnote 4):
# 1) Serial number must be in range 0-9. This affects $serial
# 2) Single-sentence messages can be up to 62 characters (372 bits) in length; multi-sentence message payloads
#      are <=60 characters (360 bits). This affects $max_fragment_size and $max_single_fragment_message
# 3) Maximum message size is 540 characters (3240 bits). This affects $max_message_size

$max_message_size = 3240; # 540 six-bit characters
$max_fragment_size = 360; # 60 six-bit characters
$max_single_fragment_message = 372; # 62 six-bit characters

# Set default values for all switches so that absent values don't cause "uninitialized value" errors

$ais_sent = "!AIVDM";
$batch = 0;  # Assume batch mode is OFF
$channel = "A";
$in_string = "";

# Parse the command line to get message parameters

GetOptions
  (
  "help" => \$help,
  "ais_sent=s" => \$ais_sent,
  "batch" => \$batch,
  "channel=s" => \$channel,
  "payload=s" => \$in_string,
  )
  or die "Usage: $0 --help for command line assistance\n";

if ($help)
  {
  help ();
  exit;
  }

# Set the output based upon the --batch switch. If in batch mode, only direct command to print

if ($batch == 0)
    {
    *OUTDEV = STDOUT;
    open (OUTDEV, ">");
    }
  else
    {
    open (OUTDEV, ">", File::Spec->devnull);
    }

print OUTDEV "\n***** AIS NMEA (Version: $version, Build date: $build_date) *****\n\n";

$strlen = length ($in_string);

# If string length not an even multiple of 6, pad it...

$pad = (6 - ($strlen % 6)) % 6;

if ($pad > 0)
  {
  print OUTDEV "Input (length=$strlen bits) is not 6-bit aligned... padding added\n\n";
  $in_string .= "0" x $pad;
  $strlen = $strlen + $pad;
  }

# If message is too big, halt the program...

if ($strlen > $max_message_size)
  {
  printf OUTDEV "Maximum message size of %d bytes (%d bits) exceeded... %d bits\n\n", $max_message_size/6, $max_message_size, $strlen;
  exit 1;
  }

printf OUTDEV "AIS sentence type: %s   Channel: %s   Payload length: %d bits (%d 6-bit words)  Pad: %d bit(s)\n\n", $ais_sent, $channel, $strlen, $strlen/6, $pad;

# Determine the number of fragments needed...
# If more than one fragment, set $serial to be a value 0-9

if ($strlen <= $max_single_fragment_message)
    { $num_frags = 1; }
  else
    {
    $num_frags = ceil ($strlen / $max_fragment_size);
    ($sec,$x,$x,$x,$x,$x,$x,$x,$x) = localtime(time);
    $serial = $sec % 10;
    }
print OUTDEV "$num_frags fragment(s)";
if ($num_frags > 1)
  { print OUTDEV ", using serial number $serial"; }
print OUTDEV "\nBinary payload: $in_string\n\n";

# Error checking. Ensure that string is composed of 0s and 1s; NMEA value is 0 or 1; and
# Channel is A or B

for ($i=0; $i<$strlen; $i++)
  {
  $tyui = substr ($in_string, $i, 1);
  if ($tyui ne "0" && $tyui ne "1")
    {
    print OUTDEV "Input contains non-binary value... $tyui\n";
    exit;
    }
  }

if ($channel ne "A" && $channel ne "B")
  {
  print OUTDEV "Channel must be A or B... $channel\n";
  exit;
  }

print OUTDEV "Armored ASCII message: ";

# Manipulate $in_string from binary to the NMEA 6-bit "armored" ASCII

for ($i=0; $i<$strlen/6; $i++)
  {

# One character = 6 bits, so get the 6-bit block

  $in_buffer [$i] = substr ($in_string, $i*6, 6);

# Convert the 6 binary bits to an integer value

  $in_buffer [$i] = convert ($in_buffer [$i], 6);

#  Now convert the value to an ASCII character

  if ($in_buffer [$i] > 39)
    { $in_buffer [$i] += 8; }
  $in_buffer [$i] += 48;

  $in_buffer [$i] = chr ($in_buffer [$i]);

  print OUTDEV $in_buffer [$i];
  }

print OUTDEV "\n\n";

# Prepare AIS Message output
# Each fragment contains:
#  1. AIS sentence type indicator
#  2. Fragment information
#  3. ASCII string for data
#  4. Compute and add checksum

print OUTDEV "AIS Sentence(s):\n";

for ($i=1; $i<=$num_frags; $i++)
  {
  $AIS_message = $ais_sent . ",";

  if ($num_frags == 1)
      {
      $AIS_message .= "1,1,," . $channel . ",";
      $str_strt = 0; $str_end = $strlen/6;
      }
    else
      {
      $AIS_message .= $num_frags . "," . $i . "," . $serial . "," . $channel . ",";
      $str_strt = ($i-1) * $max_fragment_size/6;
      $str_end = $str_strt + min ($max_fragment_size/6, $strlen/6 - $str_strt);
      }

  for ($j=$str_strt; $j<$str_end; $j++)
    { $AIS_message .= $in_buffer [$j]; }

# Pad value will be 0 on all fragments except the only one (num_frags=1) or last one in a set, in which case we actually
# use the $pad value

  if ($num_frags == 1 || $i == $num_frags)
      { $AIS_message .= "," . $pad; }
    else
      { $AIS_message .= ",0"; }

# Compute the checksum over the entirety of the message up to this point...

  $checksum = nmea_checksum ($AIS_message);
  $AIS_message .= "*" . $checksum;
  print "$AIS_message\n";
  }

print OUTDEV "\nDone!\n";


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# nmea_checksum
# ********************************************************

# The NMEA checksum is computed on the entire sentence including the AIVDM/AIVDO tag but excluding the leading "!"
# The checksum is merely a byte-by-byte XOR of the sentence

sub nmea_checksum
  {
  my ($sentence) = @_;
  my ($char, $i, $n, $len, $sum);

  $len = length ($sentence);
  $sum = 0;

  if (substr ($sentence,0,1) eq "!")
      { $n = 1; }
    else
      { $n = 0; }

  for ($i=$n; $i<$len; $i++)
    { $sum ^= ord (substr ($sentence,$i,1)); }

# Ensure that $sum is two upper-case hex digits

  $sum = sprintf ("%02X", $sum);
  return $sum;
  }

# ********************************************************
# convert
# ********************************************************

# Converts six binary bits to an integer value

sub convert
  {
  my ($str, $n) = @_;
  my ($j, $sum);

  $sum=0;

  for ($j=0; $j<$n; $j++)
    { 
    if (substr ($str,$j,1) == 1) { $sum += 2 ** ($n-$j-1); }
    }

  return $sum;
  }

# ********************************************************
# help
# ********************************************************

# Display the help file

sub help
{
print<< "EOT";
Program usage:
  AIS_NMEA.pl --ais_sent=TYPE --payload=STRING --channel=CHAN
  AIS_NMEA.pl --help

 where:
  --ais_sent   AIS sentence type (default = '!AIVDM')
  --payload    Binary string composing the AIS sentence
  --channel    AIS communication channel (A [default] or B)
  --help       Display the help file

Sample program use:

  AIS_NMEA.pl --ais_sent='!AIVDO' --payload=000001001000100101010111001011100
  000000000000000000001000010010111001111100101111001000100010101101100011000
  101001001101101011001011000111010000000000000101000001010000 --channel=B

EOT
return;
}

