#!/usr/bin/perl
use List::Util qw[min];
binmode (STDOUT,":utf8");

# Program AIS_parser.pl
# (c) Gary C. Kessler, 2018-2023

# AIS sentence parser, compliant with NMEA 0183
## See README.TXT for background commentary and version information. ##

# See also https://gpsd.gitlab.io/gpsd/AIVDM.html for more information
# on the parameters used in the command line, as well as ITU-R Rec. M.1371-5, Annex 8.

$build_date = "11/09/2023";
$version = "4.8.6";

setparms ();

print "\n***** AIS Parser (Version: $version, Build date: $build_date) *****\n\n";

# Parse the command line to ensure that the parameters are error-free.

$message_num = 0;
($switch, $parameter) = parsecommandline ();

# We are here if there are exactly two arguments. Case statement below aims to convert whatever input we have
# to a binary string ($message)

ARGS:
  {

# Binary string on command line ("-b").

  $switch eq "-b" && do
    {
    $message = $parameter;
    $msg_len = length ($message);
    print "Binary input from command line. Input length: $msg_len bits\n";
    print_binary_string ($message, $msg_len);
    start_parse ($message, $msg_len, 0);
    };

# Binary string in file ("-bf")

  $switch eq "-bf" && do
    {
    $filename = $parameter;
    open (INFILE, '<:encoding(UTF-8)', $filename)
      or die "Could not open file '$filename'\n\n";

    $message = <INFILE>;
    chomp $message;
    close INFILE;

    $msg_len = length ($message);
    print "Binary input from file: $filename Input length: $msg_len bits\n";
    print_binary_string ($message, $msg_len);
    start_parse ($message, $msg_len, 0);
    };

# !AIVDM sentence on command line ("-s")

  $switch eq "-s" && do
    {
    $line = $parameter;
    print "AIS sentence input from command line.\n";
    $message = process_sentence ($line);
    $msg_len = length ($message);
    print_binary_string ($message, $msg_len);
    start_parse ($message, $msg_len, 0);
    };

# One or more !AIVDM sentences in file ("-sf")

  $switch eq "-sf" && do
    {
    $filename = $parameter;
    print "AIS sentence(s) input from file $filename.\n\n";
    open (INFILE, '<:encoding(UTF-8)', $filename)
      or die "Could not open file '$filename'\n\n";

    while (not (eof INFILE))
      {
      $message = "";
      $numlinesin = 0;
      $line = <INFILE>;
      chomp $line;
      $numlinesin++;
      $tempmsg = process_sentence ($line);
      $message .= $tempmsg;

# Loop here in case this message has more than a single sentence

      @fields_main = split (",", $line);
      if ($fields_main[1] > 1)
        {
        for ($i=2; $i<=$fields_main[1]; $i++)
          {
          $line = <INFILE>;
          chomp $line;
          $numlinesin++;
          $tempmsg = process_sentence ($line);
          $message .= $tempmsg;
          }
        }

      print "$numlinesin sentence(s) processed\n\n";
      $msg_len = length ($message);
      print_binary_string ($message, $msg_len);
      $message_num++;
      start_parse ($message, $msg_len, $message_num);
      print "****************************************************\n";
      };

    };
    close INFILE;
  }

if ($switch eq "-sf")
  { print "\n--- End of File ---\n\n"; }

# PROGRAMMING NOTE: In the routines below, there is a lot of use of a variable named $tyui. In most of the program,
# $tyui represents a value that is used one time in a localized section of the code. The use of the value can be found
# in the context of the value being requested and the $command string switch label.

# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# start_parse
# ********************************************************

# Here is where we start the parsing process

sub start_parse
  {
  my ($message,$msg_len,$msg_num) = @_;
  my ($mmsi, $type);

  if ($msg_num == 0)
      { print "--- Start of Transmission ---\n\n"; }
    else
      { print "--- Start of Transmission $msg_num ---\n\n"; }

# All files begin with the same three fields...

################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI

  $type = oct ("0b".substr ($message,0,6));
  printf "Message Type = %d (%s)\n", $type, $message_type [$type];

# LEGACY CODE... from the beginning...

  if (not ($type >= 1 && $type <= 27))
    {
    print "\n*** Message type $type not supported ***\n\n";
    return;
    }

  printf "Repeat indicator = %d\n", oct ("0b".substr ($message,6,2));

# Note that $mmsi is a global variable

  $mmsi = sprintf ("%09d", oct ("0b".substr ($message,8,30)));
  mmsi_handler ($mmsi);

## Case statement -- decode message based upon type

# First -- $message is the raw bit string INCLUDING the padding bits, added to ensure 6-bit alignment. For purposes of decoding, we use the
# "trimmed" message, the result of stripping the $num_pad trailing bits from $message

  $trimmed_len = $msg_len - $num_pad;
  $trimmed = substr ($message,0,$trimmed_len);

  DECODE:
    {
    $type >= 1 && $type <= 3 && do
      {
      decode_010203 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    ($type == 4 || $type == 11) && do
      {
      decode_0411 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 5 && do
      {
      decode_05 ($trimmed,$trimmed_len);
      last DECODE;
      };
    ($type == 6 || $type == 8) && do
      {
      decode_0608 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    ($type == 7 || $type == 13) && do
      {
      decode_0713 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 9 && do
      {
      decode_09 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 10 && do
      {
      decode_10 ($trimmed,$trimmed_len);
      last DECODE;
      };
    ($type == 12 || $type == 14) && do
      {
      decode_1214 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 15 && do
      {
      decode_15 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 16 && do
      {
      decode_16 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 17 && do
      {
      decode_17 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 18 && do
      {
      decode_18 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 19 && do
      {
      decode_19 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 20 && do
      {
      decode_20 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 21 && do
      {
      decode_21 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 22 && do
      {
      decode_22 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 23 && do
      {
      decode_23 ($trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 24 && do
      {
      decode_24 ($mmsi,$trimmed,$trimmed_len);
      last DECODE;
      };
    ($type == 25 || $type == 26) && do
      {
      decode_2526 ($type,$trimmed,$trimmed_len);
      last DECODE;
      };
    $type == 27 && do
      {
      decode_27 ($trimmed,$trimmed_len);
      last DECODE;
      };
    };

  print "\n--- End of Transmission ---\n\n";
  return;
  }

# ********************************************************
# signed_bin2dec
# ********************************************************

# Accepts binary string and returns signed integer value

sub signed_bin2dec
  {
  my ($binstr) = @_;
  my ($sign, $value, $decimal);
  
  $sign = oct ("0b".substr ($binstr,0,1));
  $value = oct ("0b".substr ($binstr,1,length($binstr)-1));
  if ($sign == 0)
      { $decimal = $value; }
    else
      { $decimal = -1 * (2**(length($binstr)-1) - $value); }
  return $decimal;
  }

# ********************************************************
# decode_string
# ********************************************************

# Accept a binary string representing six-bit ASCII characters and convert to ASCII

sub decode_string
  {
# Parameter list
  my ($message) = @_;
  my ($i,$n,$len);
  my $vocabulary = '@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_ !"#$%&\'()*+,-./0123456789:;<=>?';
  my $string = "";

  $len = length ($message);

  for ($i=0; $i<$len; $i=$i+6)
    {
    $n = substr ($vocabulary,oct ("0b".substr($message,$i,6)),1);
    $string = $string . sprintf ("%s", $n);
    }

  return $string;
  }

# ********************************************************
# default_handler
# ********************************************************

# Many of the values have a default that indicates that information is not available. This common routine will handle those simple cases.

sub default_handler
  {
  my ($label,$value,$units,$default) = @_;

  printf $label, $value;
  if ($value == $default)
      { print " (Information not available)\n"; }
    else
      { print "$units\n"; }

  return;
  }

# *********************************************
# air_press_handler
# ********************************************************

# Air pressure is a bit odd because there is a minimum, maximum, and "not available" default, AND the value has to be changed from the encoding.
# Also, different message types use a different "not available" default value....

sub air_press_handler
  {
  my ($label,$value,$unknown) = @_;

  print $label;
  if ($value == $unknown)
      { printf " %d (Information not available)", $value; }
    else
      {
      printf "%d hPa [millibar]", $value+799;
      if ($value == 0)
          { print " or less"; }
        else
          {
          if ($value == 402)
            { print " or more"; }
          }
      };
  print "\n";
  return;
  }

# ********************************************************
# dimension_handler
# ********************************************************

# Handle the bow, stern, port, and starboard dimensions

sub dimension_handler
  {
  my ($label,$value,$big) = @_;

  printf "Dimension to %s = %d", $label, $value;
  if ($value > 0)
      {
      print " m.";
      if ($value == $big)
        { print " or greater"; }
      }
    else
      { print " (Information not available)"; }

  print "\n";

  return;
  }

# ********************************************************
# display_ascii
# ********************************************************

# Print output of the data bytes to show both hex and ASCII values.
# Binary string is in $bitstring, which is extracted from the larger $message string, starting at bit position $strt, for $n bits
# Display $linemax bits per line

sub display_ascii
  {
  my ($message,$strt,$n,$linemax) = @_;
  my ($bitstring,$bpos,$j,$k,$line_len);
  my $pad = 0;
  my $end = $strt + $n - 1;
  
# Extract the bitstring from the larger message. Pad bitstring to be octet-aligned, if necessary

  $bitstring = substr ($message,$strt,$n);
  if ($n % 8 != 0)
    {
    $pad = 8 - (($n % 8) % 8);
    $bitstring .= "0" x $pad ;
    $n += $pad;
    }
  
  $bpos = 0;
  printf "8-bit ASCII (Message bits %d-%d)", $strt, $end;
  if ($pad != 0) { print " plus $pad padding bit(s)"; }
  print "\n";
  $lines = $n / $linemax;
  if ($lines != int ($lines)) { $lines = int ($lines) + 1; }
 
  for ($i=0; $i<$lines; $i++)
    {
    $line_len = min ($linemax,$n-$bpos);
    printf ("  **  %04d: ", $strt+$bpos);
    for ($j=0; $j<$line_len; $j+=8)
      { printf (" %02X", oct ("0b".substr ($bitstring,$bpos+$j,8))); }
    if ($line_len < $linemax)
      {
      for ($j=1; $j<=$linemax-$line_len; $j+=8) { print "   "; };
      }

    print "   ";
    for ($j=0; $j<$line_len; $j+=8)
      {
      $k = oct ("0b".substr ($bitstring,$bpos+$j,8));
  
#   Print a dot (.) for unprintable characters

      if ($k <= 0x1f || ($k >= 0x7f && $k <= 0x9f))
          { print "."; }
        else
          { print chr($k); }
      }

    print "\n";
    $bpos += $linemax;
    };

  print "\n";
  return;
  }

# ********************************************************
# long_lat_handler
# ********************************************************

# Display the latitude and longitude with leading zeros and N/S or E/W, AND handle the default case
# For some reason, NMEA does things in "long, lat" order, so...

# NOTE: Need to put the line 'binmode (STDOUT,":utf8");' early in the program
# to support output of UTF-8 so that we can use a degree symbol!

sub long_lat_handler
  {
  my ($label_long,$long,$default_long,$label_lat,$lat,$default_lat) = @_;

  printf "%s =", $label_long;
  if ($long == $default_long)
      { printf " %.1f (Information not available)\n", $long; }
    else
      { long_lat_display ($long,1); }

  printf "%s =", $label_lat;
  if ($lat == $default_lat)
      { printf " %.1f (Information not available)\n", $lat; }
    else
      { long_lat_display ($lat,0); }

  if ($long ne $default_long && $lat ne $default_lat)
    {
    print "  ** Use this URL for a Google map of this position:\n";
    printf "  ** https://www.google.com/maps/place/%.6f,%.6f\n", $lat,$long;
    }

  return;
  }

# ********************************************************
# Display the latitude or longitude in both GPS and
# degrees/minutes (DD MM.MMMMM)
# ********************************************************

sub long_lat_display
  {
  my ($pos,$n) = @_;
  my ($dir,$format,$min,$deg,$string);
  my @label = (
    ["N", "S"],
    ["E", "W"],
    );

# $pos is the given latitude or longitude (in degrees)
# $n = 0 for latitude, 1 for longitude (relates to # of required digits)

  $format = " %0" . ($n+9) . ".6f\xB0%s";
  if ($pos >= 0)
      { $dir = $label [$n][0]; }
    else
      { $dir = $label [$n][1]; }
  printf $format, abs ($pos), $dir;

  $deg = abs(int($pos));
  $min = (abs($pos) - $deg) * 60;

  $format = " (%0" . ($n+2) . "d\xB0%05.2f'%s)\n";
  printf ($format, $deg, $min,$dir);
  return;
  }

# ********************************************************
# mid_handler
# ********************************************************

# Display the country from the MID

sub mid_handler
  {
  my ($mid) = @_;
  my ($country, @fields, $line,$name,$num);

  $country = "invalid";
  open (MIDFILE, '<:encoding(UTF-8)', "MaritimeIdentificationDigits-ITU.csv")
    or return "N/A";

  while (not (eof MIDFILE) && $country eq "invalid")
    {
    $line = <MIDFILE>;
    chomp $line;
    @fields = split (",", $line);
    $num = $fields [0];
    $name = $fields [1];
    if ($num eq $mid)
      { $country = substr ($name,0,length($name)-1); last; }
    }
  close MIDFILE;
  return $country;
  }

# ********************************************************
# mmsi_handler
# ********************************************************

# Display the MMSI and provide some rudimentary information about it

# 0MIDXXXXX     Group of ships
# 00MIDXXXX     Coastal stations
# 111MIDXXX     SAR (Search and Rescue) aircraft
# MIDXXXXXX     Ship
# 8MIDXXXXX 	  Diver’s radio (not used in the U.S. in 2013)
# 970MIDXXX     AIS SART (Search and Rescue Transmitter)
# 972XXXXXX     MOB (Man Overboard) device
# 974XXXXXX     EPIRB (Emergency Position Indicating Radio Beacon) AIS
# 98MIDXXXX     Auxiliary craft associated with a parent ship
# 99MIDXXXX     Aids to Navigation

sub mmsi_handler
  {
  my ($mmsi) = @_;
  my ($i,$j,$ok,$mid,$prefix);

  $mmsi = sprintf ("%09d", $mmsi);
  printf "Maritime Mobile Service Identity (MMSI) = %s\n", $mmsi;
  $prefix = substr ($mmsi,0,1);
  $ok = 0; # Set $ok to 1 any sort of message is displayed

  PREFIX:
    {
    $prefix == 0 && do
      {
      if (substr ($mmsi,1,1) != "0")
          {
          $mid = substr ($mmsi,1,3);
          printf "  ** This MMSI (prefix: 0) represents a group of ships.\n";
          }
        else
          {
          $mid = substr ($mmsi,2,3);
          printf "  ** This MMSI (prefix: 00) represents a coastal station.\n";
          }
      printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
      printf "  ** Country: %s\n", mid_handler ($mid);

      $ok = 1;
      last PREFIX;
      };

    $prefix == 1 && do
      {
      if (substr ($mmsi,1,2) eq "11")
          {
          $mid = substr ($mmsi,3,3);
          printf "  ** This MMSI (prefix: 111) represents a Search and Rescue (SAR) aircraft.\n";
          printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
          printf "  ** Country: %s\n", mid_handler ($mid);
          $i = substr ($mmsi,6,1);
          if ($i == 1)
              { print "  ** (The aircraft is possibly fixed-wing.)\n"; }
            else
              {
              if ($i == 5)
                { print "  ** (The aircraft is possibly a helicopter.)\n"; }
              }
          }
        else
          { printf "  ** The MMSI prefix (1%s) is invalid\n", substr ($mmsi,1,2); }

      $ok = 1;
      last PREFIX;
      };

    $prefix >= 2 && $prefix <= 7 && do
      {
      $mid = substr ($mmsi,0,3);
      printf "  ** This MMSI represents a ship.\n";
      printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
      printf "  ** Country: %s\n", mid_handler ($mid);
      printf "  ** Try https://www.fleetmon.com/vessels/?s=%09d for more information\n", $mmsi;

      $ok = 1;
      last PREFIX;
      };

    $prefix == 8 && do
      {
      $mid = substr ($mmsi,1,3);
      printf "  ** This MMSI (prefix: 8) represents a Handheld VHF transceiver with DSC and GNSS.\n";
      printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
      printf "  ** Country: %s\n", mid_handler ($mid);

      $ok = 1;
      last PREFIX;
      };

    $prefix == 9 && do
      {
      $i = substr ($mmsi,1,1);
      FFLIST:
        {
        $i == 7 && do
          {
          $j = substr ($mmsi,2,1);
          if ($j == 0)
              {
              $mid = substr ($mmsi, 3,3);
              printf "  ** This MMSI (prefix: 970) represents a Search and Rescue Transmitter (SART).\n";
              printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
              printf "  ** Country: %s\n", mid_handler ($mid);
              $ok = 1;
              }
            else
              {
              if ($j == 2)
                  {
                  printf "  ** This MMSI (prefix: 972) represents a Man overboard (MOB) DSC and/or AIS device.\n";
                  $ok = 1;
                  }
                else
                  {
                  if ($j == 4)
                    {
                    printf "  ** This MMSI (prefix: 974) represents an Emergency Position Indicating Radio Beacon (EPIRB) device.\n";
                    $ok = 1;
                    }
                  }
              }
          last FFLIST;
          };
        $i == 8 && do
          {
          $mid = substr ($mmsi,2,3);
          printf "  ** This MMSI (prefix: 98) represents a craft associated with a parent ship.\n";
          printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
          printf "  ** Country: %s\n", mid_handler ($mid);
          $ok = 1;
          last FFLIST;
          };
        $i == 9 && do
          {
          $mid = substr ($mmsi,2,3);
          printf "  ** This MMSI (prefix: 99) represents an aid-to-navigation (AtoN).\n";
          printf "  ** The Maritime Identification Digits (MID) are %s\n", $mid;
          printf "  ** Country: %s\n", mid_handler ($mid);
          $ok = 1;
          last FFLIST;
          };
        };
      last PREFIX;
      };
    };

# If $ok is still 0, it means that the MMSI was totally not recognized and no other output has been generated

  if ($ok == 0)
    {  print "  ** This MMSI appears to be invalid.\n"; }

  return;
  }

# ********************************************************
# mmsi_367_35
# ********************************************************

# Special MMSI handler for Types 6/8, DAC 367, FI 35 messages, where there are special meaninsg to MMSIs <= 100001023

sub mmsi_367_35
  {
  my ($mmsi) = @_;

  printf "  Maritime Mobile Service Identity (MMSI) = %09d\n", $mmsi;
  
  MMSI_LIST:
    {
    $mmsi == 0000000000 && do
      {
      print "  ** Invalid MMSI, indicating no vessels in queue [default]\n";
      last MMSI_LIST;
      };
    $mmsi == 1000000000 && do
      {
      print "  ** Invalid MMSI, indicating additional lockage for MMSI previous in sequence\n";
      last MMSI_LIST;
      };
    $mmsi == 1000000001 && do
      {
      print "  ** Invalid MMSI, indicating a non-AIS vessel (e.g., pleasure boat)\n";
      last MMSI_LIST;
      };
    $mmsi == 1000000002 && do
      {
      print "  ** Invalid MMSI, indicating ice r debris\n";
      last MMSI_LIST;
      };
    $mmsi >= 1000000003 && $mmsi <= 1000001023 && do
      {
      print "  ** Unknown MMSI\n";
      last MMSI_LIST;
      };
    }

  return;
  }

# ********************************************************
# two_max_handler
# ********************************************************

# Handle the various default options for speed and height parameters, where they
# are generally in the range 0-x, x+1 means "x+1 or greater", and x+2 means not available.
# In some cases, values are in the range, say, 0-25.0, coded as 0-250 (factor of 10), where 251 means "25.1 or greater" and
# 252 means "ino not available." So, in the parser program, just use 25.1 and 25.2 for thsoe two maximum values.

sub two_max_handler
  {
  my ($label,$value,$units,$max1,$max2) = @_;
  printf $label, $value;
  if ($value == $max2)
      { print " (Information not available)"; }
    else
      {
      print $units;
      if (abs($value) == $max1)
        { print " or greater"; }
      }
  print "\n";
  return;
  }

# ********************************************************
# three_max_handler
# ********************************************************

# Handle the very few situations where there are three "maximum" values, not unlike two_max_handler. The situation here is that max1 means "or greater,"
# max2 means "data not available,"and max3 means "sensor not available". The default parameter indictes which if max2 and max3 are the default!

# In some cases, values are in the range, say, 0-25.0, coded as 0-250 (factor of 10), where 251 means "25.1 or greater" and
# 252 means "ino not available." So, in the parser program, just use 25.1 and 25.2 for thsoe two maximum values.

sub three_max_handler
  {
  my ($label,$value,$units,$max1,$max2,$max3,$default) = @_;

  printf $label, $value;
  
  WHAT_DO_I_DO_WITH_MAX:
    {
    $value == $max3 && do
      {
      print " (Sensor not available";
      if ($default == 3) { print " [default]"; }
      print ")";
      last WHAT_DO_I_DO_WITH_MAX;
      };
    $value == $max2 && do
      {
      print " (Information not available";
      if ($default == 2) { print " [default]"; }
      print ")";
      last WHAT_DO_I_DO_WITH_MAX;
      };
    do
      {
      print $units;
      if (abs($value) == $max1) { print " or greater"; }
      last WHAT_DO_I_DO_WITH_MAX;
      };
    };

  print "\n";
  return;
  }

# ********************************************************
# timestamp_handler
# ********************************************************

# The timestamp will be 0-59 seconds; values 60-63 have special meaning

sub timestamp_handler
  {
  my ($ts_local) = @_;

  my @special = (
    "Time stamp is not available [default]", #0 (60)
    "Positioning system is in manual input mode", #1 (61)
    "Electronic Position Fixing System operates in estimated (dead reckoning) mode", #2 (62)
    "Positioning system is inoperative", #3 (63)
    );

  printf "Timestamp = %d", $ts_local;
  if ($ts_local <= 59)
      { print " seconds\n"; }
    else
      { printf " (%s)\n", $special [$ts_local-60]; }

  return;
  }

# ********************************************************
# comm_state
# ********************************************************

# Common code for parsing communication status
# Ref: IALA Technical Clarifications on ITU Rec. ITU-R M.1371-1, Edition 1.4, December 2003,
# Tables 10, 11, and 12 (see also M.1371-5, Tables 18-20 and Table 46)

# Message Types 1-4 and 11 send a 19-bit state value; the message type determines the communication state.
#  Types 1, 2, 4, and 11 use SOTDMA so flag is set to 0 when this routine is called; Type 3 uses ITDMA and flag is set to 1
# Message Types 9, 18, and 26 prepend a 20th bit (flag) to indicate SOTDMA (0) or ITDMA (1)

sub comm_state # (msgtype,flag,state)
  {
  my ($type, $flag, $state) = @_;
  my $tyui;

  my @sync_state = (
    "UTC direct", #0
    "UTC indirect", #1
    "Station is synchronized to a base station (Base direct)", #2
    "Station is synchronized to another station", #3
    );

#  Format for Message Types 1, 2, 4, 11, and (Types 9, 18, or 26 with Flag = 0)
################################################
#  Bits     Len   Description
#   0-1       2   Sync state
#   2-4       3   Slot Time-Out
#   5-18     14   Submessage
#### Submessage depends upon Slot Time-out
# Slot Time-Out    Submessage
# 0                Slot offset
# 1                UTC hour (5 bits) and minute (7 bits)
# 2, 4, 6          Slot number
# 3, 5, 7          Number of other stations being received

  if ($flag eq "0")
      {
      print "--- SOTDMA Communication State ---\n";
      $tyui = oct ("0b".substr ($state,0,2));
      printf "Synchronization state = %d (%s)\n", $tyui, $sync_state [$tyui];
      $tyui = oct ("0b".substr ($state,2,3));
      printf "Slot timeout = %d slots remaining until slot change\n", $tyui;

      SUBMESSAGE:
        {
        $tyui == 0 && do
          {
          printf "Slot offset = %d\n", oct ("0b".substr ($state,5,14));
          last SUBMESSAGE;
          };
        $tyui == 1 && do
          {
          printf "UTC time (HH:MM) = %02d:%02d\n", oct ("0b".substr ($state,5,5)), oct ("0b".substr ($state,10,7));
          last SUBMESSAGE;
          };
        ($tyui == 2 || $tyui == 4 || $tyui == 6) && do
          {
          printf "Slot number used for this transmission = %d\n", oct ("0b".substr ($state,5,14));
          last SUBMESSAGE;
          };
        ($tyui == 3 || $tyui == 5 || $tyui == 7) && do
          {
          printf "Number of stations currently being received = %d\n", oct ("0b".substr ($state,5,14));
          last SUBMESSAGE;
          };
        };
      }
    else

#  Format for Message Type 3 and (Types 9, 18, & 26 with Flag = 1)
################################################
#  Bits     Len   Description
#   0-1       2   Sync state
#   2-14     13   Offset to next slot to be used
#  15-17      3   Number of consecutive slots to allocate
#  18         1   Keep flag

      {
      print "--- ITDMA Communication State ---\n";
      $tyui = oct ("0b".substr ($state,0,2));
      printf "Synchronization state = %d (%s)\n", $tyui, $sync_state [$tyui];
      printf "Offset to next slot to be used = %d\n", oct ("0b".substr ($state,2,13));
      printf "Number of consecutive slots to allocate = %d\n", oct ("0b".substr ($state,15,3)) + 1;
      printf "Slot remains allocated for next frame = %s (1 = TRUE, 0 = FALSE)\n", substr ($state,18,1);
      }

  return;
  }

# ********************************************************
# ship_type
# ********************************************************

# Convert numerical ship type value to a descriptive string

sub ship_type # (shiptype)
  {
# Parameter list
  my ($shiptype) = @_;
  my @descriptors = (
    "Not available [default]", #0
    "Reserved for future use", #1
    "Reserved for future use", #2
    "Reserved for future use", #3
    "Reserved for future use", #4
    "Reserved for future use", #5
    "Reserved for future use", #6
    "Reserved for future use", #7
    "Reserved for future use", #8
    "Reserved for future use", #9
    "Reserved for future use", #10
    "Reserved for future use", #11
    "Reserved for future use", #12
    "Reserved for future use", #13
    "Reserved for future use", #14
    "Reserved for future use", #15
    "Reserved for future use", #16
    "Reserved for future use", #17
    "Reserved for future use", #18
    "Reserved for future use", #19
    "Wing in ground (WIG), all ships of this type", #20
    "Wing in ground (WIG), Hazardous category A", #21
    "Wing in ground (WIG), Hazardous category B", #22
    "Wing in ground (WIG), Hazardous category C", #23
    "Wing in ground (WIG), Hazardous category D", #24
    "Wing in ground (WIG), Reserved for future use", #25
    "Wing in ground (WIG), Reserved for future use", #26
    "Wing in ground (WIG), Reserved for future use", #27
    "Wing in ground (WIG), Reserved for future use", #28
    "Wing in ground (WIG), Reserved for future use", #29
    "Fishing", #30
    "Towing ahead or alongside", #31
    "Towing astern, length exceeds 200m or breadth exceeds 25m", #32
    "Engaged in dredging or underwater operations", #33
    "Engaged in diving operationss", #34
    "Engaged in military operationss", #35
    "Sailing vessels", #36
    "Pleasure Craft", #37
    "Reserved", #38
    "Reserved", #39
    "High speed craft (HSC), all ships of this type", #40
    "High speed craft (HSC), Hazardous category A", #41
    "High speed craft (HSC), Hazardous category B", #42
    "High speed craft (HSC), Hazardous category C", #43
    "High speed craft (HSC), Hazardous category D", #44
    "High speed craft (HSC), Reserved for future use", #45
    "High speed craft (HSC), Reserved for future use", #46
    "High speed craft (HSC), Reserved for future use", #47
    "High speed craft (HSC), Reserved for future use", #48
    "High speed craft (HSC), No additional information", #49
    "Pilot Vessel", #50
    "Search and Rescue vessel", #51
    "Tug", #52
    "Port Tender", #53
    "Vessel with anti-pollution facilities or equipment", #54
    "Law Enforcement", #55
    "(spare)", #56
    "Spare - Local Vessel", #57
    "Medical Transport", #58
    "Noncombatant ship according to RR Resolution No. 18", #59
    "Passenger, all ships of this type", #60
    "Passenger, Hazardous category A", #61
    "Passenger, Hazardous category B", #62
    "Passenger, Hazardous category C", #63
    "Passenger, Hazardous category D", #64
    "Passenger, Reserved for future use", #65
    "Passenger, Reserved for future use", #66
    "Passenger, Reserved for future use", #67
    "Passenger, Reserved for future use", #68
    "Passenger, No additional information", #69
    "Cargo, all ships of this type", #70
    "Cargo, Hazardous category A", #71
    "Cargo, Hazardous category B", #72
    "Cargo, Hazardous category C", #73
    "Cargo, Hazardous category D", #74
    "Cargo, Reserved for future use", #75
    "Cargo, Reserved for future use", #76
    "Cargo, Reserved for future use", #77
    "Cargo, Reserved for future use", #78
    "Cargo, No additional information", #79
    "Tanker, all ships of this type", #80
    "Tanker, Hazardous category A", #81
    "Tanker, Hazardous category B", #82
    "Tanker, Hazardous category C", #83
    "Tanker, Hazardous category D", #84
    "Tanker, Reserved for future use", #85
    "Tanker, Reserved for future use", #86
    "Tanker, Reserved for future use", #87
    "Tanker, Reserved for future use", #88
    "Tanker, No additional information", #89
    "Other Type, all ships of this type", #90
    "Other Type, Hazardous category A", #91
    "Other Type, Hazardous category B", #92
    "Other Type, Hazardous category C", #93
    "Other Type, Hazardous category D", #94
    "Other Type, Reserved for future use", #95
    "Other Type, Reserved for future use", #96
    "Other Type, Reserved for future use", #97
    "Other Type, Reserved for future use", #98
    "Other Type, no additional information", #99
    );

  if ($shiptype < 0 || $shiptype > 99)
    { $shiptype = 99; }

  return $descriptors [$shiptype];
  }

sub setparms
{

# This routine includes tables of AIS parameters. If the table is completely full -- e.g., 16 choices for a four-bit variable,
# the array is here. If there are "gaps" -- e.g., 20 choices for a five-bit variable -- refer to the subroutines above!

# Note that these are all global variables.

# ********************************************************
# Accuracy

@accuracy = (
  "Unaugmented GNSS fix, >10 m [default]", #0
  "DGPS-quality fix, <10 m", #1
  );

# ********************************************************
# Acknowledge required flag

@ack_req = (
  "Reply not required", #0
  "Reply required", #1
  );

# ********************************************************
# Additional vessel type information

@add_vessel_type_info = (
  "Unknown or not in list [default]", #0
  "Ferry vessel (people and vehicles)", #1
	"Ferry vessel (people only)", #2
  "Cruise ship", #3
  "Excursion vessel", #4
  "Reserved for future use", #5
  "Reserved for future use", #6
  "Reserved for future use", #7
  "Reserved for future use", #8
  "Reserved for future use", #9
  "Reserved for future use", #10
  "Reserved for future use", #11
  "Reserved for future use", #12
  "Reserved for future use", #13
  "Reserved for future use", #14
  "Reserved for future use", #15
  );

# ********************************************************
# Addressed mode

@addressed = (
  "Broadcast", #0
  "Addressed", #1
  );

# ********************************************************
# AIS version indicator

@ais_version = (
  "Station compliant with Recommendation ITU-R M.1371-1", #0
  "Station compliant with Recommendation ITU-R M.1371-3 (or later)", #1
  "Station compliant with Recommendation ITU-R M.1371-5 (or later)", #2
  "Station compliant with future editions", #3
  );

# ********************************************************
# Area shape

@area_shape = (
  "Circle/Point", #0
  "Rectangle", #1
  "Sector", #2
  "Polyline", #3
  "Polygon", #4
  "Associated Text", #5
  "(reserved)", #6
  "(reserved)", #7
  );

# ********************************************************
# Assigned-mode flag

@assigned = (
  "Autonomous mode [default]", #0
  "Assigned mode", #1
  );

# ********************************************************
# AtoN Position flag

@aton_pos = (
  "AtoN is on position", #0
  "AtoN is off position", #1
  );

# ********************************************************
# AtoN light status

@aton_status_light = (
  "No light or no monitoring", #0
  "Light on", #1
  "Light off", #2
  "Light error", #3
  );

# ********************************************************
# AtoN unit status

@aton_status_unit = (
  "Good health", #0
  "Alarm", #1
  );

# ********************************************************
# AtoN RACON status

@aton_status_racon = (
  "No RACON installed", #0
  "RACON not monitored", #1
  "RACON operational", #2
  "RACON error", #3
  );

# ********************************************************
# AtoN Types

@aton_type = (
  "Type not specified [default]", #0
  "Reference point", #1
  "RACON (radar transponder marking a navigation hazard)", #2
  "Fixed structure off shore", #3
  "Spare, reserved for future use", #4
  "Light, without sectors", #5
  "Light, with sectors", #6
  "Leading Light Front", #7
  "Leading Light Rear", #8
  "Beacon, Cardinal N", #9
  "Beacon, Cardinal E", #10
  "Beacon, Cardinal S", #11
  "Beacon, Cardinal W", #12
  "Beacon, Port hand", #13
  "Beacon, Starboard hand", #14
  "Beacon, Preferred Channel port hand", #15
  "Beacon, Preferred Channel starboard hand", #16
  "Beacon, Isolated danger", #17
  "Beacon, Safe water", #18
  "Beacon, Special mark", #19
  "Cardinal Mark N", #20
  "Cardinal Mark E", #21
  "Cardinal Mark S", #22
  "Cardinal Mark W", #23
  "Port hand Mark", #24
  "Starboard hand Mark", #25
  "Preferred Channel Port hand", #26
  "Preferred Channel Starboard hand", #27
  "Isolated danger", #28
  "Safe Water", #29
  "Special Mark", #30
  "Light Vessel / LANBY / Rigs", #31
  );

# ********************************************************
# Band flag

@band = (
  "Unit may not be able to switch frequency", #0
  "Unit can use any part of the marine channel", #1
  );

# ********************************************************
# Battery status

@battery_status = (
  "Good", #0
  "Low voltage", #1
  );
 
# ********************************************************
# Beaufort scale

@Beaufort = (
  "Calm: <1 kn (Sea like a mirror)", #0
  "Light air: 1-3 kn (Ripples without crests)", #1
  "Light breeze: 4-6 kn (Small wavelets; crests not breaking)", #2
  "Gentle breeze: 7-10 kn (Large wavelets; crests begin to break)", #3
  "Moderate breeze: 11-16 kn (Small waves)", #4
  "Fresh breeze: 17-21 kn (Moderate waves)", #5
  "Strong breeze: 22-27 kn (Large waves with foam crests, some spray)", #6
  "High wind: 28-33 kn (Sea heaps up; foam begins to streak)", #7
  "Gale: 34-40 kn (Moderately high waves with breaking crests)", #8
  "Strong gale: 41-47 kn (High waves)", #9
  "Storm: 48-55 kn (Very high waves; reduced visibility)", #10
  "Violent storm: 56-63 kn (Exceptionally high waves)", #11
  "Hurricane force: >63 kn (Huge waves)", #12
  "Information not available", #13
  "not used", #14
  "not used", #15
  );

# ********************************************************
# Channel A/B Band

@channel_band = (
  "Default", #0
  "12.5 kHz", #1
  );

# ********************************************************
# Carrier sense (CS) unit flag

@CS = (
  "Class B SOTDMA unit", #0
  "Class B CS unit", #1
  );

# ********************************************************
# Direction of the ship

@direction = (
   "Up or in", #0
   "Down or out", #1
   );

# ********************************************************
# Source of information for dam and hydroelectric plant discharge

@discharge_source = (
  "measured", #0
  "calculated", #1
  "observed", #2
  "unknown (default)", #3
  );

# ********************************************************
# Display flag

@display = (
  "No visual display", #0
  "Has visual display", #1
  );

# ********************************************************
# Digital selective calling (DSC) flag

@DSC = (
  "No DCS available", #0
  "Unit attached to VHF radio with DSC", #1
  );

# ********************************************************
# Data Terminal Equipment (DTE) Status

@dte = (
   "DTE ready", #0
   "DTE not ready [default]", #1
   );

# ********************************************************
# Environmental Management message parameters
  
 @em_data_timeout = (
  "No timeout period [default]", #0
  "10 minutes", #1
  "1 hour", #2
  "6 hours", #3
  "12 hours", #4
  "24 hours", #5
  "5 minutes", #6
  "(reserved)", #7
  );

@em_salinity_type = (
  "Measured", #0
  "Calculated using PSS-78", #1
  "Calculated using some other method", #2
  "Reserved", #3
  );

@em_sensor_data_desc = (
  "No data [default]", #0
  "Raw real time", #1
  "Real time with Quality Control", #2
  "Predicted [based on historical statistics]", #3
  "Forecast [predicted, refined with real-time information]", #4
  "Nowcast [a continuous forecast]", #5
	"Reserved", #6
  "Sensor not available", #7
  );

@em_sensor_owner = (
  "Unknown [default]", #0
  "U.S. NOAA National Ocean Service (hydrographic office)", #1
  "U.S. Army Corps of Engineers (inland waterway authority)", #2
  "Coastal directorate", #3
  "NOAA National Weather Service (meteorological service)", #4
  "State or Port Authority", #5
  "U.S. Coast Guard", #6
  "U.S. Navy", #7
  "Energy or Environmental Agency", #8
  "Transportation Agency", #9
  "Academic or Research Institution", #10
  "Private or Commercial Service", #11
  "Marine Exchange or Port Service", #12
  "US Geologic Survey", #13
  "Other", #14
  "(not to be used)", #15
  );

@em_sensor_report_type = (
  "Sensor Site Location", #0
  "Station ID", #1
  "Wind Report", #2
  "Tide/River Stage (Water Level) Report", #3
  "Two-Dimensional (x & y) Vertical Current Profile Report ", #4
  "Three-Dimensional (x, y, & z) Vertical Current Profile Report", #5
  "Horizontal Current Profile", #6
  "Sea State", #7
  "Salinity", #8
  "Weather", #9
  "Air Gap", #10
  "Wind Report v2", #11
  "(reserved for future use)", #12
  "(reserved for future use)", #13
  "(reserved for future use)", #14
  "(reserved for future use)", #15
  );

@em_vert_ref_datum = (
  "Mean Lower Low Water (MLLW)", #0
  "International Great Lakes Datum (IGLD-85)", #1
  "Local river datum", #2
  "Station Datum (STND)", #3
  "Mean Higher High Water (MHHW)", #4
  "Mean High Water (MHW)", #5
  "Mean Sea Level (MSL)", #6
  "Mean Low Water (MLW)", #7
  "National Geodetic Vertical Datum (NGVD-29)", #8
  "North American Vertical Datum (NAVD-88)", #9
  "World Geodetic System (WGS-84)", #10
  "Lowest Astronomical Tide (LAT)", #11
  "Pool", #12
  "Gauge", #13
  "Unknown/unavailable [default]", #14
  "(reserved)", #15
  "(reserved)", #16
  "(reserved)", #17
  "(reserved)", #18
  "(reserved)", #19
  "(reserved)", #20
  "(reserved)", #21
  "(reserved)", #22
  "(reserved)", #23
  "(reserved)", #24
  "(reserved)", #25
  "(reserved)", #26
  "(reserved)", #27
  "(reserved)", #28
  "(reserved)", #29
  "(reserved)", #30
  "(not to be used)", #31
  );

@em_water_level_trend = (
  "increasing", #0
  "decreasing", #1
  "steady", #2
  "unknown/data unavailable [default]", #3
  );

@em_water_level_type = (
  "relative to reference datum", #0
  "water depth", #1
  );

@em_wmo = (
  "Rain", #0
  "Snow", #1
  "Rain and snow)", #2
  "None [default]", #3
  );
  
# ********************************************************
# END OF Environmental Management MESSAGE PARAMETERS
  
# ********************************************************
# EPFD Type

@epfd_type = (
  "undefined [default]", #0
  "GPS", #1
  "GLONASS", #2
  "Combined GPS/GLONASS", #3
  "Loran-C", #4
  "Chayka", #5
  "Integrated navigation system", #6
  "Surveyed", #7
  "Galileo", #8
  "not used", #9
  "not used", #10
  "not used", #11
  "not used", #12
  "not used", #13
  "not used", #14
  "not used", #15
  );

# ********************************************************
# GNSS Position status flag

@gnss_position_status = (
  "Current GNSS position", #0
  "Not GNSS position [default]", #1
  );

# ********************************************************
# GPS event status

@gps_event_status = (
  "start of event", #0
  "end of event", #1
  );

# ********************************************************
# GPS event type

@gps_event_type = (
  "jamming", #0
  "spoofing", #1
  );

# ********************************************************
# Ice Type

@ice_type = (
  "None", #0
  "Present", #1
  "reserved", #2
  "N/A [default]", #3
  );

# ********************************************************
# Reporting Interval

@interval = (
  "As given by the autonomous mode", #0
  "10 minutes", #1
  "6 minutes", #2
  "3 minutes", #3
  "1 minute", #4
  "30 seconds", #5
  "15 seconds", #6
  "10 seconds", #7
  "5 seconds", #8
  "Next Shorter Reporting Interval", #9
  "Next Longer Reporting Interval", #10
  "Reserved for future use", #11
  "Reserved for future use", #12
  "Reserved for future use", #13
  "Reserved for future use", #14
  "Reserved for future use", #15
  );

# ********************************************************
# Light status

@light_status = (
  "Light off", #0
  "Light on", #1
  );
 
# ********************************************************
# Lockage type

@lockage_type = (
  "undefined or not applicable [default]", #0
  "open pass [O]", #1
  "fast double [F]", #2
  "jackknife [J]", #3
  "knockout [K]", #4
	"navigable pass [N]", #5
  "straight [S]", #6
	"barge transfer [T]", #7
  "setover [V]", #8
  "other [Z]", #9
  "undefined", #10
  "undefined", #11
  "undefined", #12
  "undefined", #13
  "undefined", #14
  "undefined", #15
  );
 
# ********************************************************
# Maneuver

@maneuver = (
  "Not available [default]", #0
  "No special maneuver", #1
  "Special maneuver (such as regional passing arrangement)", #2
  "invalid", #3
  );

# ********************************************************
# Message Type

@message_type = (
  "invalid message type", #0
  "Position Report Class A", #1
  "Position Report Class A (Assigned Schedule)", #2
  "Position Report Class A (Response to Interrogation)", #3
  "Base Station Report", #4
  "Static and Voyage Related Data", #5
  "Binary Addressed Message", #6
  "Binary Acknowledge", #7
  "Binary Broadcast Message", #8
  "Standard SAR Aircraft Position Report", #9
  "UTC/Date Inquiry", #10
  "UTC/Date Response", #11
  "Addressed Safety-Related Message", #12
  "Safety-Related Acknowledgement", #13
  "Safety-Related Broadcast Message", #14
  "Interrogation", #15
  "Assignment Mode Command", #16
  "DGNSS Broadcast Binary Message", #17
  "Standard Class B CS Position Report", #18
  "Extended Class B CS Position Report", #19
  "Data Link Management Message", #20
  "Aid-to-Navigation Report", #21
  "Channel Management", #22
  "Group Assignment Command", #23
  "Static Data Report", #24
  "Single Slot Binary Message", #25
  "Multiple Slot Binary Message", #26
  "Long Range AIS Broadcast Message", #27
  "invalid message type", #28
  "invalid message type", #29
  "invalid message type", #30
  "invalid message type", #31
  );

# ********************************************************
# Mooring Position

@mooring_position = (
  "Undefined [default]", #0
  "Port-side to", #1
  "Starboard-side to", #2
  "Mediterranean mooring", #3
  "Mooring buoy", #4
  "Anchorage", #5
  "reserved for future use", #6
  "reserved for future use", #7
  );

# ********************************************************
# Message 22 flag

@msg22 = (
  "Unit cannot accept a channel assignment via Message Type 22", #0
  "Unit can accept a channel assignment via Message Type 22", #1
  );

# ********************************************************
# Navigation Status

@nav_stat = (
  "Under way using engine", #0
  "At anchor", #1
  "Not under command", #2
  "Restricted manoeuverability", #3
  "Constrained by her draught", #4
  "Moored", #5
  "Aground", #6
  "Engaged in Fishing", #7
  "Under way sailing", #8
  "Reserved for future amendment of Navigational Status for HSC", #9
  "Reserved for future amendment of Navigational Status for WIG", #10
  "Reserved for future use", #11
  "Reserved for future use", #12
  "Reserved for future use", #13
  "AIS-SART is active", #14
  "Not defined [default]", #15
  );

# ********************************************************
# Notice Description

@notice_desc = (
  "Caution Area: Marine mammals habitat", #0
  "Caution Area: Marine mammals in area - reduce speed", #1
  "Caution Area: Marine mammals in area - stay clear", #2
  "Caution Area: Marine mammals in area - report sightings", #3
  "Caution Area: Protected habitat - reduce speed", #4
  "Caution Area: Protected habitat - stay clear", #5
  "Caution Area: Protected habitat - no fishing or anchoring", #6
  "Caution Area: Derelicts (drifting objects)", #7
  "Caution Area: Traffic congestion", #8
  "Caution Area: Marine event", #9
  "Caution Area: Divers down", #10
  "Caution Area: Swim area", #11
  "Caution Area: Dredge operations", #12
  "Caution Area: Survey operations", #13
  "Caution Area: Underwater operation", #14
  "Caution Area: Seaplane operations", #15
  "Caution Area: Fishery - nets in water", #16
  "Caution Area: Cluster of fishing vessels", #17
  "Caution Area: Fairway closed", #18
  "Caution Area: Harbour closed", #19
  "Caution Area: Risk (define in Associated Text field)", #20
  "Caution Area: Underwater vehicle operation", #21
  "(reserved for future use)", #22
  "Environmental Caution Area: Storm front (line squall)", #23
  "Environmental Caution Area: Hazardous sea ice", #24
  "Environmental Caution Area: Storm warning (storm cell or line of storms)", #25
  "Environmental Caution Area: High wind", #26
  "Environmental Caution Area: High waves", #27
  "Environmental Caution Area: Restricted visibility (fog, rain, etc.)", #28
  "Environmental Caution Area: Strong currents", #29
  "Environmental Caution Area: Heavy icing", #30
  "(reserved for future use)", #31
  "Restricted Area: Fishing prohibited", #32
  "Restricted Area: No anchoring.", #33
  "Restricted Area: Entry approval required prior to transit", #34
  "Restricted Area: Entry prohibited", #35
  "Restricted Area: Active military OPAREA", #36
  "Restricted Area: Firing – danger area", #37
  "Restricted Area: Drifting Mines", #38
  "(reserved for future use)", #39
  "Anchorage Area: Anchorage open", #40
  "Anchorage Area: Anchorage closed", #41
  "Anchorage Area: Anchoring prohibited", #42
  "Anchorage Area: Deep draft anchorage", #43
  "Anchorage Area: Shallow draft anchorage", #44
  "Anchorage Area: Vessel transfer operations", #45
  "Anchorage Area: Other (define in Associated Text field)", #46
  "Ice Report: Ice Edge", #47
  "Ice Report: New Ice (<10cm ocean, <5cm lake)", #48
  "Ice Report: Young Ice (10-30cm)", #49
  "Ice Report: Thin 1st Year Ice (30-70cm ocean, 5-15cm lake)", #50
  "Ice Report: Medium 1st Year Ice (70-120cm ocean, 15-30cm lake)", #51
  "Ice Report: Thick 1st Year Ice (120-200cm ocean, 30-70cm lake)", #52
  "Ice Report: Old/Very Thick Year Ice (>200cm ocean, >70cm lake)", #53
  "Ice Report: Undetermined or unknown thickness", #54
  "(reserved for future use)", #55
  "Security Alert - Level 1", #56
  "Security Alert - Level 2", #57
  "Security Alert - Level 3", #58
  "(reserved for future use)", #59
  "(reserved for future use)", #60
  "(reserved for future use)", #61
  "(reserved for future use)", #62
  "(reserved for future use)", #63
  "Distress Area: Vessel disabled and adrift", #64
  "Distress Area: Vessel sinking", #65
  "Distress Area: Vessel abandoning ship", #66
  "Distress Area: Vessel requests medical assistance", #67
  "Distress Area: Vessel flooding", #68
  "Distress Area: Vessel fire/explosion", #69
  "Distress Area: Vessel grounding", #70
  "Distress Area: Vessel collision", #71
  "Distress Area: Vessel listing/capsizing", #72
  "Distress Area: Vessel under assault", #73
  "Distress Area: Person overboard", #74
  "Distress Area: SAR area", #75
  "Distress Area: Pollution response area", #76
  "(reserved for future use)", #77
  "(reserved for future use)", #78
  "(reserved for future use)", #79
  "Instruction: Contact VTS at this point/juncture", #80
  "Instruction: Contact Port Administration at this point/juncture", #81
  "Instruction: Do not proceed beyond this point/juncture", #82
  "Instruction: Await instructions prior to proceeding beyond this point/juncture", #83
  "Proceed to this location - await instructions", #84
  "Clearance granted - proceed to berth", #85
  "(reserved for future use)", #86
  "(reserved for future use)", #87
  "Information: Pilot boarding position", #88
  "Information: Icebreaker waiting area", #89
  "Information: Places of refuge", #90
  "Information: Position of icebreakers", #91
  "Information: Location of response units", #92
  "VTS active target", #93
  "Rouge or suspicious vessel", #94
  "Vessel requesting non-distress assistance", #95
  "Chart Feature: Sunken vessel", #96
  "Chart Feature: Submerged object", #97
  "Chart Feature: Semi-submerged object", #98
  "Chart Feature: Shoal area", #99
  "Chart Feature: Shoal area due north", #100
  "Chart Feature: Shoal area due east", #101
  "Chart Feature: Shoal area due south", #102
  "Chart Feature: Shoal area due west", #103
  "Chart Feature: Channel obstruction", #104
  "Chart Feature: Reduced vertical clearance", #105
  "Chart Feature: Bridge closed", #106
  "Chart Feature: Bridge partially open", #107
  "Chart Feature: Bridge fully open", #108
  "(reserved for future use)", #109
  "(reserved for future use)", #110
  "(reserved for future use)", #111
  "Report from ship: Icing info", #112
  "(reserved for future use)", #113
  "Report from ship: Miscellaneous information (see Associated Text field)", #114
  "(reserved for future use)", #115
  "(reserved for future use)", #116
  "(reserved for future use)", #117
  "(reserved for future use)", #118
  "(reserved for future use)", #119
  "Route: Recommended route", #120
  "Route: Alternative route", #121
  "Route: Recommended route through ice", #122
  "(reserved for future use)", #123
  "(reserved for future use)", #124
  "Other - Define in Associated Text field", #125
  "Cancellation - cancel area as identified by Message Linkage ID", #126
  "Undefined [default]", #127
  );

# ********************************************************
# Off-position Indicator

@offpos = (
  "On position [default]", #0
  "Off position", #1
  );

# ********************************************************
# Precipitation Type

@precip_type = (
  "reserved", #0
  "Rain", #1
  "Thunderstorm", #2
  "Freezing rain", #3
  "Mixed/ice", #4
  "Snow", #5
  "reserved", #6
  "N/A [default]", #7
  );

# ********************************************************
# Power level

@power = (
  "Low", #0
  "High", #1
  );

# ********************************************************
# Power supply type

@power_supply = (
  "Alternating current (A)", #0
  "Direct current (DC)", #1
  );
  
# ********************************************************
# Receiver Autonomous Integrity Monitoring (RAIM) flag

@raim = (
  "RAIM not in use [default]", #0
  "RAIM in use", #1
  );

# ********************************************************
# Retransmit flag

@retransmit = (
  "No retransmit [default]", #0
  "Retransmitted", #1
  );

# ********************************************************
# Route Type

@routetype = (
  "Not available [default]", #0
  "Mandatory route", #1
  "Recommended route", #2
  "Alternative route", #3
  "Recommended route through ice", #4
  "Ship route plan", #5
  "(Reserved for future use)", #6
  "(Reserved for future use)", #7
  "(Reserved for future use)", #8
  "(Reserved for future use)", #9
  "(Reserved for future use)", #10
  "(Reserved for future use)", #11
  "(Reserved for future use)", #12
  "(Reserved for future use)", #13
  "(Reserved for future use)", #14
  "(Reserved for future use)", #15
  "(Reserved for future use)", #16
  "(Reserved for future use)", #17
  "(Reserved for future use)", #18
  "(Reserved for future use)", #19
  "(Reserved for future use)", #20
  "(Reserved for future use)", #21
  "(Reserved for future use)", #22
  "(Reserved for future use)", #23
  "(Reserved for future use)", #24
  "(Reserved for future use)", #25
  "(Reserved for future use)", #26
  "(Reserved for future use)", #27
  "(Reserved for future use)", #28
  "(Reserved for future use)", #29
  "(Reserved for future use)", #30
  "Cancellation (cancel route as identified by Message Linkage ID)", #31
  );

# ********************************************************
# Sender Classification

@senderclass = (
  "Ship [default]", #0
  "Authority", #1
  "(Reserved for future use)", #2
  "(Reserved for future use)", #3
  "(Reserved for future use)", #4
  "(Reserved for future use)", #5
  "(Reserved for future use)", #6
  "(Reserved for future use)", #7
  );

# ********************************************************
# Service Availability AND Service List
# Each Type of Service List entry takes on one of the Service Availability values

@service_avail = (
  "Service not available or requested [default]", #0
  "Service available", #1
  "No data or unknown", #2
  "(not to be used)", #3
  );

@service_list = (
  "", # List in standard starts at 1
  "Agent", #1
  "Bunker/fuel", #2
  "Chandler", #3
  "Stevedore", #4
  "Electrical", #5
  "Potable water", #6
  "Customs house", #7
  "Cartage", #8
  "Crane(s)", #9
  "Lift(s)", #10
  "Medical facilities", #11
  "Navigation repair", #12
  "Provisions", #13
  "Ship repair", #14
  "Surveyor", #15
  "Steam", #16
  "Tugs", #17
  "Waste disposal (solid)", #18
  "Waste disposal (liquid)", #19
  "Waste disposal (hazardous)", #20
  "Reserved ballast exchange", #21
  "Additional services are also available", #22
  "(Reserved for regional use)", #23
  "(Reserved for regional use)", #24
  "(Reserved for future use)", #25
  "(Reserved for future use)", #26
  );

# ********************************************************
# Slot Increments

@slot_inc = (
  "", #0
  "1,125 slots", #1
  "375 slots", #2
  "225 slots", #3
  "125 slots", #4
  "75 slots", #5
  "45 slots", #6
  "undefined -- disregard this assignment", #7
  );

# ********************************************************
# Station Type

@station_type = (
  "All types of mobiles [default]", #0
  "Reserved for future use", #1
  "All types of Class B mobile stations", #2
  "SAR airborne mobile station", #3
  "Aid to Navigation station", #4
  "Class B shipborne mobile station (IEC62287 only)", #5
  "Regional use and inland waterways", #6
  "Regional use and inland waterways", #7
  "Regional use and inland waterways", #8
  "Regional use and inland waterways", #9
  "Reserved for future use", #10
  "Reserved for future use", #11
  "Reserved for future use", #12
  "Reserved for future use", #13
  "Reserved for future use", #14
  "Reserved for future use", #15
  );

# ********************************************************
# Weather Tendency

@tendency = (
  "Steady", #0
  "Decreasing", #1
  "Increasing", #2
  "N/A [default]", #3
  );

# ********************************************************
# Tx/Rx Mode

@txrx_mode = (
  "TxA/TxB, RxA/RxB [default]", #0
  "TxA, RxA/RxB", #1
  "TxB, RxA/RxB", #2
  "reserved for future use", #3
  );

# ********************************************************
# Virtual AtoN flag

@vaton = (
  "Real AtoN at indicated position [default]", #0
  "Virtual AtoN simulated by nearby AIS station", #1
  );

# ********************************************************
# Waterways management vessel name descriptor type

@wm_descriptor = (
  "Name [default]", #0
  "IMO number", #1
  "State number", #2
  "reserved for future use", #3
  "reserved for future use", #4
  "reserved for future use", #5
  "reserved for future use", #6
  "Other", #7
  );

# ********************************************************
# Waterways management message subtype

@wm_subtype = (
  "Sequence - Vessel list", #0
  "Name/Position - WM place identification", #1
  "Vessel name", #2
  "reserved for future use", #3
  );

# ********************************************************
# Vessel status (lock)

@vessel_status = (
  "In the lock", #0
  "Vessel arrived at check-in point", #1
  "reserved for future use", #2
  "reserved for future use", #3
  );

# ********************************************************
# VTS id. type

@vts_id_type = (
  "MMSI", #0
  "IMO number", #1
  "Call sign", #2
  "Other (should be 0)", #3
  );

# ********************************************************
# Waterways management message type

@wm_type = (
  "Lock", #0
  "Gate", #1
  "Narrows", #2
  "Bridge", #3
  "Restricted channel - one vessel at a time - could be alternating directions - no passing or overtaking", #4
  "Estimated Arrival Time", #5
  "Assigned Arrival Time", #6
  "Traffic Advisory", #7
  "Cleared to Enter/Proceed", #8
  "Not Cleared to Enter/Do Not Proceed", #9
  "Proceed to Berth", #10
  "Proceed to _____ (defined in Linked Text message", #11
  "Vessels Awaiting Lockage", #12
  "Ice Convoy (specify speed and spacing in Linked Text message", #13
  "undefined", #14
  "undefined", #15
  );
}

# ********************************************************
# decode -- AIS MESSAGE DECODING ROUTINES
# ********************************************************
# The decode_X routines decode Type X messages, where X is the type value.

# *** Decode_010203 decodes Types 1-3, Position Report Class A messages ***

sub decode_010203
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my ($tyui,$tyui2);

################################################
#  Bits     Len   Description
#  38-41      4   Navigation Status
#  42-49      8   Rate of Turn (ROT)
#  50-59     10   Speed Over Ground (SOG)
#  60         1   Position Accuracy
#  61-88     28   Longitude
#  89-115    27   Latitude
# 116-127    12   Course Over Ground (COG)
# 128-136     9   True Heading (HDG)
# 137-142     6   Time Stamp
# 143-144     2   Maneuver Indicator
# 145-147     3   Spare
# 148         1   RAIM flag
# 149-167    19   Radio Status

  $tyui = oct ("0b".substr ($message,38,4));
  printf "Navigation Status = %d (%s)\n", $tyui, $nav_stat [$tyui];

# Rate-of-Turn conversion is to take (ROTAIS/4.733)**2, preserving the sign...
# ROT values 0, +/-127, and 128 have special meaning...

  $tyui = signed_bin2dec (substr ($message,42,8));
  print "Rate of Turn (ROTAIS) = $tyui";

  ROT_DECODE:
    {
    $tyui == 0 && do
      {
      print " (not turning)\n";
      last ROT_DECODE;
      };
    $tyui == -127 && do
      {
      print " (turning left >5\xB0/30s)\n"; 
      last ROT_DECODE;
      };
    $tyui == 127 && do
      {
      print " (turning right >5\xB0/30s)\n"; 
      last ROT_DECODE;
      };
    ($tyui >= 128 || $tyui <= -127) && do
      {
      print " (no turn information available [default])\n"; 
      last ROT_DECODE;
      };

# If we get here, we have an actual rate-of-turn value!!!

  do
    {
    $tyui2 = ($tyui / 4.733) ** 2;
    if ($tyui < 0) { $tyui2 = -$tyui2; }
    printf " (%+.1f\xB0/min)\n", $tyui2;    #exit;
    };
  }

  two_max_handler ("Speed over ground (SOG) = %.1f", (oct ("0b".substr ($message,50,10)) / 10)," kn",102.2,102.3);
  $tyui = substr ($message,60,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,61,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,89,27)) / 600000),91);
  default_handler ("Course over ground (COG) = %05.1f", oct ("0b".substr ($message,116,12)) / 10, "\xB0", 360);
  default_handler ("True heading = %03d", oct ("0b".substr ($message,128,9)), "\xB0", 511);
  timestamp_handler (oct ("0b".substr ($message,137,6)));
  $tyui = oct ("0b".substr ($message,143,2)); printf "Maneuver indicator = %d (%s)\n", $tyui, $maneuver [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,145,3));
  $tyui = substr ($message,148,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];
  if ($type == 1 || $type == 2)
      { $tyui = "0"; }
    else
      { $tyui = "1"; }
  comm_state ($type, $tyui, substr ($message,149,19));

  return;
  }

# *** Decode_0411 decodes Type 4, Base Station Report and Type 11, UTC/Date Response messages ***

sub decode_0411
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-51     14   UTC year
#  52-55      4   UTC month
#  56-60      5   UTC day
#  61-65      5   UTC hour
#  66-71      6   UTC minute
#  72-77      6   UTC second
#  78         1   Position Accuracy
#  79-106    28   Longitude
# 107-133    27   Latitude
# 134-137     4   Type of EPFD
# 138-147    10   Spare
# 148         1   RAIM flag
# 149-167    19   Radio Status

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  default_handler ("UTC year = %d",oct ("0b".substr ($message,38,14)),"",0);
  default_handler ("UTC month = %d",oct ("0b".substr ($message,52,4)),"",0);
  default_handler ("UTC day = %d",oct ("0b".substr ($message,56,5)),"",0);
  default_handler ("UTC hour = %02d",oct ("0b".substr ($message,61,5)),"",24);
  default_handler ("UTC minute = %02d",oct ("0b".substr ($message,66,6)),"",60);
  default_handler ("UTC second = %02d",oct ("0b".substr ($message,72,6)),"",60);
  $tyui = substr ($message,78,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,79,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,107,27)) / 600000),91);
  $tyui = oct ("0b".substr ($message,134,4)); printf "EPFD type = %d (%s)\n", $tyui, $epfd_type [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,138,10));
  $tyui = substr ($message,148,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];

  comm_state ($type, "0", substr ($message,149,19));

  return;
  }

# *** Decode_05 decodes Type 5, Static and Voyage Related Data messages ***

sub decode_05
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-39      2   AIS Version
#  40-69     30   IMO Number
#  70-111    42   Call Sign
# 112-231   120   Vessel Name
# 232-239     8   Ship Type
# 240-248     9   Dimension to Bow
# 249-257     9   Dimension to Stern
# 258-263     6   Dimension to Port
# 264-269     6   Dimension to Starboard
# 270-273     4   EPFD Type
# 274-277     4   ETA - UTC month
# 278-282     5   ETA - UTC day
# 283-287     5   ETA - UTC hour
# 288-293     6   ETA - UTC minute
# 294-301     8   Draught
# 302-421   120   Destination
# 422         1   DTE
# 423         1   Spare

  $tyui = oct ("0b".substr ($message,38,2)); printf "AIS Version = %d (%s)\n", $tyui, $ais_version [$tyui];
  printf "IMO Number = %d\n", oct ("0b".substr ($message,40,30));
  printf "Call sign = %s\n", decode_string (substr ($message,70,42));
  printf "Vessel name = %s\n", decode_string (substr ($message,112,120));
  $tyui = oct ("0b".substr ($message,232,8)); printf "Ship Type = %d (%s)\n", $tyui, ship_type ($tyui);
  dimension_handler ("bow", oct ("0b".substr ($message,240,9)), 511);
  dimension_handler ("stern", oct ("0b".substr ($message,249,9)), 511);
  dimension_handler ("port", oct ("0b".substr ($message,258,6)), 63);
  dimension_handler ("starboard", oct ("0b".substr ($message,264,6)), 63);
  $tyui = oct ("0b".substr ($message,270,4)); printf "EPFD type = %d (%s)\n", $tyui, $epfd_type [$tyui];
  default_handler ("ETA - UTC month = %d",oct ("0b".substr ($message,274,4)),"",0);
  default_handler ("ETA - UTC day = %d",oct ("0b".substr ($message,278,5)),"",0);
  default_handler ("ETA - UTC hour = %02d",oct ("0b".substr ($message,283,5)),"",24);
  default_handler ("ETA - UTC minute = %02d",oct ("0b".substr ($message,288,6)),"",60);
  printf "Draught = %.1f m\n", oct ("0b".substr ($message,294,8)) / 10;
  printf "Destination = %s\n", decode_string (substr ($message,302,120));
  print "   ** See https://unece.org/trade/cefact/unlocode-code-list-country-and-territory for UN/LOCODE list\n";
  $tyui = substr ($message,422,1); printf "Data terminal equipment (DTE) = %d (%s)\n", $tyui, $dte [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,423,1));

  return;
  }

# *** Decode_0608 decodes Type 6, Binary Addressed and Type 8, Binary Broadcast Messages          ***
# *** See https://www.iala-aism.org/asm/ for the best list of known Types 6 and 8 message formats ***
# There are dozens of Type 6 and Type 8 formats but all have a common header, which is handled here prior to a call to other subroutines

sub decode_0608
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my ($addressed_flag,$dac,$fi,$i,$j,$len,$lines,$linemax,$tdacfi,$tyui,$z);

# List of Message Type (T), DAC, and FI for all Type 6 and 8 Application Specific Messages
# Currently support:

  my @T68ASM = (
    [ 600000, "Monitoring aids to navigation (Zeni Lite Buoy Co., Ltd.)"], # Supported
    [ 600100, "Text using 6-bit ASCII (ITU-R M.1371-3)"], # Supported
    [ 600102, "Interrogation for specified functional message (ITU-R M.1371-1)"], # Supported
    [ 600103, "Capability interrogation (ITU-R M.1371-1)"], # Supported
    [ 600104, "Capability reply (ITU-R M.1371-1)"], # Supported
    [ 600116, "Number of persons on board (IMO Circ. 289)"], # Supported
    [ 600118, "Clearance time to enter port (IMO Circ. 289)"], # Supported
    [ 600119, "Extended ship static and voyage related data (ITU-R M.1371-1)"], # Supported
    [ 600120, "Berthing data (IMO Circ. 289)"], # Supported
    [ 600123, "Area notice - addressed (IMO Circ. 289)"], # Supported
    [ 600125, "Dangerous cargo indication (IMO Circ. 289)"],
    [ 600128, "Route information - addressed (IMO Circ. 289)"], # Supported
    [ 600130, "Text description - addressed (IMO Circ. 289)"], # Supported
    [ 600132, "Tidal window (IMO Circ. 289)"],
    [ 600140, "Number of persons on board (ITU-R M.1371-1)"], # Supported
    [ 620021, "ETA at lock/bridge/terminal (EU)"],
    [ 620022, "RTA at lock/bridge/terminal (EU)"],
    [ 620055, "Number of persons on board (EU)"],
    [ 621902, "Route suggestion (Danish Maritime Authority)"],
    [ 621903, "Route suggestion reply (Danish Maritime Authority)"],
    [ 623510, "Aids to Navigation monitoring data (UK Trinity House)"], # Supported
    [ 623520, "Buoy position monitoring (UK Trinity House)"], # Supported
    [ 625010, "Aids to Navigation monitoring data (UK Trinity House)"], # Supported
    [ 631601, "Water Flow Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 631601, "Weather Station Message (Saint Lawrence Seaway Development Corporation)"],
    [ 631601, "Water Level Message (Saint Lawrence Seaway Development Corporation)"],
    [ 631601, "Wind Information Message (Saint Lawrence Seaway Development Corporation)"],
    [ 631602, "Estimated Lock Times Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 631602, "Lockage Order Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 631632, "Version Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 636601, "Water Level Message (Saint Lawrence Seaway Development Corporation)"],
    [ 636601, "Wind Information Message (Saint Lawrence Seaway Development Corporation)"],
    [ 636601, "Water Flow Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 636601, "Weather Station Message (Saint Lawrence Seaway Development Corporation)"],
    [ 636602, "Lockage Order Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 636602, "Estimated Lock Times Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 636632, "Version Message (Seaway) (Saint Lawrence Seaway Development Corporation)"],
    [ 636716, "Passenger and Crew Count - Addressed (USCG RDC)"], # Supported
    [ 636722, "Geographic Notice - addressed (USCG RDC)"], # Supported -- Not on IALA list
    [ 636729, "Linked Text (USCG RDC)"], # Supported -- Not on IALA list
    [ 636735, "Waterways Management Message (USCG RDC)"], # Supported -- Not on IALA list
    [ 642116, "Collision alarm (Sena and Vans Co., Ltd.)"],
    [ 800100, "Text using 6-bit ASCII (ITU-R M.1371-3)"], # Supported
    [ 800116, "VTS targets (targets derived by means other than AIS) (ITU-R M.1371-1)"], # Supported
    [ 800117, "VTS-generated/Synthetic targets (IMO Circ. 289)"],
    [ 800118, "Advice of waypoints (AWP) and/or route plan of VTS (ITU-R M.1371-1)"], # Supported
    [ 800119, "Marine traffic signal (IMO Circ. 289)"],
    [ 800121, "Weather observation report from ship (IMO Circ. 289)"],
    [ 800122, "Area notice - broadcast (IMO Circ. 289)"], # Supported
    [ 800124, "Extended ship static and voyage-related data (IMO Circ. 289)"],
    [ 800126, "Environmental (IMO Circ. 289)"],
    [ 800127, "Route information - broadcast (IMO Circ. 289)"], # Supported
    [ 800129, "Text description - broadcast (IMO Circ. 289)"], # Supported
    [ 800131, "Meteorological and Hydrographic data (IMO Circ. 289)"], # Supported
    [ 820010, "Inland ship static and voyage related data)"],
    [ 820019, "Control Message (EU)"],
    [ 820023, "EMMA warning (EU)"],
    [ 820024, "Water levels (EU)"],
    [ 820025, "Present Bridge Clearance message (EU)"],
    [ 820040, "Signal status (EU)"],
    [ 820041, "Signal Lights message (EU)"],
    [ 821901, "Intended route (Danish Maritime Authority)"],
    [ 822450, "Envelope/volume (gmv)"],
    [ 836619, "Lock Status (USCG RDC)"],
    [ 836620, "Dam Status (USCG RDC)"], # Supported -- Not on IALA list
    [ 836722, "Geographic Notice - broadcast (USCG RDC)"], # Supported
    [ 836723, "Satellite Ship Weather 1-Slot (USCG RDC)"], # Supported -- Not on IALA list
    [ 836724, "Satellite Ship Weather Small (USCG RDC)"], # Supported -- Not on IALA list
    [ 836725, "Satellite Ship Weather Tiny (USCG RDC)"], # Supported -- Not on IALA list
    [ 836726, "GPS Jamming and Spoofing Report (USCG RDC)"], # Supported -- Not on IALA list
    [ 836729, "Linked Text (USCG RDC)"], # Supported
    [ 836733, "Environmental Message (USCG RDC)"], # Supported
    [ 836735, "Waterways Management Message (USCG RDC)"], # Supported
    [ 871011, "Offshore Unit Dimensions (PETROBRAS)"]
    );

  my $t68asm_len = scalar @T68ASM;

# This routine parses the common header for all Type 6 and 8 messages

# Type 6, Binary Addressed Message (COMMON HEADER) 
# Type 8, Binary Broadcast Message (COMMON HEADER)
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len   Bits     Len   Description
#   0-5       6    0-5       6   Message Type         |
#   6-7       2    6-7       2   Repeat Indicator     | <-- Handled above
#   8-37     30    8-37     30   MMSI                 |
#  38-39      2   38-39      2   Sequence Number (6)/Spare (8)
#  40-69     30                  Destination MMSI
#  70         1                  Retransmit flag
#  71         1                  Spare
#  72-81     10   40-49     10   DAC
#  82-87      6   50-55      6   FI

# Set addressed_flag so that we can handle the offset for Type 6 addressed messages

  if ($type == 6)
      {
      $addressed_flag = 1;
      printf "Sequence number = %d\n", oct ("0b".substr ($message,38,2));
      print "Target "; mmsi_handler (oct ("0b".substr ($message,40,30)));
      $tyui = oct ("0b".substr ($message,70,1)); printf "Retransmit flag = %d (%s)\n", $tyui, $retransmit [$tyui];
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,71,1));
      }
    else
      {
      $addressed_flag = 0;
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));
      };

  $dac = oct ("0b".substr ($message,40+($addressed_flag*32),10));
  $fi = oct ("0b".substr ($message,50+($addressed_flag*32),6));
  $tdacfi = $type * 100000 + $dac * 100 + $fi;

  for ($i=0; $i<$t68asm_len; $i++)
    { if ($tdacfi == $T68ASM [$i][0]) { last; } }

  print "APPLICATION SPECIFIC MESSAGE TYPE: ";
  if ($i < $t68asm_len)
      { printf "%s\n", $T68ASM [$i][1]; }
    else
      { print "unknown\n"; }

# Get here if the Type and DAC/FI are valid combinations

  if ($i < $t68asm_len)
      {
      print "Designated Area Code (DAC) = $dac\n";
      printf "  ** Authority/Registrant: %s\n", dac_handler ($type,$dac);
      print "Functional Identifier (FI) = $fi\n";

# Process supported messages...

      PIECESOF8:
        {
        ($tdacfi == 600000) && do
          {
          decode_06_000_00 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600100 || $tdacfi == 800100) && do
          {
          decode_0608_001_00 ($addressed_flag,$message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600102 || $tdacfi == 600103) && do
          {
          decode_06_001_02_03 ($fi,$message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600104) && do
          {
          decode_06_001_04 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600116 || $tdacfi == 600140) && do
          {
          decode_06_001_16_40 ($fi,$message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 800116) && do
          {
          decode_08_001_16 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600118) && do
          {
          decode_06_001_18 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 800118) && do
          {
          decode_08_001_18 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600119) && do
          {
          decode_06_001_19 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600120) && do
          {
          decode_06_001_20 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600123 || $tdacfi == 800122) && do
          {
          decode_0608_001_22_23 ($addressed_flag,$message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600128 || $tdacfi == 800127) && do
          {
          decode_0608_001_27_28 ($addressed_flag,$message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 600130 || $tdacfi == 800129 || $tdacfi == 636729 || $tdacfi == 836729) && do
          {
          decode_0608_001_29_30__367_29 ($addressed_flag,$message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 800131) && do
          {
          decode_08_001_31 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 623510 || $tdacfi == 625010) && do
          {
          decode_06_235_250_10 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 623520) && do
          {
          decode_06_235_20 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 836620) && do
          {
          decode_08_366_20 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 636716) && do
          {
          decode_06_367_16 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 836723) && do
          {
          decode_08_367_23 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 836724) && do
          {
          decode_08_367_24 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 836725) && do
          {
          decode_08_367_25 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 836726) && do
          {
          decode_08_367_26 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 836733) && do
          {
          decode_08_367_33 ($message,$msg_len);
          last PIECESOF8;
          };
        ($tdacfi == 636735 || $tdacfi == 836735) && do
          {
          decode_0608_367_35 ($addressed_flag,$message,$msg_len);
          last PIECESOF8;
          };

# Get here for a valid DAC/FI combination but unsupported message

        do
          {
          printf "--- Valid but unsupported ASM ---\n";
          };
        };
      }

# Get here if unknown ASM type (i.e., unknown DAC/FI pair)

    else
      {
      print "Designated Area Code (DAC) = $dac\n";
      print "Functional ID (FI) = $fi\n";
      print "  *** Unknown DAC/FI pair. See https://www.iala-aism.org/asm/ for more information. ***\n";
      };

  return;
  }

# *** dac_handler displays the registrant associated with the Designated Area Code (DAC).

sub dac_handler
  {
  my ($type,$dac) = @_;
  my ($i,$dactype);
  my @dac_list = (
    [6000, "Zeni Lite Buoy Co., Ltd."],
    [6001, "ITU/IMO"],
    [6200, "EU"],
    [6219, "Danish Maritime Authority"],
    [6235, "UK Trinity House"],
    [6250, "UK Trinity House"],
    [6316, "Saint Lawrence Seaway Development Corporation"],
    [6366, "Saint Lawrence Seaway Development Corporation"],
    [6367, "USCG Research & Development Center (RDC)"],
    [6421, "Sena and Vans Co., Ltd."],
    [8001, "ITU/IMO"],
    [8200, "EU"],
    [8219, "Danish Maritime Authority"],
    [8224, "GMV Innovating Solutions S.L."],
    [8366, "USCG Research & Development Center (RDC)"],
    [8367, "USCG Research & Development Center (RDC)"],
    [8710, "PETROBRAS"],
    );
  my $dac_list_len = scalar @dac_list;

  $dactype = $type * 1000 + $dac;
  for ($i=0; $i < $dac_list_len; $i++)
    {
    if ($dactype == $dac_list [$i][0])
      { return $dac_list [$i][1]; }
    }
    
  return "(Unknown)";
  }
    
# *** Decode_06_000_00 decodes Type 6, DAC 0, FI 0 (Monitoring aids to navigation) [Zeni Lite Buoy Co., Ltd.] messages ***

sub decode_06_000_00
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $x;

################################################
#  Bits     Len   Description
#  88-103    16   Sub-application ID (set to 1)
# 104-115    12   Voltage data
# 116-125    10   Current data
# 126         1   Power supply type
# 127         1   Light status
# 128         1   Battery status
# 129         1   Off-position status
# 130-135     6   Spare (set to 0)

  printf "Sub-application identifier = %d\n", oct ("0b".substr ($message,88,16));
  printf "Lantern supply voltage = %.1f volts (V)\n", oct ("0b".substr ($message,104,12)) / 10;
  printf "Lantern drain current = %.1f amps (A)\n", oct ("0b".substr ($message,116,10)) / 10;
  $x = oct ("0b".substr ($message,126,1)); printf "Power supply type = %d (%s)\n", $x, $power_supply [$x];
  $x = oct ("0b".substr ($message,127,1)); printf "Light status = %d (%s)\n", $x, $light_status [$x];
  $x = oct ("0b".substr ($message,128,1)); printf "Battery status = %d (%s)\n", $x, $battery_status [$x];
  $x = oct ("0b".substr ($message,129,1)); printf "Off-position status = %d (%s)\n", $x, $offpos [$x];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,130,6));
  return;
  }

# *** Decode_0608_001_00 decodes Type 6 and Type 8, DAC 1, FI 00 (Text using 6-bit ASCII - Addressed//Broadcast) [ITU Rec. M.1371] messages ***

sub decode_0608_001_00
  {
# Parameter list
  my ($addressed_flag,$message,$msg_len) = @_;
  my ($c_len,$tyui);

# Type 6 - "Text using 6-bit ASCII - Addressed" (DAC = 1; FI = 0)
# Type 8 - "Text using 6-bit ASCII - Broadcast" (DAC = 1; FI = 0)
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  88         1    56         1   Acknowledgement Required flag
#  89-99     11    57-67     11   Text Sequence Number
# 100-1005  906    68-1003  936   Text String (<=151 [6]/<=156 [8] six-bit characters)
#1006-1008    3  1004-1008    5   Spare (padding for octet-alignment, as needed)

  $tyui = substr ($message,268,1); printf "Acknowledged required flag = %d (%s)\n", $tyui, $ack_req [$tyui];
  $tyui = oct ("0b".substr ($message,57+($addressed_flag*32),11)); printf "Text Sequence Number = %d\n", $tyui;

# $c_len adjusts for the octet-alignment bits

  $c_len = $msg_len - 68 - ($addressed_flag*32);
  $c_len -= ($c_len % 6);
  printf "Text = %s\n", decode_string (substr ($message,68+($addressed_flag*32),$c_len));

  return;
  }

# *** Decode_06_001_02_03 decodes Type 6, DAC 1, FI 2 (Interrogation for a specific functional message) and FI 3
# (Capability interrogation) [ITU Rec. M.1371] messages ***

sub decode_06_001_02_03
  {
# Parameter list
  my ($fi,$message,$msg_len) = @_;

################################################
#  Bits     Len   Description
#  88-97     10   Requested DAC code
#  98-103     6   Requested FI code          [FI = 2]
# 104-167    64   Spare (set to 0)           [FI = 2]

#  98-167    70   Spare (set to 0)           [FI = 3]

  printf "Requested DAC code = %d\n", oct ("0b".substr ($message,88,10));
  if ($fi == 2)
      {
      printf "Requested FI code = %d\n", oct ("0b".substr ($message,98,6));
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,104,64));
      }
    else
      { printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,98,70)); }
  return;
  }

# *** Decode_06_001_04 decodes Type 6, DAC 1, FI 4 (Capability reply) [ITU Rec. M.1371] messages ***

sub decode_06_001_04
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($bpos,$i,$j);

################################################
#  Bits     Len   Description
#  88-97     10   DAC code being interrogated
#  98-225   128   FI capability table
# 226-351   126   Spare (set to 0)

  printf "Requested DAC code for interrogation = %d\n", oct ("0b".substr ($message,88,10));

  printf "\n----- Functional Identifier (FI) Capability Table for Designated Area Code (DAC) %d -----\n\n", oct ("0b".substr ($message,88,10));
  print "FI availability (A): - = Not supported [default] (0); Y = Supported (1)\n";
  print "Spare bit (S): Unused and reserved for future use; should be 0\n\n";
  print "       FI  A  S       FI  A  S       FI  A  S       FI  A  S\n";
  print "       --  -  -       --  -  -       --  -  -       --  -  -\n";

  for ($i=0; $i<=15; $i++)
    {
    $bpos = $i*2 + 98;
    for ($j=0; $j<=3; $j++)
      {
      printf "       %2d  ", $i + ($j*16);
      if (oct(substr ($message,$bpos+($j*32),1)) == 0) { print "-  "; }
        else { print "Y  "; }
      printf "%d", oct("0b".substr ($message,$bpos+($j*32)+1,1));
      if ($j == 3) { print "\n"; }
      }
    }

  printf "\n   ------------------------------------------------------------------------------------\n\n";
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,226,126));
  return;
  }

# *** Decode_06_001_16_40 decodes Type 6, DAC 1, FI 16 (Number of Persons On Board - Addressed) [IMO Circular 289] and FI 40
# (Number of Persons on Board) [ITU Rec. M.1371] messages ***

sub decode_06_001_16_40
  {
# Parameter list
  my ($fi,$message,$msg_len) = @_;

################################################
#  Bits     Len   Description
#  88-100    13   Number of persons
# 101-135    35   Spare (set to 0)          FI = 16
# 101-103     3   Spare (set to 0)          FI = 40

  two_max_handler ("Number of persons on board = %d", oct ("0b".substr ($message,88,13)),"",8191,0);
  if ($fi == 16)
      { printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,101,35)); }
    else
      { printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,101,3)); }
  return;
  }

# *** Decode_08_001_16 decodes Type 8, DAC 1, FI 16 (VTS Targets) [ITU-R M.1371-1] messages ***

sub decode_08_001_16
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$id,$num,$tyui);

################################################
#  Bits     Len   Description
#  56-175   120   VTS Target 1
# 176-295   120   VTS Target 2 (optional)
# 296-415   120   VTS Target 3 (optional)
# 416-535   120   VTS Target 4 (optional)
# 536-655   120   VTS Target 5 (optional)
# 656-775   120   VTS Target 6 (optional)
# 776-895   120   VTS Target 7 (optional)

# Target Report
#-----------------------------------------------
#  Bits     Len   Description
#   0-1       2   Identifier Type
#   2-43     42   Target ID
#  44-47      4   Spare (set to 0)
#  48-71     24   Latitude
#  72-96     25   Longitude
#  97-105     9   Course over ground
# 106-111     6   Timestamp
# 112-119     8   Speed over ground

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

  $num = ($msg_len - 56) / 120;
  for ($i = 0; $i < $num; $i++)
    {
    printf "VTS TARGET REPORT %d\n", $i+1;
    $tyui = oct ("0b".substr ($message,56+$i*120,2)); printf "  VTS target identifier type = %d (%s)\n", $tyui, $vts_id_type [$tyui];
    $id = oct ("0b".substr ($message,58+$i*120,42));
    TARGET_ID:
      {
      ($tyui == 0) && do
        {
        print "  ";
        mmsi_handler ($id);
        next TARGET_ID;
        };
      ($tyui == 1) && do
        {
        printf "  IMO Number = $id\n", $id;
        next TARGET_ID;
        };
      ($tyui == 2) && do
        {
        printf "  Call Sign = %s\n", decode_string (sprintf ("%042b",$id)) =~ s/@//gr;
        next TARGET_ID;
        };
      ($tyui == 3) && do
        {
        printf "  Target id. = %d\n", $id;
        next TARGET_ID;
        };
      };
    printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,100+$i*120,4));
    long_lat_handler ("  Longitude", (signed_bin2dec (substr ($message,128+$i*120,25)) / 60000),181, "  Latitude", (signed_bin2dec (substr ($message,104+$i*120,24)) / 60000),91);
    default_handler ("  Course over ground (COG) = %d", oct ("0b".substr ($message,153+$i*120,9)), "\xB0", 360);
    default_handler ("  Timestamp = %d", oct ("0b".substr ($message,162+$i*120,6)), " seconds", 60);
    default_handler ("  Speed over ground (SOG) = %d", oct ("0b".substr ($message,168+$i*120,8)), " kn", 255);
    }
  return;
  }

# *** Decode_06_001_18 decodes Type 6, DAC 1, FI 18 (Clearance Time to Enter Port - Addressed) [IMO Circular 289] messages ***

sub decode_06_001_18
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  88-97     10   Message Linkage ID
#  98-101     4   Clearance time to enter port -- UTC month
# 102-106     5   Clearance time to enter port -- UTC day
# 107-111     5   Clearance time to enter port -- UTC hour
# 112-117     6   Clearance time to enter port -- UTC minute
# 118-237   120   Name of Port and Berth
# 238-267    30   Destination
# 268-292    25   Longitude
# 293-316    24   Latitude
# 317-359    43   Spare (set to 0)

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,88,10)),"",0);
  default_handler ("Clearance time to enter port -- UTC month = %d",oct ("0b".substr ($message,98,4)),"",0);
  default_handler ("Clearance time to enter port -- UTC day = %d",oct ("0b".substr ($message,102,5)),"",0);
  default_handler ("Clearance time to enter port -- UTC hour = %02d",oct ("0b".substr ($message,107,5)),"",24);
  default_handler ("Clearance time to enter port -- UTC minute = %02d",oct ("0b".substr ($message,112,6)),"",60);
  printf "Name of port & berth = %s\n", decode_string (substr ($message,118,120));
  printf "Destination = %s\n", decode_string (substr ($message,238,30));
  print "   ** See https://unece.org/trade/cefact/unlocode-code-list-country-and-territory for UN/LOCODE list\n";
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,268,25)) / 60000),181, "Latitude", (signed_bin2dec (substr ($message,293,24)) / 60000),91);
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,317,43));
  return;
  }

# *** Decode_08_001_18 decodes Type 8, DAC 1, FI 18 (Advice of Waypoints/VTS Route Plan) [ITU-R.M.1371-1] messages ***

sub decode_08_001_18
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$num,$tyui);
  my @spare = (4,1,6,3,0,5,2,7,4,1,6,3,0);

################################################
#  Bits     Len   Description
#  56-59      4   Number of waypoints
#  60-126    67   Waypoint 1 detail (optional)
# 127-193    67   Waypoint 2 detail (optional)
# 194-260    67   Waypoint 3 detail (optional)
# 261-327    67   Waypoint 4 detail (optional)
# 328-394    67   Waypoint 5 detail (optional)
# 395-361    67   Waypoint 6 detail (optional)
# 462-528    67   Waypoint 7 detail (optional)
# 529-595    67   Waypoint 8 detail (optional)
# 596-662    67   Waypoint 9 detail (optional)
# 663-729    67   Waypoint 10 detail (optional)
# 730-796    67   Waypoint 11 detail (optional)
# 797-863    67   Waypoint 12 detail (optional)
# 864-983   120   Route description

# Waypoint detail
#-----------------------------------------------
#  Bits     Len   Description
#   0-27     28   Longitude
#  28-54     27   Latitude
#  55-66     12   Turning radius

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  $num = oct ("0b".substr ($message,56,4)); print "Number of waypoints = $num\n";
  for ($i = 1; $i <= $num; $i++)
    {
    print "ADVISED WAYPOINT/VTS ROUTE PLAN #$i\n";
    long_lat_handler ("  Longitude", (signed_bin2dec (substr ($message,60+($i-1)*67,28)) / 600000),181,
                       "  Latitude", (signed_bin2dec (substr ($message,88+($i-1)*67,27)) / 600000),91);
    default_handler ("  Turning radius = %d", oct ("0b".substr ($message,115+($i-1)*67,12)), " m", 0);
    }

  $tyui = (decode_string (substr ($message,60+($num*67),120)) =~ s/@//gr);
  if ($tyui eq "") { $tyui = "<null>"; }
  print "Route description = $tyui\n";
  if ($spare[$num] != 0)
    { printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,180+($num*67),$spare[$num])); }
  return;
  }

# *** Decode_06_001_19 decodes Type 6, DAC 1, FI 19 (Extended ship static and voyage related data) [ITU M.1371] messages ***

sub decode_06_001_19
  {
# Parameter list
  my ($message,$msg_len) = @_;

################################################
#  Bits     Len   Description
#  88-98     11   Height over keel
#  99-103     5   Spare (set to 0)

  two_max_handler ("Height over keel = %.1f", (oct ("0b".substr ($message,88,11))/10)," m",204.7,0);
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,99,5));
  return;
  }

# *** Decode_06_001_20 decodes Type 6, DAC 1, FI 20 (Berthing Data - Addressed) [IMO Circular 289] messages ***

sub decode_06_001_20
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$tyui,$svc_ind);

################################################
#  Bits     Len   Description
#  88-97     10   Message Linkage ID
#  98-106     9   Berth Length
# 107-114     8   Water depth at berth
# 115-117     3   Mooring Position
# 118-121     4   UTC month
# 122-126     5   UTC day
# 127-131     5   UTC hour
# 132-137     6   UTC minute
# 138         1   Services Availability
# 139-190    52   Type of Services Available
# 191-310   120   Name of Berth
# 311-335    25   Centre of Berth, Longitude
# 336-359    24   Centre of Berth, Latitude

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,88,10)),"",0);
  default_handler ("Berth length = %d",oct ("0b".substr ($message,98,9))," m",0);
  default_handler ("Water depth at berth = %d",oct ("0b".substr ($message,107,8))/10," m",0);
  $tyui = oct ("0b".substr ($message,115,3)); printf "Mooring position = %d (%s)\n", $tyui, $mooring_position [$tyui];
  default_handler ("Berth date and time - UTC month = %d",oct ("0b".substr ($message,118,4)),"",0);
  default_handler ("Berth date and time - UTC day = %d",oct ("0b".substr ($message,122,5)),"",0);
  default_handler ("Berth date and time - UTC hour = %02d",oct ("0b".substr ($message,127,5)),"",24);
  default_handler ("Berth date and time - UTC minute = %02d",oct ("0b".substr ($message,132,6)),"",60);

  $tyui = substr ($message,138,1);
  if ($tyui == 0)
      {
      print "Service availability = 0 (Service types unknown)\n";
      printf "Types of service = %52s (should be 0)\n";
      }
    else
      {
      print "Service availability = 1 (Service types listed below)\n";
      for ($i=0; $i<22; $i++)
        {
        $svc_ind = oct ("0b".substr ($message,139+($i*2),2));
        printf "  ** %s service =  %d (%s)\n", $service_list [$i+1], $svc_ind, $service_avail [$svc_ind];
        }
      for ($i=22; $i<26; $i++)
        {
        printf "  ** %s =  %d (Should be 0)\n", $service_list [$i+1], oct ("0b".substr ($message,139+($i*2),2));
        }
      }

  printf "Name of berth = %s\n", decode_string (substr ($message,191,120));
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,311,25)) / 60000),181, "Latitude", (signed_bin2dec (substr ($message,336,24)) / 60000),91);

  return;
  }

###########
# *** Decode_0608_001_22_23 decodes Type 6, DAC 1, FI 23 (Area Notice - Addressed) and
# Type 8, DAC 1, FI 22 (Area Notice - Broadcast) [IMO Circular 289] messages ***

sub decode_0608_001_22_23
  {
# Parameter list
  my ($addressed_flag,$message,$msg_len) = @_;
  my ($i,$j,$num_shapes,$sa_base,$scale_factor,$shape,$tyui);

# Type 6 - "Area Notice - Addressed" (DAC = 1; FI = 23)
# Type 8 - "Area Notice - Broadcast" (DAC = 1; FI = 22)
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  88-97     10    56-65     10   Message Linkage ID
#  98-104     7    66-72      7   Notice Description
# 105-108     4    73-76      4   UTC month
# 109-113     5    77-81      5   UTC day
# 114-118     5    82-86      5   UTC hour
# 119-124     6    87-92      6   UTC minute
# 125-142    18    93-110    18   Duration
# 143-229    87   111-197    87   Sub-area 1
# 230-316    87   198-284    87   Sub-area 2 (if present)
# 317-403    87   285-371    87   Sub-area 3 (if present)
# 404-490    87   372-458    87   Sub-area 4 (if present)
# 491-577    87   459-545    87   Sub-area 5 (if present)
# 578-664    87   546-632    87   Sub-area 6 (if present)
# 665-751    87   633-719    87   Sub-area 7 (if present)
# 752-838    87   720-806    87   Sub-area 8 (if present)
# 839-925    87   807-893    87   Sub-area 9 (if present)
# 926-1012   87   894-980    87   Sub-area 10 (if present)

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,56+($addressed_flag*32),10)),"",0);
  $tyui = oct ("0b".substr ($message,66+($addressed_flag*32),7)); printf "Notice Description = %d (%s)\n", $tyui, $notice_desc [$tyui];
  default_handler ("UTC month = %d",oct ("0b".substr ($message,73+($addressed_flag*32),4)),"",0);
  default_handler ("UTC day = %d",oct ("0b".substr ($message,77+($addressed_flag*32),5)),"",0);
  default_handler ("UTC hour = %02d",oct ("0b".substr ($message,82+($addressed_flag*32),5)),"",24);
  default_handler ("UTC minute = %02d",oct ("0b".substr ($message,87+($addressed_flag*32),6)),"",60);

  $tyui = oct ("0b".substr ($message,93+($addressed_flag*32),18));
  if ($tyui == 0)
      { print "Duration = 0 (Cancel Area Notice)\n"; }
    else
      {
      if ($tyui == 262143)
          { print "Duration = 262,143 (Undefined [default])\n"; }
        else
          { printf "Duration = %d minutes (%.2f days) until the end of Area Notice\n", $tyui, $tyui/1440; } # 1,440 min/day
      };

# Now handle the sub-areas

  $num_shapes = ($msg_len-(111+($addressed_flag*32)))/87;
  for ($i=1; $i <= $num_shapes; $i++)
    {
    print "--- Sub-area #$i ---\n";

# Start counting at bit value 111 + ($i-1)*87

    $sa_base = 111 + ($addressed_flag*32) + (($i-1) * 87);

# Fields common to all sub-areas

################################################
#  Bits     Len   Description
#   0-2       3   Area Shape

    $shape = oct ("0b".substr ($message,$sa_base,3)); printf "Area Shape = %d (%s)\n", $shape, $area_shape [$shape];

# The following fields are common to shapes 0-4

################################################
#  Bits     Len   Description
#   3-4       2   Scale Factor

    if ($shape <= 4)
      {
      $tyui = oct ("0b".substr ($message,$sa_base+3,2));
      $scale_factor = 10**$tyui; printf "Scale Factor = %d (Factor of %d)\n", $tyui, $scale_factor;
      }

# The following fields are common to shapes 0-2

################################################
#  Bits     Len   Description       FIELD
#   5-29     25   Longitude
#  30-53     24   Latitude
#  54-56      3   Precision

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

    if ($shape <=2)
      {
      long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,$sa_base+5,25)) / 60000),181,
       "Latitude", (signed_bin2dec (substr ($message,$sa_base+30,24)) / 60000),91);
      default_handler ("Precision of latitude/longitude = %d",oct ("0b".substr ($message,$sa_base+54,3))," decimal places",4);
      };

# Now handle the other shapes individually

    SHAPE:
      {

# SHAPE = 0
################################################
#  Bits     Len   Description
#  57-68     12   Radius
#  69-86     18   Not used

      $shape == 0 && do
        {
        $tyui = oct ("0b".substr ($message,$sa_base+57,12));
        if ($tyui == 0)
            { print "Radius = 0 (point)\n"; }
          else
            { printf "Radius of circle = %d (%d m, %.3f nm)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+69,18));
        last SHAPE;
        };

# SHAPE = 1
################################################
#  Bits     Len   Description
#  57-64      8   E Dimension
#  65-72      8   N Dimension
#  73-81      9   Orientation
#  82-86      5   Not used

      $shape == 1 && do
        {
        $tyui = oct ("0b".substr ($message,$sa_base+57,8));
        if ($tyui == 0)
            { print "East Dimension = 0 (North-South line; default)\n"; }
          else
            { printf "East Dimension = %d (%d m, %.3f nm from corner point)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        $tyui = oct ("0b".substr ($message,$sa_base+65,8));
        if ($tyui == 0)
            { print "North Dimension = 0 (East-West line; default)\n"; }
          else
            { printf "North Dimension = %d (%d m, %.3f nm from corner point)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        $tyui = oct ("0b".substr ($message,$sa_base+73,9));
        if ($tyui == 0)
            { print "Orientation = 0 (no rotation [default])\n"; }
          else
            {
            if ($tyui >= 1 && $tyui <= 359)
                { printf "Orientation of area = %03d\xb0T\n", $tyui; }
              else
                { print "Orientation = $tyui (invalid; reserved for future use)\n"; }
            };

        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+82,5));
        last SHAPE;
        };

# SHAPE = 2
################################################
#  Bits     Len   Description
#  57-68     12   Radius
#  69-77      9   Left Boundary
#  78-86      9   Right Boundary

      $shape == 2 && do
        {
        $tyui = oct ("0b".substr ($message,$sa_base+57,12));
        if ($tyui == 0)
            { print "Radius = 0 (point)\n"; }
          else
            { printf "Radius of circle = %d (%d m, %.3f nm)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        $tyui = oct ("0b".substr ($message,$sa_base+69,9));
        if ($tyui == 0)
            { print "Orientation of left boundary = 0 (no rotation [default])\n"; }
          else
            {
            if ($tyui >= 1 && $tyui <= 359)
                { printf "Orientation of left boundary edge = %03d\xb0T\n", $tyui; }
              else
                { print "Orientation of left boundary = $tyui (invalid; reserved for future use)\n"; }
            };
        $tyui = oct ("0b".substr ($message,$sa_base+78,9));
        if ($tyui == 0)
            { print "Orientation of right boundary = 0 (no rotation [default])\n"; }
          else
            {
            if ($tyui >= 1 && $tyui <= 359)
                { printf "Orientation of right boundary edge = %03d\xb0T\n", $tyui; }
              else
                { print "Orientation of right boundary = $tyui (invalid; reserved for future use)\n"; }
            };

        last SHAPE;
        };

# SHAPES 3 AND 4
################################################
#  Bits     Len   Description
#   5-14     10   Point 1 Angle
#  15-24     10   Point 1 Distance
#  25-34     10   Point 2 Angle
#  35-44     10   Point 2 Distance
#  45-54     10   Point 3 Angle
#  55-64     10   Point 3 Distance
#  65-74     10   Point 4 Angle
#  75-84     10   Point 4 Distance
#  85-86      2   Not used

      ($shape == 3 || $shape == 4) && do
        {
        for ($j=0; $j<4; $j++)
          {
          printf "Point %d to %d =", $j, $j+1;
          $tyui = oct ("0b".substr ($message,$sa_base+5+($j*20),10));
          if ($tyui < 720)
              { printf " Bearing: %3d (%05.1f\xB0T)", $tyui, $tyui/2; }
            else
              {
              if ($tyui == 720)
                  { print " Bearing: $tyui (not available [default])"; }
                else
                  { print " Bearing: $tyui (reserved for future use)"; }
              };
          $tyui = oct ("0b".substr ($message,$sa_base+15+($j*20),10));
          if ($tyui == 0)
              { print "   Distance: 0 (no point)\n"; }
            else
              { printf "   Distance: %d (%d m, %.3f nm)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
          };
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+85,2));
        last SHAPE;
        };

# SHAPE 5
################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (5)      0
#   3-86     84   Text                1   (14 6-bit characters)

      $shape == 5 && do
        {
        printf "Text = %s\n", decode_string (substr ($message,$sa_base+3,84));
        last SHAPE;
        };

     do
        { print "Area Shape = $shape (invalid)\n"; };
      };
    };

  return;
  }

# *** Decode_0608_001_27_28 decodes Type 6, DAC 1, FI 28 (Route Information - Addressed) and
# Type 8, DAC 1, FI 27 (Route Information - Broadcast) [IMO Circular 289] messages***

sub decode_0608_001_27_28
  {
# Parameter list
  my ($addressed_flag,$message,$msg_len) = @_;
  my ($i,$numpoints,$tyui);

# Type 6 - "Route Information - Addressed" (DAC = 1; FI = 27)
# Type 8 - "Route Information - Broadcast" (DAC = 1; FI = 28)
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#   88-97     10    56-65     10   Message Linkage ID
#   98-100     3    66-68      3   Sender Classification
#  101-105     5    69-73      5   Route Type
#  106-109     4    74-77      4   UTC month
#  110-114     5    78-82      5   UTC day
#  115-119     5    83-87      5   UTC hour
#  120-125     6    88-93      6   UTC minute
#  126-143    18    94-111    18   Duration
#  144-148     5   112-116     5   Number of waypoints
#  149-176    28   117-144    28   Longitude 1 (if present)
#  177-203    27   145-171    27   Latitude 1 (if present)
#  204-231    28   172-199    28   Longitude 2 (if present)
#  232-258    27   200-226    27   Latitude 2 (if present)
#  259-286    28   227-254    28   Longitude 3 (if present)
#  287-313    27   255-281    27   Latitude 3 (if present)
#  314-341    28   282-309    28   Longitude 4 (if present)
#  342-368    27   310-336    27   Latitude 4 (if present)
#  369-396    28   337-364    28   Longitude 5 (if present)
#  397-423    27   365-391    27   Latitude 5 (if present)
#  424-451    28   392-419    28   Longitude 6 (if present)
#  452-478    27   420-446    27   Latitude 6 (if present)
#  479-506    28   447-474    28   Longitude 7 (if present)
#  507-533    27   475-501    27   Latitude 7 (if present)
#  534-561    28   502-529    28   Longitude 8 (if present)
#  562-588    27   530-556    27   Latitude 8 (if present)
#  589-616    28   557-584    28   Longitude 9 (if present)
#  617-643    27   585-611    27   Latitude 9 (if present)
#  644-671    28   612-639    28   Longitude 10 (if present)
#  672-698    27   640-666    27   Latitude 10 (if present)
#  699-726    28   667-694    28   Longitude 11 (if present)
#  727-753    27   695-721    27   Latitude 11 (if present)
#  754-781    28   722-749    28   Longitude 12 (if present)
#  782-808    27   750-776    27   Latitude 12 (if present)
#  809-836    28   777-804    28   Longitude 13 (if present)
#  837-863    27   805-831    27   Latitude 13 (if present)
#  864-891    28   832-859    28   Longitude 14 (if present)
#  892-918    27   860-886    27   Latitude 14 (if present)
#  929-946    28   887-914    28   Longitude 15 (if present)
#  947-973    27   915-941    27   Latitude 15 (if present)
#  974-1001   28   942-969    28   Longitude 16 (if present)
# 1002-1028   27   970-996    27   Latitude 16 (if present)

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,56+($addressed_flag*32),10)),"",0);
  $tyui = oct ("0b".substr ($message,66+($addressed_flag*32),3)); printf "Sender classification = %d (%s)\n", $tyui, $senderclass [$tyui];
  $tyui = oct ("0b".substr ($message,69+($addressed_flag*32),5)); printf "Route type = %d (%s)\n", $tyui, $routetype [$tyui];
  default_handler ("UTC month = %d",oct ("0b".substr ($message,74+($addressed_flag*32),4)),"",0);
  default_handler ("UTC day = %d",oct ("0b".substr ($message,78+($addressed_flag*32),5)),"",0);
  default_handler ("UTC hour = %02d",oct ("0b".substr ($message,83+($addressed_flag*32),5)),"",24);
  default_handler ("UTC minute = %02d",oct ("0b".substr ($message,88+($addressed_flag*32),6)),"",60);

  $tyui = oct ("0b".substr ($message,94+($addressed_flag*32),18));
  if ($tyui == 0)
      { print "Duration = 0 (Cancel Route)\n"; }
    else
      {
      if ($tyui == 262143)
          { print "Duration = 262,143 (Not available [default])\n"; }
        else
          { printf "Duration = %d minutes (%.2f days) until the end of route validity\n", $tyui, $tyui/1440; } # 1,440 min/day
      };

  $numpoints = oct ("0b".substr ($message,112+($addressed_flag*32),5)); printf "Number of waypoints: %d\n", $numpoints;
  if ($numpoints > 0)
    {
    for ($i=0; $i<$numpoints; $i++)
      {
      printf "-- Waypoint #%d --\n", $i+1;
      long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,117+($i*55)+($addressed_flag*32),28)) / 600000),181,
      "Latitude", (signed_bin2dec (substr ($message,145+($i*55)+($addressed_flag*32),27)) / 600000),91);
      }
    }

  return;
  }

# *** Decode_0608_001_29_30__367_29 decodes Type 6, DAC 1, FI 30 Text Description - Addressed) and
# Type 8, DAC 1, FI 29 (Text Description - Broadcast) [IMO Circular 289] messages, as well as
# Types 6 and 8, DAC 367, FI 29 (Linked Text - Addresssed/Broadcast) [USCG RDC] messages***

sub decode_0608_001_29_30__367_29
  {
# Parameter list
  my ($addressed_flag,$message,$msg_len) = @_;

# Type 6 - "Text Description - Addressed" (DAC = 1; FI = 30)
# Type 8 - "Text Description - Broadcast" (DAC = 1; FI = 29)
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  88-97     10    56-65     10   Message Linkage ID
#  98-1027  930    66-1031  966   Free text

  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,56+($addressed_flag*32),10)),"",0);
  printf "Text = %s\n", decode_string (substr ($message,66+($addressed_flag*32),min($msg_len-66-($addressed_flag*32),930+($addressed_flag*36))));

  return;
  }

# *** Decode_08_001_31 decodes Type 8, DAC = 1, FI = 31 (Meteorological and Hydrographic Data) [IMO Circular 289] messages ***

sub decode_08_001_31
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  56-80     25   Longitude
#  81-104    24   Latitude
# 105         1   Position accuracy             ** Not currently supported; set to 1
# 106-110     5   UTC day
# 111-115     5   UTC hour
# 116-121     6   UTC minute
# 122-128     7   Average Wind Speed
# 129-135     7   Gust Speed
# 136-144     9   Wind Direction
# 145-153     9   Wind Gust Direction
# 154-164    11   Air Temperature
# 165-171     7   Relative Humidity             ** Not currently supported; set to 101
# 172-181    10   Dew Point                     ** Not currently supported; set to 501
# 182-190     9   Air Pressure
# 191-192     2   Pressure Tendency
# 193-200     8   Maximum Visibility
# 201-212    12   Water Level                   ** Not currently supported; set to 4001
# 213-214     2   Water Level Trend             ** Not currently supported; set to 3
# 215-222     8   Surface Current Speed         ** Not currently supported; set to 255
# 223-231     9   Surface Current Direction     ** Not currently supported; set to 360
# 232-239     8   Surface Current Speed #2      ** Not currently supported; set to 255
# 240-248     9   Surface Current Direction #2  ** Not currently supported; set to 360
# 249-253     5   Measurement Depth #2          ** Not currently supported; set to 31
# 254-261     8   Surface Current Speed #3      ** Not currently supported; set to 255
# 262-270     9   Surface Current Direction #3  ** Not currently supported; set to 360
# 271-275     5   Measurement Depth #3          ** Not currently supported; set to 31
# 276-283     8   Wave Height
# 284-289     6   Wave Period
# 290-298     9   Wave Direction
# 299-306     8   Swell Height
# 307-312     6   Swell Period
# 313-321     9   Swell Direction
# 322-325     4   Sea State
# 326-335    10   Water Temperature             ** Not currently supported; set to 501
# 336-338     3   Precipitation
# 339-347     9   Salinity                      ** Not currently supported; set to 511
# 348-349     2   Ice
# 350-359    10   Spare

# NOTE: Longitude and latitude are scaled by multiplying by 60,000

# Handle DAC = 1, FI = 31 (Meteorological and Hydrographic Data) messages

  print "--- Meteorological and Hydrographic Data Application-Specific Message (ASM) ---\n";
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,56,25)) / 60000),181, "Latitude", (signed_bin2dec (substr ($message,81,24)) / 60000),91);
  $tyui = substr ($message,105,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  default_handler ("UTC day = %d",oct ("0b".substr ($message,106,5)),"",0);
  default_handler ("UTC hour = %02d",oct ("0b".substr ($message,111,5)),"",24);
  default_handler ("UTC minute = %02d",oct ("0b".substr ($message,116,6)),"",60);
  two_max_handler ("10-minute average wind speed = %d", (oct ("0b".substr ($message,122,7)))," kn",126,127);
  two_max_handler ("10-minute maximum wind (gust) speed = %d", (oct ("0b".substr ($message,129,7)))," kn",126,127);
  default_handler ("Wind direction = %03d", oct ("0b".substr ($message,136,9)), "\xB0", 360);
  default_handler ("Wind gust direction = %03d", oct ("0b".substr ($message,145,9)), "\xB0", 360);
  default_handler ("Air temperature = %.1f", signed_bin2dec (substr ($message,154,11)), "\xB0C", -102.4);
  default_handler ("Relative humidity = %d", oct ("0b".substr ($message,165,7)), "\%", 101);
  default_handler ("Dew point = %.1f", (signed_bin2dec (substr ($message,172,10),10)/10), "\xB0C", 50.1);
  air_press_handler ("Air pressure = ", oct ("0b".substr ($message,182,9)), 511);
  $tyui = oct ("0b".substr ($message,191,2)); printf "Pressure tendency (trend) = %d (%s)\n", $tyui, $tendency [$tyui];
  default_handler ("Maximum visibility = %.1f", (oct ("0b".substr ($message,193,8))/10), " nm", 127);

# Handle water level

  $tyui = oct ("0b".substr ($message,201,12)) / 100;
  print "Water level = ";
  if ($tyui == 40.01)
      { printf "%.2f (Information not available)\n", $tyui; }
    else
      { printf "%.2f m\n", $tyui-10; }

  $tyui = oct ("0b".substr ($message,213,2)); printf "Water level tendency (trend) = %d (%s)\n", $tyui, $tendency [$tyui];
  two_max_handler ("Surface current speed = %.1f", (oct ("0b".substr ($message,215,8))/10)," kn",25.1,25.5);
  default_handler ("Surface current direction = %03d", oct ("0b".substr ($message,223,9)), "\xB0", 360);
  two_max_handler ("Current speed #2 = %.1f", (oct ("0b".substr ($message,232,8))/10)," kn",25.1,25.5);
  default_handler ("Current direction #2 = %03d", oct ("0b".substr ($message,240,9)), "\xB0", 360);
  default_handler ("Measurement depth #2 = %d", (oct ("0b".substr ($message,249,5))), " m", 31);
  two_max_handler ("Current speed #3 = %.1f", (oct ("0b".substr ($message,254,8))/10)," kn",25.1,25.5);
  default_handler ("Current direction #3 = %03d", oct ("0b".substr ($message,262,9)), "\xB0", 360);
  default_handler ("Measurement depth #3 = %d", (oct ("0b".substr ($message,271,5))), " m", 31);
  two_max_handler ("Wave height = %.1f", (oct ("0b".substr ($message,276,8))/10)," m",25.1,25.5);
  default_handler ("Wave period = %d", oct ("0b".substr ($message,284,6)), " sec.", 63);
  default_handler ("Wave direction = %03d", oct ("0b".substr ($message,290,9)), "\xB0", 360);
  two_max_handler ("Swell height = %.1f", (oct ("0b".substr ($message,299,8))/10)," m",25.1,25.5);
  default_handler ("Swell period = %d", oct ("0b".substr ($message,307,6)), " sec.", 63);
  default_handler ("Swell direction = %03d", oct ("0b".substr ($message,313,9)), "\xB0", 360);
  $tyui = oct ("0b".substr ($message,322,4)); printf "Sea state (Beaufort scale) = %d (%s)\n", $tyui, $Beaufort [$tyui];
  default_handler ("Water temperature = %.1f", (signed_bin2dec (substr ($message,326,10),10)/10), "\xB0C", 50.1);
  $tyui = oct ("0b".substr ($message,336,3)); printf "Precipitation = %d (%s)\n", $tyui, $precip_type [$tyui];

# Handle the oddities of salinity!

  $tyui = oct ("0b".substr ($message,339,9)) / 10;
  print "Salinity = ";
  if ($tyui == 51.1)
      { printf "%.1f (Sensor not available)", $tyui; }
    else
      {
      if ($tyui == 51.0)
          { printf "%.1f (Information not available)", $tyui; }
        else
          { 
          printf "%.1f\%", $tyui;
          if ($tyui > 50)
            { print " or more"; }
          }
        };

  print "\n";

  $tyui = oct ("0b".substr ($message,348,2)); printf "Ice = %d (%s)\n", $tyui, $ice_type [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,350,10));

  return;
  }

# *** Decode_06_235_250_10 decodes Type 6, DAC 235/250, FI 10 (Aids to navigation monitoring data) [UK Trinity House] messages ***

sub decode_06_235_250_10
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($x,$y);
  my @state = ("Off", "On");

################################################
#  Bits     Len   Description
#  88-97     10   Analogue (internal supply voltage to AIS unit)
#  98-107    10   Analogue (external – from hardware analogue input No 1)
# 108-117    10   Analogue (external – from hardware analogue input No 2)
# 118-122     5   Status (internal)
# 123-130     8   Status (external)
# 131         1   Off position status
# 132-135     4   Spare (set to 0)

  printf "Internal supply voltage to AIS unit = %.2f volts (V)\n", oct ("0b".substr ($message,88,10)) / 20;
  printf "External voltage from hardware analogue input no. 1 = %.2f volts (V)\n", oct ("0b".substr ($message,98,10)) / 20;
  printf "External voltage from hardware analogue input no. 2 = %.2f volts (V)\n", oct ("0b".substr ($message,108,10)) / 20;
  $x = oct ("0b".substr ($message,118,5)); print "Internal AtoN system status = $x\n";
  $y = $x % 2; printf "  ** Unit status: %d (%s)\n", $y, $aton_status_unit [$y];
  $y = int ($x/2) % 4; printf "  ** Light status: %d (%s)\n", $y, $aton_status_light [$y];
  $y = int ($x/8); printf "  ** Radar beacon (RACON) status: %d (%s)\n", $y, $aton_status_racon [$y];
  printf "Status of external digital inputs = %d\n", oct ("0b".substr ($message,123,8));
  for ($y = 0; $y < 8; $y++)
    {
    $x = oct ("0b".substr ($message,130-$y,1));
    printf "  ** Digital input #%d: %d (%s)\n", $y, $x, $state [$x];
    }
  $x = oct ("0b".substr ($message,131,1)); printf "Off-position status = %d (%s)\n", $x, $offpos [$x];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,132,4));
  return;
  }

# *** Decode_06_235_20 decodes Type 6, DAC 235, FI 20 (Buoy position monitoring) [UK Trinity House] messages ***

sub decode_06_235_20
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($x,$y);

################################################
#  Bits     Len   Description
#  88-291   204   AtoN name
# 292-296     5   UTC day
# 297-301     5   UTC hour
# 302-307     6   UTC minute
# 308-335    28   Longitude
# 336-362    27   Latitude
# 363         1   Off position status
# 364-367     4   Spare (set to 0)

  printf "AtoN name = %s\n", decode_string (substr ($message,88,204));
  default_handler ("UTC day = %d",oct ("0b".substr ($message,292,5)),"",0);
  default_handler ("UTC hour = %02d",oct ("0b".substr ($message,297,5)),"",24);
  default_handler ("UTC minute = %02d",oct ("0b".substr ($message,302,6)),"",60);
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,308,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,336,27)) / 600000),91);
  $x = oct ("0b".substr ($message,363,1)); printf "Off-position status = %d (%s)\n", $x, $offpos [$x];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,364,4));
  return;
  }

# *** Decode_0608_367_22 decodes Type 6 and Type 8 with DAC = 367, FI = 22 (Area Notice - Addressed and Area Notice - Broadcast
#  messages, respectively) [USCG RDC] ***
# This format is based on USCG documentation; it is *similar* to Type 6, DAC=001/FI=23 and Type 8, DAC=001/FI=22 messages
#  but different enough to warrant a different decode routine....

sub decode_0608_367_22
  {
# Parameter list
  my ($addressed_flag,$message,$msg_len) = @_;
  my ($i,$j,$num_shapes,$sa_base,$scale_factor,$shape,$tyui);

# Type 6 - "Geographic Notice - Addressed" (DAC = 367; FI = 22) per USCG
# Type 8 - "Geographic Notice - Broadcast" (DAC = 367; FI = 22) per USCG
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  88-93      6    56-61      6   Message Version (2)
#  94-103    10    62-71     10   Message Linkage ID
# 104-110     7    72-78      7   Notice Description
# 111-114     4    79-82      4   UTC month
# 115-119     5    83-87      5   UTC day
# 120-124     5    88-92      5   UTC hour
# 125-130     6    93-98      6   UTC minute
# 131-148    18    99-116    18   Duration
#                 117         1   Action (set to 0)
# 149-151     3   118-119     2   Spare
# 152-247    96   120-215    96   Sub-area 1
# 248-343    96   216-311    96   Sub-area 2 (if present)
# 344-439    96   312-407    96   Sub-area 3 (if present)
# 440-535    96   408-503    96   Sub-area 4 (if present)
# 536-631    96   504-599    96   Sub-area 5 (if present)
# 632-727    96   600-695    96   Sub-area 6 (if present)
# 728-823    96   696-791    96   Sub-area 7 (if present)
# 824-919    96   792-887    96   Sub-area 8 (if present)
# 920-1015   96   888-983    96   Sub-area 9 (if present)

# NOTE: Longitude and latitude are scaled by multiplying by 600,000
# NOTE: Addressed_flag = 1 for Type 6, = 0 for Type 8

  printf "Message Version = %d (Should be 2)\n", oct ("0b".substr ($message,56+($addressed_flag*32),6));
  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,62+($addressed_flag*32),10)),"",0);
  $tyui = oct ("0b".substr ($message,72+($addressed_flag*32),7)); printf "Notice Description = %d (%s)\n", $tyui, $notice_desc [$tyui];
  default_handler ("UTC month = %d",oct ("0b".substr ($message,79+($addressed_flag*32),4)),"",0);
  default_handler ("UTC day = %d",oct ("0b".substr ($message,83+($addressed_flag*32),5)),"",0);
  default_handler ("UTC hour = %02d",oct ("0b".substr ($message,88+($addressed_flag*32),5)),"",24);
  default_handler ("UTC minute = %02d",oct ("0b".substr ($message,93+($addressed_flag*32),6)),"",60);

  $tyui = oct ("0b".substr ($message,99+($addressed_flag*32),18));
  if ($tyui == 0)
      { print "Duration = 0 (Cancel Area Notice)\n"; }
    else
      {
      if ($tyui == 262143)
          { print "Duration = 262,143 (Undefined [default])\n"; }
        else
          { printf "Duration = %d minutes (%.2f days) until the end of Area Notice\n", $tyui, $tyui/1440; } # 1,440 min/day
      };

  if ($addressed_flag == 1)
      { printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,149,3)); }
    else
      {
      print "Action =";
      if (oct ("0b".substr ($message,117,1)) == 0)
          { print " 0 (Advisement)\n"; }
        else
          { print " 1 (Directive)\n"; };
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,118,2));
      }

# Now handle the sub-areas

  $num_shapes = ($msg_len-(120+($addressed_flag*32)))/96;
  for ($i=1; $i <= $num_shapes; $i++)
    {
    print "--- Sub-area #$i ---\n";

# Start counting at bit value 120 + ($i-1)*96

    $sa_base = 120 + ($addressed_flag*32) + (($i-1) * 96);

# Fields common to all sub-areas

################################################
#  Bits     Len   Description
#   0-2       3   Area Shape

    $shape = oct ("0b".substr ($message,$sa_base,3)); printf "Area Shape = %d (%s)\n", $shape, $area_shape [$shape];

# The following fields are common to shapes 0-4

################################################
#  Bits     Len   Description
#   3-4       2   Scale Factor

    if ($shape <= 4)
      {
      $tyui = oct ("0b".substr ($message,$sa_base+3,2));
      $scale_factor = 10**$tyui; printf "Scale Factor = %d (Factor of %d)\n", $tyui, $scale_factor;
      }

# The following fields are common to shapes 0-2

################################################
#  Bits     Len   Description       FIELD
#   5-32     28   Longitude
#  33-59     27   Latitude
#  60-62      3   Precision

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

    if ($shape <=2)
      {
      long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,$sa_base+5,28)) / 600000),181,
       "Latitude", (signed_bin2dec (substr ($message,$sa_base+33,27)) / 600000),91);
      default_handler ("Precision of latitude/longitude = %d",oct ("0b".substr ($message,$sa_base+60,3))," decimal places",4);
      };

# Now handle the other shapes individually

    SHAPE:
      {

# SHAPE = 0 (Circle/Point)
################################################
#  Bits     Len   Description
#  63-74     12   Radius
#  75-95     21   Not used

      $shape == 0 && do
        {
        $tyui = oct ("0b".substr ($message,$sa_base+63,12));
        if ($tyui == 0)
            { print "Radius = 0 (point)\n"; }
          else
            { printf "Radius of circle = %d (%d m, %.3f nm)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+75,21));
        last SHAPE;
        };

# SHAPE = 1 (Rectangle/Line)
################################################
#  Bits     Len   Description
#  63-70      8   E Dimension
#  71-78      8   N Dimension
#  79-87      9   Orientation
#  88-95      8   Not used

      $shape == 1 && do
        {
        $tyui = oct ("0b".substr ($message,$sa_base+63,8));
        if ($tyui == 0)
            { print "East Dimension = 0 (North-South line; default)\n"; }
          else
            { printf "East Dimension = %d (%d m, %.3f nm from corner point)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        $tyui = oct ("0b".substr ($message,$sa_base+71,8));
        if ($tyui == 0)
            { print "North Dimension = 0 (East-West line; default)\n"; }
          else
            { printf "North Dimension = %d (%d m, %.3f nm from corner point)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        $tyui = oct ("0b".substr ($message,$sa_base+79,9));
        if ($tyui == 0)
            { print "Orientation = 0 (no rotation [default])\n"; }
          else
            {
            if ($tyui >= 1 && $tyui <= 359)
                { printf "Orientation of area = %03d\xb0T\n", $tyui; }
              else
                { print "Orientation = $tyui (invalid; reserved for future use)\n"; }
            };

        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+88,8));
        last SHAPE;
        };

# SHAPE = 2 (Sector)
################################################
#  Bits     Len   Description
#  63-74     12   Radius
#  75-83      9   Left Boundary
#  84-92      9   Right Boundary
#  93-95      3   Not used

      $shape == 2 && do
        {
        $tyui = oct ("0b".substr ($message,$sa_base+63,12));
        if ($tyui == 0)
            { print "Radius = 0 (point)\n"; }
          else
            { printf "Radius of circle = %d (%d m, %.3f nm)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
        $tyui = oct ("0b".substr ($message,$sa_base+75,9));
        if ($tyui == 0)
            { print "Orientation of left boundary = 0 (no rotation [default])\n"; }
          else
            {
            if ($tyui >= 1 && $tyui <= 359)
                { printf "Orientation of left boundary edge = %03d\xb0T\n", $tyui; }
              else
                { print "Orientation of left boundary = $tyui (invalid; reserved for future use)\n"; }
            };
        $tyui = oct ("0b".substr ($message,$sa_base+84,9));
        if ($tyui == 0)
            { print "Orientation of right boundary = 0 (no rotation [default])\n"; }
          else
            {
            if ($tyui >= 1 && $tyui <= 359)
                { printf "Orientation of right boundary edge = %03d\xb0T\n", $tyui; }
              else
                { print "Orientation of right boundary = $tyui (invalid; reserved for future use)\n"; }
            };
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+93,3));

        last SHAPE;
        };

# SHAPES 3 (Polyline) AND 4 (Polygon)
################################################
#  Bits     Len   Description
#   5-14     10   Point 1 Angle
#  15-25     11   Point 1 Distance
#  26-35     10   Point 2 Angle
#  36-46     11   Point 2 Distance
#  47-56     10   Point 3 Angle
#  57-67     11   Point 3 Distance
#  68-77     10   Point 4 Angle
#  78-88     11   Point 4 Distance
#  89-95      7   Not used

      ($shape == 3 || $shape == 4) && do
        {
        for ($j=0; $j<4; $j++)
          {
          printf "Point %d to %d =", $j, $j+1;
          $tyui = oct ("0b".substr ($message,$sa_base+5+($j*21),10));
          if ($tyui < 720)
              { printf " Bearing: %3d (%05.1f\xB0T)", $tyui, $tyui/2; }
            else
              {
              if ($tyui == 720)
                  { print " Bearing: $tyui (not available [default])"; }
                else
                  { print " Bearing: $tyui (reserved for future use)"; }
              };
          $tyui = oct ("0b".substr ($message,$sa_base+15+($j*21),11));
          if ($tyui == 0)
              { print "   Distance: 0 (no point)\n"; }
            else
              { printf "   Distance: %d (%d m, %.3f nm)\n", $tyui, $tyui*$scale_factor, $tyui*$scale_factor/1852.001; } # 1852.001 m/nm
          };
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+89,7));
        last SHAPE;
        };

# SHAPE 5
################################################
#  Bits     Len   Description       FIELD
#   0-2       3   Area Shape (5)      0
#   3-92     90   Text                1   (15 6-bit characters)
#  93-95      3   Not used

      $shape == 5 && do
        {
        printf "Text = %s\n", decode_string (substr ($message,$sa_base+3,90));
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$sa_base+93,3));
        last SHAPE;
        };

     do
        { print "Area Shape = $shape (invalid)\n"; };

      };

    };

  return;
  }

# *** Decode_08_366_20 decodes Type 8, DAC 366, FI 20 (Dam Status) [USCG RDC, V.1a] messages ***

sub decode_08_366_20
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$tyui);
  my @discharge = ("over the dam", "through the associated hydroelectric plant");

################################################
#  Bits     Len   Description
#  56-58      3   Message version (0)
#  59-68     10   Message Linkage ID
#  69-86     18   UN-LOCODE
#  87-110    24   Navigation Unit Identifier
# 111-115     5   UTC hour
# 116-121     6   UTC minute
# 122-134    13   Dam gate total
# 135-146    12   Dam discharge
# 147-148     2   Dam discharge source
# 149-160    12   Hydroelectric plant discharge
# 161-162     2   Hydroelectric plant discharge source
# 163-167     5   Number of gates (N)
# 168-407 0-240   N octets of dam gate info

  printf "Message version: %d (Should be 0)\n", oct ("0b".substr ($message,56,3));
  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,59,10)),"",0);
  printf "Last three characters of UN/LOCODE = %s\n", decode_string (substr ($message,69,18));
  print "  ** See https://unece.org/trade/cefact/unlocode-code-list-country-and-territory for UN/LOCODE list\n";
  printf "Navigation Unit Identifier (GUID) = %s\n", decode_string (substr ($message,87,24));
  print "Timstamp of dam status information:\n";
  default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,111,5)),"",24);
  default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,116,6)),"",60);
  default_handler ("Total amount of dam opening = %.1f",oct ("0b".substr ($message,122,13))/10," m",810.1);

# Special handling for "discharge" information because there are two "maximum" values with special meaning

  for ($i = 0; $i < 2; $i++)
    {
    $tyui = oct ("0b".substr ($message,135+($i*14),12))/10;
    printf "Total discharge %s = %.1f", $discharge [$i], $tyui;
    if ($tyui == 409.1)
        { print " (not applicable [default])\n"; }
      else
        {
        if ($tyui == 409.2)
            { print " (data unknown or unavailable)\n"; }
          else
            { print " cubic meters/sec\n"; }
        }
    $tyui = oct ("0b".substr ($message,147+($i*14),2)); printf "  Discharge information source = %d (%s)\n", $tyui, $discharge_source [$tyui];
    }
    
  $tyui = oct ("0b".substr ($message,163,5)); default_handler ("Number of gates = %d",$tyui,"",31);
  $tyui = min ($tyui,30);
  if ($tyui > 0)
    {
    for ($i = 0; $i < $tyui; $i++)
      {
      printf "  Information for gate %2d: ", $i+1;
      two_max_handler ("**Opening width = %.1f", oct ("0b".substr ($message,168+($i*8),8))/10," m",25.1,25.2);
      }
    }
  return;
  }

# *** Decode_06_367_16 decodes Type 6, DAC 367, FI 16 (Passenger and Crew Count – Addressed) [USCG RDC, V.1a] messages ***

sub decode_06_367_16
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  88-93      6   Message version (1)
#  94-103    10   Message Linkage ID
# 104-108     5   UTC day
# 109-113     5   UTC hour
# 114-119     6   UTC minute
# 120-132    13   Total crew members
# 133-147    15   Total passengers
# 148-151     4   Additional vessel type information

  printf "Message version: %d (Should be 1; 0 = test message)\n", oct ("0b".substr ($message,88,6));
  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,94,10)),"",0);
  print "Timstamp of currency of POB data:\n";
  default_handler ("  UTC day = %d",oct ("0b".substr ($message,104,5)),"",0);
  default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,109,5)),"",24);
  default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,114,6)),"",60);
  two_max_handler ("Total number of crew members = %d", oct ("0b".substr ($message,120,13)),"",8191,0);
  two_max_handler ("Total number of passengers = %d", oct ("0b".substr ($message,133,15)),"",32501,32502);
  $tyui = oct ("0b".substr ($message,148,4)); printf "Additional vessel type Information: %d (%s)\n", $tyui, $add_vessel_type_info [$tyui];
  return;
  }

# *** Decode_08_367_23 decodes Type 8, DAC 367, FI 23 (Satellite Ship Weather) [USCG RDC, V.0] messages ***

sub decode_08_367_23
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  56-58      3   Message version (0)
#  59-63      5   UTC day
#  64-68      5   UTC hour
#  69-74      6   UTC minute
#  75-99     25   Longitude
# 100-123    24   Latitude
# 124-132     9   Air pressure
# 133-143    11   Air temperature
# 144-150     7   Wind speed
# 151-157     7   Wind gust
# 158-166     9   Wind direction
# 167         1   Spare

  printf "Message version: %d (Should be 0)\n", oct ("0b".substr ($message,56,3));
  print "Timstamp of weather data:\n";
  default_handler ("  UTC day = %02d",oct ("0b".substr ($message,59,5)),"",0);
  default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,64,5)),"",24);
  default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,69,6)),"",60);
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,75,25)) / 60000),181, "Latitude", (signed_bin2dec (substr ($message,100,24)) / 60000),91);
  air_press_handler ("Air pressure = ", oct ("0b".substr ($message,124,9)), 403);
  default_handler ("Air temperature = %.1f", signed_bin2dec (substr ($message,133,11))/10, "\xB0C", -102.4);
  two_max_handler ("10-minute average wind speed = %d", (oct ("0b".substr ($message,144,7)))," kn",121,122);
  two_max_handler ("10-minute maximum wind (gust) speed = %d", (oct ("0b".substr ($message,151,7)))," kn",121,122);
  default_handler ("Wind direction = %03d", oct ("0b".substr ($message,158,9)), "\xB0",360);
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,167,1));
  return;
  }

# *** Decode_08_367_24 decodes Type 8, DAC 367, FI 24 (Satellite Ship Weather Small) [USCG RDC, V.0] messages ***

sub decode_08_367_24
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  56-58      3   Message version (0)
#  59-63      5   UTC hour
#  64-69      6   UTC minute
#  70-94     25   Longitude
#  95-118    24   Latitude
# 119-127     9   Air pressure

  printf "Message version: %d (Should be 0)\n", oct ("0b".substr ($message,56,3));
  print "Timstamp of weather data:\n";
  default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,59,5)),"",24);
  default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,64,6)),"",60);
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,70,25)) / 60000),181, "Latitude", (signed_bin2dec (substr ($message,95,24)) / 60000),91);
  air_press_handler ("Air pressure = ", oct ("0b".substr ($message,119,9)), 403);
  return;
  }

# *** Decode_08_367_25 decodes Type 8, DAC 367, FI 25 (Satellite Ship Weather Tiny) [USCG RDC, V.0] messages ***

sub decode_08_367_25
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  56-58      3   Message version (0)
#  59-63      5   UTC hour
#  64-69      6   UTC minute
#  70-78      9   Air pressure
#  79-85      7   Wind speed
#  86-94      9   Wind direction
#  95         1   Spare

  printf "Message version: %d (Should be 0)\n", oct ("0b".substr ($message,56,3));
  print "Timstamp of weather data:\n";
  default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,59,5)),"",24);
  default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,64,6)),"",60);
  air_press_handler ("Air pressure = ", oct ("0b".substr ($message,70,9)), 403);
  two_max_handler ("10-minute average wind speed = %d", (oct ("0b".substr ($message,79,7)))," kn",121,122);
  default_handler ("Wind direction = %03d", oct ("0b".substr ($message,86,9)), "\xB0", 360);
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,95,1));
  return;
  }

# *** Decode_08_367_26 decodes Type 8, DAC 367, FI 26 (GPS Jamming and Spoofing Report) [USCG RDC, V.0] messages ***

sub decode_08_367_26
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  56-58      3   Message version (0)
#  59-63      5   UTC day
#  64-68      5   UTC hour
#  69-74      6   UTC minute
#  75-99     25   Longitude
# 100-123    24   Latitude
# 124         1   Type of event
# 125         1   Event status
# 126-127     2   Spare

  printf "Message version: %d (Should be 0)\n", oct ("0b".substr ($message,56,3));
  print "Timstamp of GPS event:\n";
  default_handler ("  UTC day = %02d",oct ("0b".substr ($message,59,5)),"",0);
  default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,64,5)),"",24);
  default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,69,6)),"",60);
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,75,25)) / 60000),181, "Latitude", (signed_bin2dec (substr ($message,100,24)) / 60000),91);
  $tyui = oct ("0b".substr ($message,124,1)); printf "Type of event = %d (%s)\n", $tyui, $gps_event_type [$tyui];
  $tyui = oct ("0b".substr ($message,125,1)); printf "Event status = %d (%s)\n", $tyui, $gps_event_status [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,126,2));
  return;
  }

# *** Decode_08_367_33 decodes Type 8, DAC 367, FI 33 (Environmental Management) [USCG RDC, V.3a] messages ***
# This is a very complicated message to decode because there are a dozen different message formats...

sub decode_08_367_33
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$j,$k,$lbl,$num_reports,$report_type,$tyui);
  my @vector = ("North","East","Up");

################################################
#  Bits     Len    Bits     Len   Description
#  38-39      2   Spare
#  40-49     10   DAC
#  50-55      6   FI
#  56-167   112   Sensor Report 1
# 168-279   112   Sensor Report 2 (optional)
# 280-391   112   Sensor Report 3 (optional)
# 392-503   112   Sensor Report 4 (optional)
# 504-615   112   Sensor Report 5 (optional)
# 616-727   112   Sensor Report 6 (optional)
# 728-839   112   Sensor Report 7 (optional)
# 840-951   112   Sensor Report 8 (optional)

# Sensor Report
#-----------------------------------------------
#  Bits     Len   Description
#   0-3       4   Report Type      <-- starting position is (56 + increments of 112)
#   4-8       5   UTC day
#   9-13      5   UTC hour
#  14-19      6   UTC minute
#  20-26      7   Site ID
#  27-111    85   Sensor Data      <-- starting position is (83 + increments of 112)

  $num_reports = int (($msg_len - 56) / 112);
  for ($i = 0; $i < $num_reports; $i++)
    {
    printf "SENSOR REPORT %d\n", $i+1;
    $report_type = oct ("0b".substr ($message,56+$i*112,4)); printf "  Sensor report type = %d (%s)\n", $report_type, $em_sensor_report_type [$report_type];
    print "  Time of data:\n";
    default_handler ("    UTC day = %02d",oct ("0b".substr ($message,60+$i*112,5)),"",0);
    default_handler ("    UTC hour = %02d",oct ("0b".substr ($message,65+$i*112,5)),"",24);
    default_handler ("    UTC minute = %02d",oct ("0b".substr ($message,70+$i*112,6)),"",60);
    printf "  Site identifier = %d\n", oct ("0b".substr ($message,76+$i*112,7));
    
    SENSOR_REPORT_HANDLER:
      {

# Sensor Type 0 - Site Location
################################################
#  Bits     Len   Description
#   0-5       6   Message Version 
#   6-33     28   Longitude
#  34-60     27   Latitude
#  61-63      3   Precision
#  64-75     12   Altitude
#  76-79      4   Sensor Owner
#  80-82      3   Data Timeout
#  83-84      2   Spare

      $report_type == 0 && do
        {
        printf "  Message version number = %d (should be 3)\n", oct ("0b".substr ($message,83+$i*112,6));
        long_lat_handler ("  Longitude", (signed_bin2dec (substr ($message,89+$i*112,28)) / 600000),181, "  Latitude", (signed_bin2dec (substr ($message,117+$i*112,27)) / 600000),91);
        default_handler ("  Precision of latitude/longitude = %d",oct ("0b".substr ($message,144+$i*112,3))," decimal places",5);
        two_max_handler ("  Altitude of sensor relative to MSL = %.1f", (signed_bin2dec (substr ($message,147+$i*112,12))/10)," m",200.1,200.2);
        $tyui = oct ("0b".substr ($message,159+$i*112,4)); printf "  Sensor owner = %d (%s)\n", $tyui, $em_sensor_owner [$tyui];
        $tyui = oct ("0b".substr ($message,163+$i*112,3)); printf "  Data timeout = %d (%s)\n", $tyui, $em_data_timeout [$tyui];
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,166+$i*112,2));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 1 - Station ID
################################################
#  Bits     Len   Description
#   0-83     84   Name
#  84         1   Spare

      $report_type == 1 && do
        {
        print "  Station ID = ", decode_string (substr ($message,83+$i*112,84)) =~ s/@//gr, "\n";
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,167+$i*112,1));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 2 - Wind Report
################################################
#  Bits     Len   Description
#   0-6       7   Wind Speed
#   7-13      7   Wind Gust
#  14-22      9   Wind Direction
#  23-31      9   Wind Gust Direction
#  32-34      3   Sensor Data Description
#  35-41      7   Forecast Wind Speed
#  48-48      7   Forecast Wind Gust
#  49-57      9   Forecast Wind Direction
#  58-62      5   UTC Date
#  63-67      5   UTC Hour
#  68-73      6   UTC Minute
#  74-81      8   Duration of Forecast
#  82-84      3   Spare

      $report_type == 2 && do
        {
        two_max_handler ("  10-minute average wind speed = %d", (oct ("0b".substr ($message,83+$i*112,7)))," kn",121,122);
        two_max_handler ("  10-minute maximum wind (gust) speed = %d", (oct ("0b".substr ($message,90+$i*112,7)))," kn",121,122);
        default_handler ("  Wind direction = %03d", oct ("0b".substr ($message,97+$i*112,9)), "\xB0", 360);
        default_handler ("  Wind gust direction = %03d", oct ("0b".substr ($message,106+$i*112,9)), "\xB0", 360);
        $tyui = oct ("0b".substr ($message,115+$i*112,3)); printf "  Wind sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        two_max_handler ("  Predicted average wind speed = %d", (oct ("0b".substr ($message,118+$i*112,7)))," kn",121,122);
        two_max_handler ("  Predicted maximum wind (gust) speed = %d", (oct ("0b".substr ($message,125+$i*112,7)))," kn",121,122);
        default_handler ("  Predicted wind direction = %03d", oct ("0b".substr ($message,132+$i*112,9)), "\xB0", 360);
        print "  Valid time of forecast:\n";
        default_handler ("    UTC day = %d",oct ("0b".substr ($message,141+$i*112,5)),"",0);
        default_handler ("    UTC hour = %02d",oct ("0b".substr ($message,146+$i*112,5)),"",24);
        default_handler ("    UTC minute = %02d",oct ("0b".substr ($message,151+$i*112,6)),"",60);
        $tyui = oct ("0b".substr ($message,157+$i*112,8));
        if ($tyui == 0)
            { print "  Duration of forecast = 0 (Cancel forecast)\n"; }
          else
            { printf "  Duration of forecast = %d minutes (H:MM %01d:%02d)\n", $tyui, int ($tyui/60), $tyui % 60; };
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,165+$i*112,3));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 3	- Tide/River Stage (Water Level) Report
################################################
#  Bits     Len   Description
#   0         1   Water Level Type
#   1-16     16   Water Level
#  17-18      2   Water Level Trend
#  19-23      5   Vertical Reference Datum 
#  24-26      3   Sensor Data Description
#  27         1   Forecast Water Level Type
#  28-43     16   Forecast Water Level
#  44-48      5   UTC Date
#  49-53      5   UTC Hour
#  54-59      6   UTC Minute
#  60-67      8   Duration of Forecast
#  68-84     17   Spare

      $report_type == 3 && do
        {
        $tyui = oct ("0b".substr ($message,83+$i*112,1)); printf "  Water level type = %d (%s)\n", $tyui, $em_water_level_type [$tyui];
        two_max_handler ("  Water level = %.2f", (signed_bin2dec (substr ($message,84+$i*112,16))/100)," m",327.67,-327.68);
        $tyui = oct ("0b".substr ($message,100+$i*112,2)); printf "  Water level trend = %d (%s)\n", $tyui, $em_water_level_trend [$tyui];
        $tyui = oct ("0b".substr ($message,102+$i*112,5)); printf "  Vertical reference datum = %d (%s)\n", $tyui, $em_vert_ref_datum [$tyui];
        $tyui = oct ("0b".substr ($message,107+$i*112,3)); printf "  Water level sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        $tyui = oct ("0b".substr ($message,110+$i*112,1)); printf "  Forecast water level type = %d (%s)\n", $tyui, $em_water_level_type [$tyui];
        two_max_handler ("  Forecast water level = %.2f", (signed_bin2dec (substr ($message,111+$i*112,16))/100)," m",327.67,-327.68);
        print "  Valid time of forecast:\n";
        default_handler ("    UTC day = %d",oct ("0b".substr ($message,127+$i*112,5)),"",0);
        default_handler ("    UTC hour = %02d",oct ("0b".substr ($message,132+$i*112,5)),"",24);
        default_handler ("    UTC minute = %02d",oct ("0b".substr ($message,137+$i*112,6)),"",60);
        $tyui = oct ("0b".substr ($message,143+$i*112,8));
        if ($tyui == 0)
            { print "  Duration of forecast = 0 (Cancel forecast)\n"; }
          else
            { printf "  Duration of forecast = %d minutes (H:MM %01d:%02d)\n", $tyui, int ($tyui/60), $tyui % 60; };
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,151+$i*112,17));
        last SENSOR_REPORT_HANDLER;
        };
     
# Sensor Type 4	- Two-Dimensional (x & y) Vertical Current Profile Report
################################################
#  Bits     Len   Description
#   0-7       8   Current 1 Speed
#   8-16      9   Current 1 Direction
#  17-25      9   Current 1 Measuring Level
#  26-33      8   Current 2 Speed
#  34-42      9   Current 2 Direction
#  43-51      9   Current 2 Measuring Level
#  52-59      8   Current 3 Speed
#  60-68      9   Current 3 Direction
#  69-77      9   Current 3 Measuring Level
#  78-80      3   Sensor Data Description
#  81-84      4   Spare

      $report_type == 4 && do
        {
        for ($j = 1; $j <= 3; $j++)
          {
          two_max_handler ("  Speed of current $j = %.1f", oct ("0b".substr ($message,83+$i*112+($j-1)*26,8))/10," kn",24.6,24.7);
          default_handler ("  Direction of current $j = %03d",oct ("0b".substr ($message,91+$i*112+($j-1)*26,9)),"\xB0", 360);
          two_max_handler ("  Measurement level of current $j below water surface = %d", oct ("0b".substr ($message,100+$i*112+($j-1)*26,9))," m",361,362);
          };
        $tyui = oct ("0b".substr ($message,161+$i*112,3)); printf "  Current sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,164+$i*112,4));
        last SENSOR_REPORT_HANDLER;
        };
        
# Sensor Type 5	- Three-Dimensional (x, y, & z) Vertical Current Profile Report
################################################
#  Bits     Len   Description
#   0-8       9   Current 1: vector component North
#   9-17      9   Current 1: vector component East
#  18-26      9   Current 1: vector component Up
#  27-35      9   Current 1 Measuring Level
#  36-44      9   Current 2: vector component North
#  45-53      9   Current 2: vector component East
#  54-62      9   Current 2: vector component Up
#  63-71      9   Current 2 Measuring Level
#  72-74      3   Sensor Data Description
#  75-84     10   Spare

      $report_type == 5 && do
        {
        for ($j = 1; $j <= 2; $j++)
          {
          print "  --- Information for current $j ---\n";
          for ($k=0; $k<3; $k++)
            {
            $lbl = sprintf ("    Speed of %s component = ", $vector[$k]) . "%.1f";
            two_max_handler ($lbl, (signed_bin2dec (substr ($message,83+$i*112+($j-1)*36+$k*9,9))/10)," kn",-25.1,-25.6);
            };
          two_max_handler ("    Measurement level below water surface = %d", oct ("0b".substr ($message,110+$i*112+($j-1)*36,9))," m",361,362);
          };
        $tyui = oct ("0b".substr ($message,155+$i*112,3)); printf "  Current sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,158+$i*112,10));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 6	- Horizontal Current Profile
################################################
#  Bits     Len   Description
#   0-8       9   Current Reading Bearing
#   9-17      9   Current 1 Distance
#  18-25      8   Current 1 Speed
#  26-34      9   Current 1 Direction
#  35-43      9   Current 1 Measuring Level
#  44-52      9   Current 2 Distance
#  53-60      8   Current 2 Speed
#  61-69      9   Current 2 Direction
#  70-78      9   Current 2 Measuring Level
#  79-81      3   Sensor Data Description
#  82-84      3   Spare

      $report_type == 6 && do
        {
        default_handler ("  Bearing of current readings from sensor position = %03d",oct ("0b".substr ($message,83+$i*112,9)),"\xB0", 360);
        for ($j = 1; $j <= 2; $j++)
          {
          print "  --- Information for current $j ---\n";
          two_max_handler ("    Distance of reading from sensor = %.1f", oct ("0b".substr ($message,92+$i*112+($j-1)*35,9))," m",361,362);
          two_max_handler ("    Speed = %.1f", oct ("0b".substr ($message,101+$i*112+($j-1)*35,8))/10," kn",24.6,24.7);
          default_handler ("    Direction = %03d",oct ("0b".substr ($message,109+$i*112+($j-1)*35,9)),"\xB0", 360);
          two_max_handler ("    Measurement level below water surface = %d", oct ("0b".substr ($message,118+$i*112+($j-1)*35,9))," m",361,362);
          };
        $tyui = oct ("0b".substr ($message,162+$i*112,3)); printf "  Current sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,165+$i*112,3));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 7	- Sea State
################################################
#  Bits     Len   Description
#   0-7       8   Swell Height
#   8-13      6   Swell Period
#  14-22      9   Swell Direction
#  23-26      4   Sea State 
#  27-29      3   Swell Sensor Data Description
#  30-39     10   Water Temperature
#  40-46      7   Water Temperature Depth
#  47-49      3   Water Temperature Sensor Data Description
#  50-57      8   Significant Wave Height
#  58-63      6   Wave Period
#  64-72      9   Wave Direction
#  73-75      3   Wave Sensor Data Description
#  76-84      9   Salinity

      $report_type == 7 && do
        {
        two_max_handler ("  Swell height = %.1f", oct ("0b".substr ($message,83+$i*112,8))/10," m",24.6,24.7);
        default_handler ("  Swell period = %d",oct ("0b".substr ($message,91+$i*112,6))," sec", 61);
        default_handler ("  Swell direction = %d",oct ("0b".substr ($message,97+$i*112,9)),"\xB0", 360);
        $tyui = oct ("0b".substr ($message,106+$i*112,4)); printf "  Sea state (Beaufort scale) = %d (%s)\n", $tyui, $Beaufort [$tyui];
        $tyui = oct ("0b".substr ($message,110+$i*112,3)); printf "  Swell sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        default_handler ("  Water temperature = %.1f",(oct ("0b".substr ($message,113+$i*112,10))/10)-10,"\xB0C", 60.1);
        two_max_handler ("  Water temperature sensor depth below water surface = %.1f", oct ("0b".substr ($message,123+$i*112,7))/10," m",12.1,12.2);
        $tyui = oct ("0b".substr ($message,130+$i*112,3)); printf "  Water temperature sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        two_max_handler ("  Significant wave height = %.1f", oct ("0b".substr ($message,133+$i*112,8))/10," m",24.6,24.7);
        default_handler ("  Wave period = %d",oct ("0b".substr ($message,141+$i*112,6))," sec", 61);
        default_handler ("  Wave direction = %d",oct ("0b".substr ($message,147+$i*112,9)),"\xB0", 360);
        $tyui = oct ("0b".substr ($message,156+$i*112,3)); printf "  Wave sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        three_max_handler ("  Salinity = %.1f", oct ("0b".substr ($message,159+$i*112,9))/10,"\x{2030}",50.1,50.2,50.3,2);
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 8	- Salinity
################################################
#  Bits     Len   Description
#   0-9      10   Water Temperature
#  11-19     10   Conductivity
#  20-35     16   Water Pressure
#  36-44      9   Salinity
#  45-46      2   Salinity Type
#  47-49      3   Salinity Sensor Data Description
#  50-84     35   Spare

      $report_type == 8 && do
        {
        default_handler ("  Water temperature = %.1f",(oct ("0b".substr ($message,83+$i*112,10))/10)-10,"\xB0C", 60.1);
        three_max_handler ("  Conductivity = %.2f", oct ("0b".substr ($message,93+$i*112,10))/100," Siemens/meter (S/m)",7.01,7.02,7.03,3);
        three_max_handler ("  Water pressure = %.1f", oct ("0b".substr ($message,103+$i*112,16))/10," decibars (10 ATM)",6000.1,6000.2,6000.3,3);
        three_max_handler ("  Salinity = %.1f", oct ("0b".substr ($message,119+$i*112,9))/10,"\x{2030}",50.1,50.2,50.3,3);
        $tyui = oct ("0b".substr ($message,128+$i*112,2)); printf "  Salinity type = %d (%s)\n", $tyui, $em_salinity_type [$tyui];
        $tyui = oct ("0b".substr ($message,130+$i*112,3)); printf "  Wave sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,133+$i*112,35));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 9	- Weather
################################################
#  Bits     Len   Description
#   0-10     11   Air Temperature
#  11-13      3   Air Temperature Sensor Data Description
#  14-15      2   Precipitation Type
#  16-23      8   Horizontal Visibility
#  24-33     10   Dew Point
#  34-36      3   Dew Point Sensor Data Description
#  37-45      9   Air Pressure
#  46-47      2   Air Pressure Trend
#  48-50      3   Air Pressure Sensor Data Description
#  51-59      9   Salinity
#  60-84     25   Spare

      $report_type == 9 && do
        {
        default_handler ("  Air temperature = %.1f", signed_bin2dec (substr ($message,83+$i*112,11))/10, "\xB0C", -102.4);
        $tyui = oct ("0b".substr ($message,94+$i*112,3)); printf "  Air temperature sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        $tyui = oct ("0b".substr ($message,97+$i*112,2)); printf "  WMO precipitation type = %d (%s)\n", $tyui, $em_wmo [$tyui];

# Special handling for horizontal visibility
        $tyui = oct ("0b".substr ($message,99+$i*112,8))/10;
        printf "  Horizontal visibility = %.1f", $tyui;
        HOR_VIS:
          {
          $tyui <= 24 && do
            {
            print " nm\n";
            last HOR_VIS;
            };
          $tyui == 24.1 && do
            {
            print " nm or greater\n";
            last HOR_VIS;
            };
          $tyui == 24.2 && do
            {
            print " (Data unavailable)\n";
            last HOR_VIS;
            };
          $tyui == 24.3 && do
            {
            print " (Sensor unavailable)\n";
            last HOR_VIS;
            }
          };

# Special handling for dew point
        $tyui = (oct ("0b".substr ($message,107+$i*112,10))/10)-20;
        printf "  Dew point = %.1f", $tyui;
        HOR_VIS:
          {
          $tyui <= 50 && do
            {
            print "\xB0C\n";
            last HOR_VIS;
            };
          $tyui == 50.1 && do
            {
            print " (Data unavailable [default])\n";
            last HOR_VIS;
            };
          $tyui == 50.2 && do
            {
            print " (less than -20\xB0C)\n";
            last HOR_VIS;
            };
          $tyui = 50.3 && do
            {
            print " (greater than 50\xB0C)\n";
            last HOR_VIS;
            }
          };

        $tyui = oct ("0b".substr ($message,117+$i*112,3)); printf "  Dew point sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        air_press_handler ("  Air pressure = ", oct ("0b".substr ($message,120+$i*112,9)), 403);
        $tyui = oct ("0b".substr ($message,129+$i*112,2)); printf "  Air pressure trend = %d (%s)\n", $tyui, $tendency [$tyui];
        $tyui = oct ("0b".substr ($message,131+$i*112,3)); printf "  Air pressure sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        three_max_handler ("  Salinity = %.1f", oct ("0b".substr ($message,134+$i*112,9))/10,"\x{2030}",50.1,50.2,50.3,3);
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,143+$i*112,25));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 10 - Air Gap
################################################
#  Bits     Len   Description
#   0-12     13   Air Draught
#  13-25     13   Air Gap
#  26-27      2   Air Gap Trend
#  28-40     13   Predicted Air Gap
#  41-45      5   UTC Date
#  46-50      5   UTC Hour
#  51-56      6   UTC Minute
#  57-59      3   Air Gap Sensor Data Description
#  68-84     25   Spare

      $report_type == 10 && do
        {
        two_max_handler ("  Air draught (waterline to highest point of ship) = %.1f", oct ("0b".substr ($message,83+$i*112,13))/10," m",819.1,0);
        two_max_handler ("  Air gap (surface of the water to the sensor) = %.1f", oct ("0b".substr ($message,96+$i*112,13))/10," m",819.1,0);
        $tyui = oct ("0b".substr ($message,109+$i*112,2)); printf "  Air gap trend = %d (%s)\n", $tyui, $tendency [$tyui];
        two_max_handler ("  Predicted air gap (surface of the water to the sensor) = %.1f", oct ("0b".substr ($message,111+$i*112,13))/10," m",819.1,0);
        print "  Time of predicted air gap:\n";
        default_handler ("    UTC day = %d",oct ("0b".substr ($message,124+$i*112,5)),"",0);
        default_handler ("    UTC hour = %02d",oct ("0b".substr ($message,129+$i*112,5)),"",24);
        default_handler ("    UTC minute = %02d",oct ("0b".substr ($message,134+$i*112,6)),"",60);
        $tyui = oct ("0b".substr ($message,140+$i*112,3)); printf "  Air gap sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,143+$i*112,25));
        last SENSOR_REPORT_HANDLER;
        };

# Sensor Type 11	- Wind report v2
################################################
#  Bits     Len   Description
#   0-6       7   Wind Speed
#   7-13      7   Wind Gust
#  14-22      9   Wind Direction
#  23-28      6   Averaging Time
#  29-31      3   Sensor Data Description
#  32-38      7   Forecast Wind Speed
#  39-45      7   Forecast Wind Gust
#  46-54      9   Forecast Wind Direction
#  55-59      5   UTC Hour
#  60-65      6   UTC Minute
#  66-73      8   Duration of Forecast
#  74-84     11   Spare

      $report_type == 11 && do
        {
        two_max_handler ("  N-minute average wind speed = %d", (oct ("0b".substr ($message,83+$i*112,7)))," kn",121,122);
        two_max_handler ("  N-minute maximum wind (gust) speed = %d", (oct ("0b".substr ($message,90+$i*112,7)))," kn",121,122);
        default_handler ("  N-minute average wind direction = %03d", oct ("0b".substr ($message,97+$i*112,9)), "\xB0", 360);
        two_max_handler ("  Averaging time (N) = %d", oct ("0b".substr ($message,106+$i*112,6)), " minutes", 61,0);
        $tyui = oct ("0b".substr ($message,112+$i*112,3)); printf "  Wind sensor data description = %d (%s)\n", $tyui, $em_sensor_data_desc [$tyui];
        two_max_handler ("  Predicted average wind speed = %d", (oct ("0b".substr ($message,115+$i*112,7)))," kn",121,122);
        two_max_handler ("  Predicted maximum wind (gust) speed = %d", (oct ("0b".substr ($message,122+$i*112,7)))," kn",121,122);
        default_handler ("  Predicted wind direction = %03d", oct ("0b".substr ($message,129+$i*112,9)), "\xB0", 360);
        print "  Valid time of forecast:\n";
        default_handler ("    UTC hour = %02d",oct ("0b".substr ($message,138+$i*112,5)),"",24);
        default_handler ("    UTC minute = %02d",oct ("0b".substr ($message,143+$i*112,6)),"",60);
        $tyui = oct ("0b".substr ($message,149+$i*112,8));
        if ($tyui == 0)
            { print "  Duration of forecast = 0 (Cancel forecast)\n"; }
          else
            { printf "  Duration of forecast = %d minutes (H:MM %01d:%02d)\n", $tyui, int ($tyui/60), $tyui % 60; };
        printf "  Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,157+$i*112,11));
        last SENSOR_REPORT_HANDLER;
        };
        
      };
      
    };
    
  return;
  }

# *** Decode_0608_367_35 decodes Type 6 and Type 8, DAC 367, FI 35 (Waterways management - Addressed//Broadcast) [USCG RDC, V.2c] messages ***

sub decode_0608_367_35
  {
# Parameter list
  my ($addressed_flag,$message,$msg_len) = @_;
  my ($i,$num_entries,$subtype,$tyui);

# Type 6 - "Text using 6-bit ASCII - Addressed" (DAC = 1; FI = 0)
# Type 8 - "Text using 6-bit ASCII - Broadcast" (DAC = 1; FI = 0)
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
#  88-97     10    56-65     10   Message Linkage ID
#  98-101     4    66-69      4   Waterways Management message type
# 102-103     2    70-71      2   WM message subtype

# Parse common WM message header

  default_handler ("Message Linkage ID = %d",oct ("0b".substr ($message,56+($addressed_flag*32),10)),"",0);
  $tyui = oct ("0b".substr ($message,66+($addressed_flag*32),4)); printf "Wayerways management message type = %d (%s)\n", $tyui, $wm_type [$tyui];
  $subtype = oct ("0b".substr ($message,70+($addressed_flag*32),2)); printf "Wayerways management message subtype = %d (%s)\n", $subtype, $wm_subtype [$subtype];

  SUBTYPE_PARSER:
    {

# Subtype 0 - Vessel Information
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
# 104-159    56    72-127    56   Vessel #1 information
# 160-215    56   128-183    56   Vessel #2 information (optional)
# 216-271    56   184-239    56   Vessel #3 information (optional)
# 272-327    56   240-295    56   Vessel #4 information (optional)
# 328-383    56   296-351    56   Vessel #5 information (optional)
# 384-439    56   352-407    56   Vessel #6 information (optional)
# 440-495    56   408-463    56   Vessel #7 information (optional)

# Vessel Information
#-----------------------------------------------
#  Bits     Len   Description
#   0-29     30   MMSI
#  30-34      5   UTC day
#  35-39      5   UTC hour
#  40-45      6   UTC minute
#  46-47      2   Vessel status
#  48-50      3   Chamber
#  51         1   Direction
#  52-55      4   Lockage Type

    ($subtype == 0) && do
      {
      $num_entries = ($msg_len - 72 - ($addressed_flag*32)) / 56;
      for ($i = 1; $i <= $num_entries; $i++)
        {
        print "Information for vessel #$i\n";
        $tyui = oct ("0b".substr ($message,72 + ($addressed_flag*32) + (($i-1)*56),30));
        if ($tyui > 0 && $tyui < 1000000000) { print "  "; mmsi_handler ($tyui); }
          else { mmsi_367_35 ($tyui); }
        default_handler ("  UTC day = %d",oct ("0b".substr ($message,102 + ($addressed_flag*32) + (($i-1)*56),5)),"",0);
        default_handler ("  UTC hour = %02d",oct ("0b".substr ($message,107 + ($addressed_flag*32) + (($i-1)*56),5)),"",24);
        default_handler ("  UTC minute = %02d",oct ("0b".substr ($message,112 + ($addressed_flag*32) + (($i-1)*56),6)),"",60);
        $tyui = oct ("0b".substr ($message,118 + ($addressed_flag*32) + (($i-1)*56),2)); printf "  Vessel status = %d (%s)\n", $tyui, $vessel_status [$tyui];
        default_handler ("  Chamber (as defined by USACE LPMS) = %d",oct ("0b".substr ($message,120 + ($addressed_flag*32) + (($i-1)*56),3)),"",0);
        $tyui = oct ("0b".substr ($message,123 + ($addressed_flag*32) + (($i-1)*56),1));
        printf "  Direction of the ship (defined by river flow or channel marking) = %d (%s)\n", $tyui, $direction [$tyui];
        $tyui = oct ("0b".substr ($message,124 + ($addressed_flag*32) + (($i-1)*56),4));
        printf "  Lockage type (see USACE ENG FORM 3102B) = %d (%s)\n", $tyui, $lockage_type [$tyui];
        }
      last SUBTYPE_PARSER;
      };

# Subtype 1 - WM	Place	Identification
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
# 104-109     6    72-77      6   Message version
# 110-137    28    78-105    28   Longitude
# 138-164    27   106-132    27   Latitude
# 165-167     3   133-135     3   Precision
# 168-176     9   136-144     9   Orientation
# 177-386   210   145-354   210   Description (up to 35 characters)
# 387-391    <8   355-359    <8   Spare (octet-alignment)

    ($subtype == 1) && do
      {
      printf "  Message version (should be 2; 0 = test message) = %d\n", oct ("0b".substr ($message,72 + ($addressed_flag*32),6));
      long_lat_handler ("  Longitude", (signed_bin2dec (substr ($message,78 + ($addressed_flag*32),28)) / 600000),181, "  Latitude", (signed_bin2dec (substr ($message,106 + ($addressed_flag*32),27)) / 600000),91);
      default_handler ("  Precision of latitude/longitude = %d",oct ("0b".substr ($message,133 + ($addressed_flag*32),3))," decimal places",4);
      default_handler ("  Orientation = %03d", oct ("0b".substr ($message,136 + ($addressed_flag*32),9)), "\xB0", 360);
      $tyui = int (($msg_len-(145 + ($addressed_flag*32))) / 6);
      printf "  Description = %s\n", decode_string (substr ($message,145 + ($addressed_flag*32),$tyui*6));
      last SUBTYPE_PARSER;
      };

# Subtype 2 - Vessel Name
################################################
#  == Type 6 ==    == Type 8 ==
#  Bits     Len    Bits     Len   Description
# 104-113    10    72-81     10   Vessel number
# 114-116     3    82-84      3   Descriptor type
# 117-326   210    85-294   210   Vessel name (up to 35 characters)
# 327        <8   295        <8   Spare (octet-alignment)

    ($subtype == 2) && do
      {
      $tyui = oct ("0b".substr ($message,72 + ($addressed_flag*32),10)); print "  Vessel number = $tyui";
      if ($tyui <= 2)
          { print " (reserved)\n"; }
        else
          { printf " (corresponds to MMSI 100000%04d in companion subtype 0 message)\n", $tyui; }
      $tyui = oct ("0b".substr ($message,82 + ($addressed_flag*32),3)); printf "  Descriptor type = %d (%s)\n", $tyui, $wm_descriptor [$tyui];
      $tyui = int (($msg_len-(85 + ($addressed_flag*32))) / 6);
      printf "  Vessel name = %s\n", decode_string (substr ($message,85 + ($addressed_flag*32),$tyui*6));
      last SUBTYPE_PARSER;
      };
    };

  return;
  }

######## END OF TYPE 6 AND 8 DECODE ROUTINES

# *** Decode_0713 decodes Type 7, Binary Acknowledge and Type 13, Safety-Related Acknowledgement messages ***

sub decode_0713
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$num);

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-69     30   Target MMSI #1
#  70-71      2   Sequence number for MMSI #1
#  72-101    30   Target MMSI #2 (if present)
# 101-103     2   Sequence number for MMSI #2
# 104-133    30   Target MMSI #3 (if present)
# 134-135     2   Sequence number for MMSI #3
# 136-165    30   Target MMSI #3 (if present)
# 166-167     2   Sequence number for MMSI #4

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));

  $num = int (($msg_len-40)/32);
  for ($i=1; $i<=$num; $i++)
    {
    print "Target #$i "; mmsi_handler (oct ("0b".substr ($message,8+($i*32),30)));
    printf "Sequence number for MMSI #$i = %d\n", oct ("0b".substr ($message,38+($i*32),2));
    };

  return;
  }

# *** Decode_09 decodes Type 9, Standard SAR Aircraft Position Report messages ***

sub decode_09
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-49     12   Altitude
#  50-59     10   Speed Over Ground (SOG)
#  60         1   Position Accuracy
#  61-88     28   Longitude
#  89-115    27   Latitude
# 116-127    12   Course Over Ground (COG)
# 128-133     6   Time Stamp
# 134-141     8   Regional Reserved
# 142         1   DTE
# 143-145     3   Spare
# 146         1   Assigned
# 147         1   RAIM flag
# 148-167    20   Radio Status

  two_max_handler ("Altitude = %d", (oct ("0b".substr ($message,38,12)))," m",4094,4095);
  two_max_handler ("Airspeed over ground (SOG) = %d", (oct ("0b".substr ($message,50,10)))," kn",1022,1023);
  $tyui = substr ($message,60,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,61,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,89,27)) / 600000),91);
  default_handler ("Course over ground (COG) = %05.1f", oct ("0b".substr ($message,116,12)) / 10, "\xB0", 360);
  timestamp_handler (oct ("0b".substr ($message,128,6)));
  printf "Regional reserved = %d (Unused; should be 0)\n", oct ("0b".substr ($message,134,8));
  $tyui = substr ($message,142,1); printf "Data terminal equipment (DTE) = %d (%s)\n", $tyui, $dte [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,143,3));
  $tyui = substr ($message,146,1); printf "Assigned-mode flag = %d (%s)\n", $tyui, $assigned [$tyui];
  $tyui = substr ($message,147,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];
  comm_state ($type, substr ($message,148,1), substr ($message,149,19));

  return;
  }

# *** Decode_10 decodes Type 10, UTC/Date Inquiry messages ***

sub decode_10
  {
# Parameter list
  my ($message,$msg_len) = @_;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-69     30   Target MMSI
#  70-71      2   Spare

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));
  print "Target "; mmsi_handler (oct ("0b".substr ($message,40,30)));
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,70,2));

  return;
  }

# *** Decode_1214 decodes Type 12, Addressed Safety-Related Message and Type 14, Safety-Related Broadcast messages ***

sub decode_1214
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my ($addressed_flag, $tyui);

################################################
#  = TYPE 12 ==   = TYPE 14 ==
#  Bits     Len   Bits     Len  Description
#  38-39      2   38-39      2   Sequence Number (12)/Spare (14)
#  40-69     30                  Destination MMSI
#  70         1                  Retransmit flag
#  71         1                  Spare
#  72-1007  936   40-1007  966   Message (<=156 [12]/<=161 [14] six-bit characters)

# Set addressed_flag to accomodate the extra 32 bit field

  if ($type == 12)
      {
      $addressed_flag = 1;
      printf "Sequence number = %d\n", oct ("0b".substr ($message,38,2));
      print "Target "; mmsi_handler (oct ("0b".substr ($message,40,30)));
      $tyui = oct ("0b".substr ($message,70,1)); printf "Retransmit flag = %d (%s)\n", $tyui, $retransmit [$tyui];
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,71,1));
      }
    else
      {
      $addressed_flag = 0;
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));
      }

  printf "Message text = %s\n", decode_string (substr ($message,40+($addressed_flag*32),min($msg_len-40,966-($addressed_flag*30))));

  return;
  }

# *** Decode_15 decodes Type 15, Interrogation messages ***

sub decode_15
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-39      2   Spare (set to 0)
#  40-69     30   Interrogated MMSI 1
#  70-75      6   Message type 1.1
#  76-87     12   Slot offset 1.1         <-- End here if no add'l. queries
#  88-89      2   Spare (set to 0)
#  90-95      6   Message type 1.2
#  96-107    12   Slot offset 1.2
# 108-109     2   Spare (set to 0)        <-- End here if no 2nd MMSI (see notes below)
# 110-139    30   Interrogated MMSI 2
# 140-145     6   Message type 2.1
# 146-157    12   Slot offset 2.2
# 158-159     2   Spare (set to 0)

## SPECIAL NOTE: AIS messages are supposed to be 8-bit aligned. When assembled as a sentence, they need two additional padding bits for 6-bit alignment...

# There are four possible formats:
# 1) One MMSI and one interrogation/slot offset (message length = 88 bits; 90 with padding)
# 2) One MMSI and two interrogations/slot offsets (message length = 110 bits but might be 112 bits for 8-bit alignment; 114 with padding)
# 3) Two MMSIs each with one interrogation/slot offset (message length = 160 bits, where type 1.2 and offset 1.2 are set to 0; 162 with padding)
# 4) Two MMSIs where the first has two interrogations/slot offsets (message length = 160 bits; 162 with padding)

# All formats start with these fields

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));
  print "1st interrogated "; mmsi_handler (oct ("0b".substr ($message,40,30)));
  $tyui = oct ("0b".substr ($message,70,6)); printf "Message type 1.1 = %d (%s)\n", $tyui, $message_type [$tyui];
  printf "Slot offset 1.1 = %d\n", (oct ("0b".substr ($message,76,12)));

# If length > 90 bits, there is something in the type 1.2 and offset 1.2 fields, but it might be 0! This starts to cover formats 2-4

  if ($msg_len > 90)
    {
    printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,88,2));
    $tyui = oct ("0b".substr ($message,90,6));
    printf "Message type 1.2 = %d", $tyui;
    if ($tyui == 0)
        { print " (unused)\n"; }
      else
        { printf " (%s)\n", $message_type [$tyui]; }
    printf "Slot offset 1.2 = %d", (oct ("0b".substr ($message,96,12)));
    if ($tyui == 0)
        { print " (unused)\n"; }
      else
        { print "\n"; }
    printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,108,2));
  
# We get here if the message length is 162 and we have a second MMSI to process

    if ($msg_len == 162)
      {
      print "2nd interrogated "; mmsi_handler (oct ("0b".substr ($message,110,30)));
      $tyui = oct ("0b".substr ($message,140,6)); printf "Message type 2.1 = %d (%s)\n", $tyui, $message_type [$tyui];
      printf "Slot offset 2.1 = %d\n", (oct ("0b".substr ($message,146,12)));
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,158,2));
      }
    }

  return;
  }

# *** Decode_16 decodes Type 16, Assignment Mode Command messages ***

sub decode_16
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$increment,$n,$offset,$tyui);

################################################
#  Bits     Len   Description
#  38-39      2   Spare (set to 0)
#  40-69     30   Destination A MMSI
#  70-81     12   Offset A
#  82-91     10   Increment A            <-- End here if no add'l. queries
#  92-121    30   Destination B MMSI
# 122-133    12   Offset B
# 134-143    10   Increment B

# This message might have one ($n=0)or two ($n=1) sets of MMSI/offset/increment parameters

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));

  $n = 0;
  if ($msg_len > 96) { $n = 1; }
  
  for ($i=0; $i<=$n; $i++)
    {
    printf "Destination %c ", 65+$i; mmsi_handler (oct ("0b".substr ($message,40+($i*52),30)));
     
# Need to know the increment value in order to interpret offset correctly

    $offset = oct ("0b".substr ($message,70+($i*52),12));
    $increment = oct ("0b".substr ($message,82+($i*52),10));
     
    if ($increment == 0)
        {
        print " Offset = $offset reports per 10-minute interval";
        if ($offset >= 20 && $offset <= 600 && ($offset % 20) == 0)
            { print "\n"; }
          else
            {
            INTERVAL:
              {
              $offset == 0 && do
                { $tyui = 20; last INTERVAL; };
              $offset > 0 && $offset < 600 && do
                { $tyui = int (($offset+20)/20)*20; last INTERVAL; };
              $offset > 600 && do
                { $tyui = 600; last INTERVAL; };
              }
            print " (Invalid value; receiver should use $tyui)\n";
            }
         
        print " Increment = 0 (Interpret offset as number of reports per time interval)\n";
        }
      else
        {
        print " Offset from current slot to first assigned slot = $offset\n";
        if ($increment <= 7)
            { printf " Increment = %d (%s)\n", $increment, $slot_inc [$increment]; }
          else
            { print " Increment = $increment (invalid)\n"; }
        }    
    }
      
  return;
  }

# *** Decode_17 decodes Type 17, DGNSS Broadcast Binary Messages ***

sub decode_17
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i,$n,$tyui,$valid);
  my @M823_Message_Types = (
    [1, "GPS Differential GNSS corrections (full set of satellites)"],
    [3, "GPS Reference station parameters"],
    [4, "GPS/GLONASS Reference Station Datum"],
    [5, "GPS Constellation health"],
    [6, "GPS Null frame"],
    [7, "GPS Radio beacon almanacs"],
    [9, "GPS Subset differential GNSS corrections"],
    [16, "GPS Special message"],
    [27, "GPS/GLONASS Extended radio beacon almanac"],
    [31, "GLONASS Differential GNSS corrections (full set of satellites)"],
    [32, "GLONASS Reference station parameters"],
    [33, "GLONASS Constellation health"],
    [35, "GLONASS Radio beacon almanacs"],
    [36, "GLONASS Special message"],
    );
  my $n_types = scalar @M823_Message_Types;
  my @M823_Health = (
    "Used by the service provider for healthy broadcasts", #0
    "Used by the service provider for healthy broadcasts", #1
    "Used by the service provider for healthy broadcasts", #2
    "Used by the service provider for healthy broadcasts", #3
    "Used by the service provider for healthy broadcasts", #4
    "Used by the service provider for healthy broadcasts", #5
    "Used by user equipment to indicate that the transmission is unmonitored", #6
    "Used by user equipment to indicate that the reference station is not working properly", #7
    );
    
################################################
#  Bits     Len   Description
#  38-39      2   Spare (set to 0)
#  40-57     18   Longitude
#  58-74     17   Latitude
#  75-79      5   Spare (set to 0)
#  80-815   736   Payload (M.823 message)

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,40,18)) / 600),181, "Latitude", (signed_bin2dec (substr ($message,58,17)) / 600),91);
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,75,5));

# ITU-R M.823 Message format
#************************************************************
# TYPE 17   M.823
#  Bits     Bits      Len   Description
#  80-85     0-5        6   Message Type
#  86-95     6-15      10   Station Identifier
#  96-108   16-28      13   Z Count
# 109-111   29-31       3   Sequence Number
# 112-116   32-36       5   Number of DGNSS Data Words (N)
# 117-119   37-39       3   Station Health
# 120-815   40-736  0-696   GNSS data words (in 24-bit increments)
  
  print "                    --- Message Payload (ITU-R Rec. M.823) ---\n";

# Rec. M.823 message handling is wierd because a) the messages are not sequentially numbered and b) there are two interpretatons of type 34 (dependenpent upon N)

  $tyui = oct ("0b".substr ($message,80,6));
  $n = oct ("0b".substr ($message,112,5));
  printf "Message type = %d", $tyui;
  M823_MESSAGES:
    {
    $tyui == 34 && ($n == 0 || $n == 1) && do
      {
      print " (GLONASS Null frame (0 or 1 data words))\n";
      last M823_MESSAGES;
      };
    $tyui == 34 && $n > 1 && do
      {
      print " (GLONASS Subset differential GNSS corrections (>1 data word))\n";
      last M823_MESSAGES;
      };
    $tyui != 34 && do
      {
      $valid = 0;
      for ($i=0; $i< $n_types; $i++)
        {
        if ($tyui == $M823_Message_Types [$i][0])
          {
          printf " (%s)\n", $M823_Message_Types [$i][1];
          $valid = 1;
          last;
          }
        }
      if ($valid == 0) { print " (Invalid message type)\n"; };
      last M823_MESSAGES;
      };
    };  
  
  printf "Station identifier = %d\n", oct ("0b".substr ($message,86,10));
  $tyui = oct ("0b".substr ($message,96,13)); printf "Z count = %d 0.6-second units (%.1f sec, %.2f min)\n", $tyui, $tyui * 0.6, $tyui * 0.01;
  printf "Sequence number = %d\n", oct ("0b".substr ($message,109,3));
  printf "Number of DGNSS data words = %d\n", $n;
  $tyui = oct ("0b".substr ($message,117,3));
  printf "Station health = %d (%s)\n", $tyui, $M823_Health [$tyui];
  
  if ($n > 0)
    {
    for ($i=0; $i<$n; $i++)
      {
      printf "Data word #%2d -- %s\n", $i+1, substr ($message,120+($i*24),24);
      }
    }
  
  return;
  }

# *** Decode_18 decodes Type 18, Standard Class B CS Position Report messages ***

sub decode_18
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-45      8   Regional Reserved
#  46-55     10   Speed Over Ground (SOG)
#  56         1   Position Accuracy
#  57-84     28   Longitude
#  85-111    27   Latitude
# 112-123    12   Course Over Ground (COG)
# 124-132     9   True Heading (HDG)
# 133-138     6   Time Stamp
# 139-140     2   Regional Reserved
# 141         1   CS Unit
# 142         1   Display Flag
# 143         1   DSC Flag
# 144         1   Band Flag
# 145         1   Message 22 Flag
# 146         1   Assigned
# 147         1   RAIM flag
# 148-167    20   Radio Status

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  printf "Regional reserved = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,8));
  two_max_handler ("Speed over ground (SOG) = %.1f", (oct ("0b".substr ($message,46,10)) / 10)," kn",102.2,102.3);
  $tyui = substr ($message,56,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,57,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,85,27)) / 600000),91);
  default_handler ("Course over ground (COG) = %05.1f", oct ("0b".substr ($message,112,12)) / 10, "\xB0", 360);
  default_handler ("True heading = %03d", oct ("0b".substr ($message,124,9)), "\xB0", 511);
  timestamp_handler (oct ("0b".substr ($message,133,6)));
  printf "Regional reserved = %d (Unused; should be 0)\n", oct ("0b".substr ($message,139,2));
  $tyui = substr ($message,141,1); printf "Carrier sense (CS) unit flag = %d (%s)\n", $tyui, $CS [$tyui];
  $tyui = substr ($message,142,1); printf "Display flag = %d (%s)\n", $tyui, $display [$tyui];
  $tyui = substr ($message,143,1); printf "Digital selective calling (DSC) flag = %d (%s)\n", $tyui, $DSC [$tyui];
  $tyui = substr ($message,144,1); printf "Band flag = %d (%s)\n", $tyui, $band [$tyui];
  $tyui = substr ($message,145,1); printf "Message 22 flag = %d (%s)\n", $tyui, $msg22 [$tyui];
  $tyui = substr ($message,146,1); printf "Assigned-mode flag = %d (%s)\n", $tyui, $assigned [$tyui];
  $tyui = substr ($message,147,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];
  comm_state ($type, substr ($message,148,1), substr ($message,149,19));

  return;
  }

# *** Decode_19 decodes Type 19, Extended Class B CS Position Report messages ***

sub decode_19
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-45      8   Regional Reserved
#  46-55     10   Speed Over Ground (SOG)
#  56         1   Position Accuracy
#  57-84     28   Longitude
#  85-111    27   Latitude
# 112-123    12   Course Over Ground (COG)
# 124-132     9   True Heading (HDG)
# 133-138     6   Time Stamp
# 139-142     4   Regional Reserved
# 143-262   120   Name
# 263-270     8   Ship Type
# 271-279     9   Dimension to Bow
# 280-288     9   Dimension to Stern
# 289-294     6   Dimension to Port
# 295-300     6   Dimension to Starboard
# 301-304     4   Type of EPFD
# 305         1   RAIM flag
# 306         1   DTE
# 307         1   Assigned
# 308-311     4   Spare

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  printf "Regional reserved = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,8));
  two_max_handler ("Speed over ground (SOG) = %.1f", (oct ("0b".substr ($message,46,10)) / 10)," kn",102.2,102.3);
  $tyui = substr ($message,56,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,57,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,85,27)) / 600000),91);
  default_handler ("Course over ground (COG) = %05.1f", oct ("0b".substr ($message,112,12)) / 10, "\xB0", 360);
  default_handler ("True heading = %03d", oct ("0b".substr ($message,124,9)), "\xB0", 511);
  timestamp_handler (oct ("0b".substr ($message,133,6)));
  printf "Regional reserved = %d (Unused; should be 0)\n", oct ("0b".substr ($message,139,4));
  printf "Vessel name = %s\n", decode_string (substr ($message,143,120));
  $tyui = oct ("0b".substr ($message,263,8)); printf "Ship Type = %d (%s)\n", $tyui, ship_type ($tyui);
  dimension_handler ("bow", oct ("0b".substr ($message,271,9)), 511);
  dimension_handler ("stern", oct ("0b".substr ($message,280,9)), 511);
  dimension_handler ("port", oct ("0b".substr ($message,289,6)), 63);
  dimension_handler ("starboard", oct ("0b".substr ($message,295,6)), 63);
  $tyui = oct ("0b".substr ($message,301,4)); printf "EPFD type = %d (%s)\n", $tyui, $epfd_type [$tyui];
  $tyui = substr ($message,305,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];
  $tyui = substr ($message,306,1); printf "Data terminal equipment (DTE) = %d (%s)\n", $tyui, $dte [$tyui];
  $tyui = substr ($message,307,1); printf "Assigned-mode flag = %d (%s)\n", $tyui, $assigned [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,308,4));

  return;
  }

# *** Decode_20 decodes a Type 20, Data Link Management Message ***

sub decode_20
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($i, $fmt, $numblocks, $pos);

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-51     12   Offset number 1
#  52-55      4   Reserved slots
#  56-58      3   Time-out
#  59-69     11   Increment
#  70-81     12   Offset number 2
#  82-85      4   Reserved slots
#  86-88      3   Time-out
#  89-99     11   Increment
# 100-111    12   Offset number 3
# 112-115     4   Reserved slots
# 116-118     3   Time-out
# 119-129    11   Increment
# 130-141    12   Offset number 4
# 142-145     4   Reserved slots
# 146-148     3   Time-out
# 149-159    11   Increment

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));

  $numblocks = int (($msg_len - 40) / 30);
  for ($i=0; $i<$numblocks; $i++)
    {
    $fmt = "Offset #" . ($i+1) . " = \%d";
    default_handler ($fmt, oct ("0b".substr ($message,40+($i*30),12)), "", 0);
    default_handler ("Number of reserved slots = %d", oct ("0b".substr ($message,52+($i*30),4)), "", 0);
    default_handler ("Time-out = %d", oct ("0b".substr ($message,56+($i*30),3)), " min.", 0);
    default_handler ("Repeat increment = %d", oct ("0b".substr ($message,59+($i*30),11)), "", 0);
    }

  return;
  }

# *** Decode_21 decodes Type 21, Aid-to-Navigation Report messages ***

sub decode_21
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my ($atontype, $name_ext_len, $timestamp);
  my ($tyui,$x);

################################################
#  Bits     Len   Description
#  38-42      5   Aid type
#  43-162   120   Name
# 163         1   Position Accuracy
# 164-191    28   Longitude
# 192-218    27   Latitude
# 219-227     9   Dimension to Bow
# 228-236     9   Dimension to Stern
# 237-242     6   Dimension to Port
# 243-248     6   Dimension to Starboard
# 249-252     4   Type of EPFD
# 253-258     6   UTC seconds
# 259         1   Off-Position Indicator
# 260-267     8   AtoN Status
# 268         1   RAIM flag
# 269         1   Virtual-aid flag
# 270         1   Assigned
# 271         1   Spare
# 272-360    88   Name Extension

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  $atontype = oct ("0b".substr ($message,38,5)); printf "AtoN type = %d (%s)\n", $atontype, $aton_type [$atontype];
  printf "AtoN name = %s\n", decode_string (substr ($message,43,120));
  $tyui = substr ($message,163,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,164,28)) / 600000),181, "Latitude", (signed_bin2dec (substr ($message,192,27)) / 600000),91);
  dimension_handler ("bow", oct ("0b".substr ($message,219,9)), 511);
  dimension_handler ("stern", oct ("0b".substr ($message,228,9)), 511);
  dimension_handler ("port", oct ("0b".substr ($message,237,6)), 63);
  dimension_handler ("starboard", oct ("0b".substr ($message,243,6)), 63);
  $tyui = oct ("0b".substr ($message,249,4)); printf "EPFD type = %d (%s)\n", $tyui, $epfd_type [$tyui];
  timestamp_handler (oct ("0b".substr ($message,253,6)));
  $tyui = substr ($message,259,1); printf "Off-position indicator = %d (%s)\n", $tyui, $offpos [$tyui];
  if ($atontype < 20 || $timestamp > 59)
      { print " (Valid only for floating AtoNs and if timestamp between 0-59)\n"; }
    else
      {
      printf " (%s)\n", $aton_pos [$tyui];
      }

  printf "AtoN status = %d (default is 0)\n", oct ("0b".substr ($message,260,8));
  $tyui = oct ("0b".substr ($message,260,8)); print "Internal AtoN system status = $tyui\n";
  $x = $tyui % 2; printf "  ** Unit status: %d (%s)\n", $x, $aton_status_unit [$x];
  $x = int ($tyui/2) % 4; printf "  ** Light status: %d (%s)\n", $x, $aton_status_light [$x];
  $x = int ($tyui/8); printf "  ** Radar beacon (RACON) status: %d (%s)\n", $x, $aton_status_racon [$x];
  $tyui = substr ($message,268,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];
  $tyui = substr ($message,269,1); printf "Virtual AtoN flag = %d (%s)\n", $tyui, $vaton [$tyui];
  $tyui = substr ($message,270,1); printf "Assigned-mode flag = %d (%s)\n", $tyui, $assigned [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,271,1));

# Message extension in use if length > 272...

  if ($msg_len > 272)
      {
      $name_ext_len = int (($msg_len - 272) / 6);
      printf "AtoN name extension = %s\n", decode_string (substr ($message,272,$name_ext_len));
      printf "** Full AtoN name = %s%s **\n", decode_string (substr ($message,43,120)), decode_string (substr ($message,272,$name_ext_len));
      }
    else
      { print "No AtoN name extension field\n"; }

  return;
  }

# *** Decode_22 decodes Type 22, Channel Management messages ***

sub decode_22
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $i;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-51     12   Channel A
#  52-63     12   Channel B
#  64-67      4   Tx/Rx mode                    ** Not currently supported; set to 0
#  68         1   Power                         ** Not currently supported; set to 0

#  If Addressed (bit 139) = 0
#  69-86     18   NE Longitude
#  87-103    17   NE Latitude
# 104-121    18   SW Longitude
# 122-138    17   SW Latitude

#  If Addressed (bit 139) = 1
#  69-98     30   MMSI_1                        ** Not currently supported
#  99-103     5   Spare                         ** Not currently supported
# 104-133    30   MMSI_2                        ** Not currently supported
# 134-138     5   Spare                         ** Not currently supported

# 139         1	  Addressed                     ** Not currently supported; set to 0
# 140         1   Channel A Band                ** Not currently supported; set to 0
# 141         1   Channel B Band                ** Not currently supported; set to 0
# 142-144     3   Zone size                     ** Not currently supported; set to 5
# 145-167    23   Spare

# NOTE: Longitude and latitude are scaled by multiplying by 600

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));

  for ($i=0; $i<2; $i++)
    {
    $tyui = oct ("0b".substr ($message,40+($i*12),12));
    printf "Channel %s = %d", chr (65+$i), $tyui;
    if ($tyui == 2087)
        { print " (Channel 87B, 161.975 MHz)\n"; }
      else
        {
        if ($tyui == 2088)
            { print " (Channel 88B, 162.025 MHz)\n"; }
          else
            { print " (Unknown channel)\n"; }
        }
    } 

  $tyui = oct ("0b".substr ($message,64,4)); printf "Tx/Rx mode = %d (%s)\n", $tyui, $txrx_mode [$tyui];
  $tyui = substr ($message,68,1); printf "Power = %d (%s)\n", $tyui, $power [$tyui];

  $tyui = substr ($message,139,1);
  if ($tyui == 0)
      {
      long_lat_handler ("Northeast longitude", (signed_bin2dec (substr ($message,69,18)) / 600),181,
                        "Northeast latitude", (signed_bin2dec (substr ($message,87,17)) / 600),91);
      long_lat_handler ("Southwest longitude", (signed_bin2dec (substr ($message,104,18)) / 600),181, 
                        "Southwest Latitude", (signed_bin2dec (substr ($message,122,17)) / 600),91);
      }
    else
      {
      print "First "; mmsi_handler (oct ("0b".substr ($message,69,30)));
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,99,5));
      print "Second "; mmsi_handler (oct ("0b".substr ($message,104,30)));
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,134,5));
      }

  printf "Addressed mode = %d (%s)\n", $tyui, $addressed [$tyui];

  for ($i=0; $i<2; $i++)
    {
    $tyui = substr ($message,140+$i,1);
    printf "Channel %s Band = %d\n", chr (65+$i), $tyui, $channel_band [$tyui];
    } 

  printf "Transition zone size = %d nm\n", oct ("0b".substr ($message,142,3));
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,145,23));

  return;
  }

# *** Decode_23 decodes Type 23, Group Assignment Command messages ***

sub decode_23
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38-39      2   Spare
#  40-57     18   NE Longitude
#  58-74     17   NE Latitude
#  75-92     18   SW Longitude
#  93-109    17   SW Latitude
# 110-113     4   Station Type
# 114-121     8   Ship Type
# 122-143    22   Spare
# 144-145     2   Tx/Rx Mode
# 146-149     4   Report Interval
# 150-153     4   Quiet Time
# 154-159     6   Spare

# NOTE: Longitude and latitude are scaled by multiplying by 600,000

  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,38,2));
  long_lat_handler ("Northeast longitude", (signed_bin2dec (substr ($message,40,18)) / 600),181,
                    "Northeast latitude", (signed_bin2dec (substr ($message,58,17)) / 600),91);
  long_lat_handler ("Southwest longitude", (signed_bin2dec (substr ($message,75,18)) / 600),181, 
                    "Southwest Latitude", (signed_bin2dec (substr ($message,93,17)) / 600),91);
  $tyui = oct ("0b".substr ($message,110,4)); printf "Station = %d (%s)\n", $tyui, $station_type [$tyui];
  $tyui = oct ("0b".substr ($message,114,8)); printf "Ship type = %d (%s)\n", $tyui, ship_type ($tyui);
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,122,22));
  $tyui = oct ("0b".substr ($message,144,2)); printf "Tx/Rx mode = %d (%s)\n", $tyui, $txrx_mode [$tyui];
  $tyui = oct ("0b".substr ($message,146,4)); printf "Report interval = %d (%s)\n", $tyui, $interval [$tyui];
  printf "Quiet time = %d", oct ("0b".substr ($message,150,4));
    if (oct ("0b".substr ($message,150,4)) == 0)
        { print " (None)\n"; }
      else
        { print " min.\n"; } ;
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,154,6));

  return;
  }

# *** Decode_24 decodes a Type 24, Static Data Report ***
# Type 24 has two formats, part A (160 bits) and part B (168 bits)

###### *** SPECIALLY HANDLE 24A as it might be 160 bits or 168 (if latter, last 8 bits are 0).

# Actually, Type 24 has 2.5 formats because part B messags may have two interpretations!

sub decode_24
  {
# Parameter list
  my ($mmsi,$message,$msg_len) = @_;
  my ($part,$shiptype);

################################################
#  Bits     Len   Description
#  38-39      2   Part Number (0=A,1=B)

$part = oct ("0b".substr ($message,38,2));
printf "Part = %d (Format %s)\n", $part, chr (65+$part);

PART24:
  {
  $part == 0 && do
    {

# Part = 0 => Format A (Length = 160 *or* might be 168)
# Standard says 168 bits but practice is to truncate after the vessel name...

################################################
#  40-159   120   Vessel Name (6-bit characters)
# 160-167     8   Spare                          <--- maybe, maybe not...

    printf "Vessel name = %s\n", decode_string (substr ($message,40,120));
    if ($msg_len == 168)
      {
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,160,8));
      }
    last PART24;
    };

  $part == 1 && do
    {

# Part = 1 => Format B (Length = 168)

################################################
#  40-47      8   Ship Type
#  48-65     18   Vendor ID (6-bit characters)
#  66-69      4   Unit Model Code
#  70-89     20   Serial Number
#  90-131    42   Call Sign (6-bit characters)

# 132-140     9   Dimension to Bow
# 141-149     9   Dimension to Stern
# 150-155     6   Dimension to Port
# 156-161     6   Dimension to Starboard
	
# 132-161    30   Mothership MMSI (if MMSI starts with "98")
	
# 162-167     6   Spare

    $shiptype = oct ("0b".substr ($message,40,8));
    printf "Ship Type = %d (%s)\n", $shiptype, ship_type ($shiptype);
    printf "Vendor ID = %s\n", decode_string (substr ($message,48,18));
    printf "Unit Model Code = %02d\n", oct ("0b".substr ($message,66,4));
    printf "Serial Number = %07d\n", oct ("0b".substr ($message,70,20));
    printf "Call Sign = %s\n", decode_string (substr ($message,90,42));

    if (substr ($mmsi,0,2) eq "98")
        {
        print "Mothership ";
        mmsi_handler (oct ("0b".substr ($message,132,30)));
        }
      else
        {
        dimension_handler ("bow", oct ("0b".substr ($message,132,9)), 511);
        dimension_handler ("stern", oct ("0b".substr ($message,141,9)), 511);
        dimension_handler ("port", oct ("0b".substr ($message,150,6)), 63);
        dimension_handler ("starboard", oct ("0b".substr ($message,156,6)), 63);
        }

    printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,162,6));

    last PART24;
    };

# If we get here, we have an invalid part type...

  do
    {
    print "\n***** Invalid part format type: $part *****\n";
    exit;
    };
  };

return;
}

# *** Decode_2526 decodes Type 25: Single Slot Binary Message and Type 26: Multiple Slot Binary Message ***
# There is also a special Type 26 - "R-Mode Navigation Message" (DAC = 1, FI = 10, 11, 14)

sub decode_2526
  {
# Parameter list
  my ($type,$message,$msg_len) = @_;
  my ($tyui,$bin_len,$dest_ind,$pad,$r,$str_len,$structure,$tdacfi);
  my $start_pos = 40;
  
# Binary data flag

  my @bd_flag = (
    "Unstructured binary data", #0
    "Binary data coded per Application Identifier", #1
    );

# List of Message Type (T), DAC, and FI for the Type 25 and 26 Application Identifier values
#    from https://www.iala-aism.org/asm/ (25 Nov 2022)

  my @APP_ID = (
    [ 2500100, "Single Slot Binary Message"],
    [ 2600100, "Multiple Slot Binary Message"],
    [ 2600110, "R-Mode Navigation Message - 1 slot"],
    [ 2600111, "R-Mode Navigation Message - 2 slots"],
    [ 2600114, "R-Mode Navigation Message - 5 slots"],
     );
  my $app_id_len = scalar @APP_ID;
  
# List the length of Gold Code and number of spare bits for Type 26, DAC=1, FI=10/11/14

  my @r_mode = ( [80,4], [300,8], [0,0], [0,0], [968,12] );
    
# TYPE 25
################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI
#  38         1   Destination Indicator
#  39         1   Binary Data flag
#
# Bits 38-39=00  Bits 38-39=01  Bits 38-39=10  Bits 38-39=11
#  Bits     Len   Bits     Len   Bits     Len   Bits     Len
#                                40-69     30   40-69     30  Destination MMSI
#                                70-71      2   70-71      2  Spare (00)
#                 40-49     10                  72-81     10  Application ID (DAC)
#                 50-55      6                  82-87      6  Application ID (FI)
#  40-167   128   56-167   112   72-167    96   88-167    80  Binary data (variable length, 0-128)

# TYPE 26
################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI
#  38         1   Destination Indicator
#  39         1   Binary Data flag
#
# Bits 38-39=00  Bits 38-39=01  Bits 38-39=10  Bits 38-39=11
#  Bits     Len   Bits     Len   Bits     Len   Bits     Len
#                                40-69     30   40-69     30  Destination MMSI
#                                70-71      2   70-71      2  Spare (00)
#                 40-49     10                  72-81     10  Application ID (DAC)
#                 50-55      6                  82-87      6  Application ID (FI)
#  40-1039 1000   56-1039  984   72-1039  968   88-1039  952  Binary data (variable length, 0-1000)
#  1040-1043  4   1040-1043  4   1040-1043  4   1040-1043  4  Spare
#  1044-1063 20   1044-1063 20   1044-1063 20   1044-1063 20  Radio Status   ** Not currently supported; set to 11100000000000000110

# TYPE 26 -- R-mode special message
################################################
#  Bits     Len   Description
#   0-5       6   Message Type
#   6-7       2   Repeat Indicator
#   8-37     30   MMSI
#  38         1   Destination Indicator (0)
#  39         1   Binary Data flag (1)
#  40-49     10   Application Identifier (DAC)
#  50-55      6   Application Identifier (FI)
#  56-63      8   Station Health
#
#     FI=10          FI=11          FI=14
#  Bits     Len   Bits     Len   Bits     Len   Description
#  64-143    80   64-363   300   64-1031  968   Gold Code
# 144-147     4  364-371     8  1032-1043  12   Spare
# 148-167    20  372-391    20  1044-1063  20   Radio Status (set to 0)

# start_pos is the starting position in the message where the binary data begins. It starts at 40 and will be adjusted if there
# is a target MMSI (Destination Indicator = 1) and/or structured data (Binary Data flag = 1)

  $dest_ind = substr ($message,38,1); printf "Destination Indicator = %d (%s)\n", $dest_ind, $addressed [$dest_ind];
  $structure = substr ($message,39,1); printf "Binary Data flag = %d (%s)\n", $structure, $bd_flag [$structure];
  
# Handle target MMSI, if present

  if ($dest_ind == 1)
    {
    print "Target ";
    mmsi_handler (oct ("0b".substr ($message,$start_pos,30)));
    printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,$start_pos+30,2));
    $start_pos += 32;
    }

  $tdacfi = 0;
  if ($structure == 1)
    {
    $dac = oct ("0b".substr ($message,$start_pos,10));
    $fi = oct ("0b".substr ($message,$start_pos+10,6));
    $tdacfi = $type * 100000 + $dac * 100 + $fi;
    for ($i=0; $i<$app_id_len; $i++)
      { if ($tdacfi == $APP_ID [$i][0]) { last; } }

    print "Application Identifier: ";
    if ($i < $app_id_len)
       { printf "%s", $APP_ID [$i][1]; }
     else
       { print "unknown"; }
    print " (DAC = $dac/FI = $fi)\n";
    $start_pos += 16;
    };

# Standard Types 25 and 26 now have a binary data field. the R-mode messages have their own format

  if ($tdacfi != 2600110 && $tdacfi != 2600111 && $tdacfi != 2600114)

# Handle standard Type 25 and 26 messages here
# Binary Data field is variable; length is message_length - starting_position - 24 (Type 26 only)

      {
      $bin_len = $msg_len - $start_pos - (($type-25) * 24);
      $pad = $bin_len % 6;
      printf "6-bit ASCII text string = %s\n", decode_string (substr ($message,$start_pos,$bin_len-$pad));
      display_ascii ($message,$start_pos,$bin_len-$pad,128);
      if ($type == 26)
        {
        printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,-24,4));
        comm_state ($type, substr ($message,-20,1), substr ($message,-19,19));
        }
      }
      
# Handle R-mode messages here

    else
      {
      printf "Station Health = %d (Not yet specified; should be 0)\n", oct ("0b".substr ($message,56,8));
      $str_len = $r_mode [$fi-10][0];
      printf "Gold Code = %s\n", decode_string (substr ($message,64,$str_len));
      display_ascii ($message,64,$str_len,128);
      printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,64+$str_len,$r_mode [$fi-10][1]));
      $start_pos = 64 + $str_len + $r_mode [$fi-10][1];
      comm_state ($type, substr ($message,$start_pos,1), substr ($message,$start_pos+1,19));
      };

  return;
  }

# *** Decode_27 decodes Type 27, Long Range AIS Broadcast Messages ***

sub decode_27
  {
# Parameter list
  my ($message,$msg_len) = @_;
  my $tyui;

################################################
#  Bits     Len   Description
#  38         1   Position Accuracy
#  39         1   RAIM flag
#  40-43      4   Navigation Status
#  44-61     18   Longitude
#  62-78     17   Latitude
#  79-84      6   Speed Over Ground (SOG)
#  85-93      9   Course Over Ground (COG)
#  94         1   GNSS Position status
#  95         1   Spare

  $tyui = substr ($message,38,1); printf "Position accuracy = %d (%s)\n", $tyui, $accuracy [$tyui];
  $tyui = substr ($message,39,1); printf "Receiver Autonomous Integrity Monitoring flag = %d (%s)\n", $tyui, $raim [$tyui];
  $tyui = oct ("0b".substr ($message,40,4)); printf "Navigation Status = %d (%s)\n", $tyui, $nav_stat [$tyui];
  long_lat_handler ("Longitude", (signed_bin2dec (substr ($message,44,18)) / 600),181, "Latitude", (signed_bin2dec (substr ($message,62,17)) / 600),91);
  default_handler ("Speed over ground (COG) = %d", oct ("0b".substr ($message,79,6)), " kn", 63);
  default_handler ("Course over ground (COG) = %03d", oct ("0b".substr ($message,85,9)), "\xB0", 511);
  $tyui = substr ($message,94,1); printf "GNSS position status = %d (%s)\n", $tyui, $gnss_position_status [$tyui];
  printf "Spare = %d (Unused; should be 0)\n", oct ("0b".substr ($message,95,1));

  return;
  }

# ********************************************************
# process_sentence
# ********************************************************

# Given a complete AIS sentence, find the talker ID and encapsulation formatter, validate the checksum,
# and pullout the payload for parsing.

sub process_sentence
  {
  my ($line) = @_;
  my $tempmsg; # returned value
  my ($ascii, $ais_sent_type, $char, $fields, $i, $numchars, $nmea_chk, $tyui);
  my @talker = (
    ["AB","NMEA 4.0 Base AIS station"],
    ["AD","MMEA 4.0 Dependent AIS Base Station"],
    ["AI","Mobile AIS station"],
    ["AN","NMEA 4.0 Aid to Navigation AIS station"],
    ["AR","NMEA 4.0 AIS Receiving Station"],
    ["AS","NMEA 4.0 Limited Base Station"],
    ["AT","NMEA 4.0 AIS Transmitting Station"],
    ["AX","NMEA 4.0 Repeater AIS station"],
    ["BS","Base AIS station (deprecated in NMEA 4.0)"],
    ["SA","NMEA 4.0 Physical Shore AIS Station"]
    );
  my @encap = (
    ["ABM","AIS Addressed Binary and Safety Related Message"],
    ["BBM","AIS Broadcast Binary Message"],
    ["MEB","Message Input for Broadcast, Command"],
    ["VDM","AIS VHF Data-Link Message"],
    ["VDO","AIS VHF Date-Link Own-Vessel Report"],
    );

  print "AIS sentence: $line\n";
  
  # $fields[0] = AIS talker ID     $fields[1] = # fragments     $fields[2] = fragment #
  # $fields[3] = serial number     $fields[4] = channel         $fields[5] = Armored ASCII payload
  # $fields[6] = # pad bits        $fields[7] = NMEA checksum

  @fields = split (/[,*]/, $line);
  $ais_sent_type = $fields [0];
  $ascii = $fields [5];
  $num_pad = $fields [6];
  $nmea_chk = $fields [7];

  $tyui = substr ($ais_sent_type,1,2);
  print "  Talker ID: $tyui " ;
  for ($i=0; $i < scalar (@talker); $i++)
    {
    if ($tyui eq $talker [$i][0]) { last; }
    }
  if ($i < scalar (@talker))
      { printf "(%s)\n", $talker [$i][1]; }
    else
      { print "(unknown)\n"; }

  $tyui = substr ($ais_sent_type,-3);
  print "  AIS Encapsulation Formatter: $tyui ", ;
  for ($i=0; $i < scalar (@encap); $i++)
    {
    if ($tyui eq $encap [$i][0]) { last; }
    }
  if ($i < scalar (@encap))
      { printf "(%s)\n", $encap [$i][1]; }
    else
      { print "(unknown)\n"; }

  printf "  No. of fragments: %d    Fragment number: %d    Serial number:", $fields [1], $fields [2];
  if ($fields [3] eq "")
      { print " <null>"; }
    else
      { printf " %d", $fields [3]; }
  printf "    AIS channel: %s\n", $fields [4];
  $numchars = length ($ascii);
  printf "  Payload: %s (%d bits/%d 6-bit words)\n", $ascii, $numchars*6, $numchars;
  print "  Number of padding bits: $num_pad\n";

# Have to do this little playing around because the sentence-based input doesn't always terminate the line properly

  @fields = split (/[*]/, $line);
  $chksum = nmea_checksum (substr($fields[0],-(length($fields[0])-1)));
  print "  Checksum: 0x$nmea_chk";
  if ($chksum == $nmea_chk)
      { print " (verified)\n"; }
    else
      { print " (bit error detected [0x$chksum])\n"; }

# Cycle through characters in string -- get ord (8-bit) in binary

  $tempmsg = "";

  for ($i=0; $i < $numchars; $i++)
    {
    $char = ord (substr ($ascii,$i,1)) - 48;
    if ($char > 40)
      { $char -= 8; }
    $tempmsg .= sprintf ("%06b", $char);
    }

  return $tempmsg;
  }

# ********************************************************
# nmea_checksum
# ********************************************************

# The NMEA checksum is computed on the entire sentence including the AIVDM/AIVDO tag but excluding the leading "!"
# The checksum is merely a byte-by-byte XOR of the sentence

sub nmea_checksum
  {
  my $sentence = $_[0];
  my $len = length ($sentence);
  my $sum = 0;
  my $i;

  for ($i=0; $i<$len; $i++)
    { $sum ^= ord (substr ($sentence,$i,1)); }

# Ensure that $sum is two upper-case hex digits

  $sum = sprintf ("%02X", $sum);
  return $sum;
  }

# ********************************************************
# print_binary_string
# ********************************************************

# Common code to display the binary string but wrap at 96 columns

sub print_binary_string
  {
  my ($message,$len) = @_;
  my ($i,$j);

  print "Binary payload ($len bits):\n";
  $j = 0;

  for ($i=0; $i<$len; $i++)
    {
    printf "%s", substr ($message,$i,1);
    $j++;
    if ($j == 96)
      { print "\n"; $j = 0; }
    }

  print "\n\n";
  return;
  }

# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line to ensure the proper number of parameters

sub parsecommandline
  {

# No arguments --> Display the help message

  if ($#ARGV < 0)
      {
      help ();
      exit;
      }

# The only correct single argument option is -h

  if ($#ARGV == 0)
    {
    if ($ARGV[0] eq "-h")
        {
        help ();
        }
      else
        {
        printf "Invalid argument '%s'.\n", $ARGV[0];
        print "Use -h to get the help file\n\n";
        };
    exit;
    };

# If there are more than two arguments, we have an error...

  if ($#ARGV > 1)
    {
    printf "%d arguments is too many.\n", $#ARGV+1;
    print "Use -h to get the help file\n\n";
    exit;
    };

# We're here because there are exactly two arguments. Make sure that the switch is correct...

  if ($ARGV[0] ne "-b" && $ARGV[0] ne "-bf" && $ARGV[0] ne "-s" && $ARGV[0] ne "-sf")
    {
    printf "Invalid switch '%s'.\n\n", $ARGV[0];
    print "Use -h to get the help file\n\n";
    exit;
    };

return ($ARGV[0], $ARGV[1]);
}

# ********************************************************
# help
# ********************************************************

# Display the help file

sub help
{
print<< "EOT";
Program usage:
  AIS_parser.pl {-bf <filename>} | {-b <string>}
  AIS_parser.pl {-sf <filename>} | {-s <string>}
  AIS_parser.pl -h

 where:
  -b <string>    provide a binary string on the command line
  -bf <filename> name a file containing a single binary string
  -s <string>    provide a complete AIVDM sentence on the command line
  -sf <filename> name a file containing one or more complete AIS sentences,
                 comprising one or more AIS messages
  -h             display the help file

Sample program use:

  AIS_parser.pl -b 00000100100010010101011100101110000000000000000000000
  100001001011100111110010111100100010001010110110001100010100100110110101100
  1011000111010000000000000101000001010000

  AIS_parser.pl -bf test_binary.txt

  AIS_parser.pl -s \\!AIVDM,1,1,,A,18UG\\;P0012G\\?Uq4EdHa\\=c\\;7@051\\@,0*53
    **** NOTE: If you enter an AIVDM sentence directly on the command line,
    **** precede the "!" and non-comma symbols (e.g., ";" or "+") with a "\\".

  AIS_parser.pl -sf test_aivdm.txt

EOT
return;
}
